
Require Import Tactics.
Require Import Dynamic.
Require Import Sequence.
Require Import Relation.
Require Import Hwf.
Require Import History.
Require Import Match.
Require Import Truncate.
Require Import Coq.Wellfounded.Inclusion.


(* Inert forms *)

Lemma vop_inert :
  forall H i i',
    trco H
    -> vop H i i'
    -> exec H i \/ exists i'', po H i i''.
Proof.
intros H i i' Htrco Hvop.
so (vop_execed_fst_or_svop _#3 Htrco Hvop) as [Hexec | Hsop].
  left; assumption.

  right; eexists; eapply sop_po; eauto; done.
Qed.


Lemma vt_inert :
  forall H i i',
    trco H
    -> vt H i i'
    -> exec H i \/ exists i'', po H i i''.
Proof.
intros H i i' Htrco Hvt.
destruct Hvt as (i'' & Hvop & Hxos).
eapply vop_inert; eauto.
Qed.


Lemma wrp_inert :
  forall H i i',
    trco H
    -> wrp H i i'
    -> exec H i \/ exists i'', po H i i''.
Proof.
intros H i i' Htrco Hwrp.
invert Hwrp; intros; eauto using vt_inert.
Qed.


Lemma wwp_inert :
  forall H i i',
    trco H
    -> wwp H i i'
    -> exec H i \/ exists i'', po H i i''.
Proof.
intros H i i' Htrco Hwwp.
invert Hwwp; intros; eauto using vt_inert, wrp_inert.
Qed.


Lemma co_inert :
  forall H i i',
    trco H
    -> co H i i'
    -> (exists i'', H {{ ev_co i i'' }}) \/ exec H i \/ (exists i'', po H i i'').
Proof.
intros H i i' Htrco Hco.
invert Hco; try (intros; eauto using wrp_inert, wwp_inert; done); [].
intros iw Hrwf _ _.
right; left.
destruct Hrwf; eapply rf_execed; eauto; done.
Qed.


Lemma cop_inert :
  forall H i i',
    trco H
    -> cop H i i'
    -> (exists i'', H {{ ev_co i i'' }}) \/ exec H i \/ exists i'', po H i i''.
Proof.
intros H i i' Htrco Hcop.
destruct Hcop as (i'' & Hco & _).
eapply co_inert; eauto; done.
Qed.


Lemma extend_init_nexec :
  forall H i p,
    trco (H; ev_init i p)
    -> exec (H; ev_init i p) i
    -> False.
Proof.
intros H i p Htrco Hexec.
discriminate_exec Hexec; intro Hexec; [].
edestruct trco_init_invert_nexec; eauto; done.
Qed.


Lemma extend_is_nexec :
  forall H i a,
    trco (H; ev_is i a)
    -> exec (H; ev_is i a) i
    -> False.
Proof.
intros H i p Htrco Hexec.
discriminate_exec Hexec; intro Hexec; [].
edestruct trco_is_invert_nexec; eauto; done.
Qed.


Lemma extend_fence_nexec :
  forall H i f,
    trco (H; ev_fence f i)
    -> exec (H; ev_fence f i) i
    -> False.
Proof.
intros H i p Htrco Hexec.
discriminate_exec Hexec; intro Hexec; [].
edestruct trco_fence_invert_nexec; eauto; done.
Qed.


Lemma extend_init_npo :
  forall H i p i',
    trco (H; ev_init i p)
    -> po (H; ev_init i p) i i'
    -> False.
Proof.
intros H i p i' Htrco Hpo.
edestruct po_irreflex; eauto; [].
eapply po_poq_trans; eauto; [].
assert ((H; ev_init i p) {{ ev_init i' p }}) as Hinit.
  eapply po_init_1; eauto; [].
  left; reflexivity.
destruct Hinit as [Heq | Hinit].
  injection Heq; intros; subst; left; reflexivity.

  right.
  eapply (po_from_before _ H nil i'); eauto; [].
  reflexivity.
Qed.


Lemma extend_is_npo :
  forall H i a i',
    trco (H; ev_is i a)
    -> po (H; ev_is i a) i i'
    -> False.
Proof.
intros H i a i' Htrco Hpo.
invert Htrco.
intros Htrco' _ _ (p & H' & ->).
eapply extend_init_npo; eauto; [].
eapply po_tail; eauto; discform.
Qed.


Lemma extend_fence_npo :
  forall H i f i',
    trco (H; ev_fence f i)
    -> po (H; ev_fence f i) i i'
    -> False.
Proof.
intros H i f i' Htrco Hpo.
remember (H; ev_fence f i) as Hh.
revert H i f i' HeqHh Htrco Hpo.
induct Hh.
  intros; discriminate.
intros h H' IH H i f i' HeqHh Htrco Hpo.
injection HeqHh; clear HeqHh; intros; subst H' h.
invert Htrco.
intros Htrco' HH _.
destruct HH as [(a & H' & ->) | (f' & H' & ->)].
  apply (extend_is_npo H' i a i'); auto; [].
  eapply po_tail; eauto; discform.

  apply (IH H' i f' i'); [| |].
    reflexivity.

    prove_trco.

    eapply po_tail; eauto; discform.
Qed.


(* Extend cases *)
(* Most of these are stronger than necessary now (we only need cases pertaining to
   init, is, and fence).  But I already had these from an earlier version of the proof,
   so it's easiest to use them.  But, if these ever become difficult to maintain, we
   should cut them back to size.
*)

Lemma po_extend_cases :
  forall H h i i',
    trco (H; h)
    -> po (H; h) i i'
    -> po H i i' \/ (exists p, h = ev_init i' p).
Proof.
intros H h i i' Htrco Hpo.
destruct Hpo as (H1 & H2 & H3 & p & Heq).
so (app_eq_cons _#5 Heq) as [(H3' & _ & Heq') | (_ & Heq')].
  left.
  unfold po; eauto; done.

  injection Heq'; intros; subst.
  right; eauto; done.
Qed.


Lemma so_extend_cases :
  forall H h b i i',
    trco (H; h)
    -> so b (H; h) i i'
    -> so b H i i'
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h b i i' Htrco Hso.
invert Hso.

(* tag *)
intros T T' Hpo Hfence Hfence' Hedge.
so (edge_fence_tail_1 _#6 Htrco Hfence Hedge) as Hedge'; clear Hedge.
destruct Hfence' as [-> | Hfence'].
  right; right; eauto; done.
destruct Hfence as [-> | Hfence].
  exfalso.
  eapply extend_fence_npo; eauto; done.
so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
  left.
  eapply so_tag; eauto; done.

  right; left; eauto.

(* post *)
intros Hpo Hfence.
destruct Hfence as [-> | Hfence].
  exfalso.
  eapply extend_fence_npo; eauto; done.
so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
  left.
  eapply so_post; eauto; done.

  right; left; eauto.

(* pre *)
intros Hpo Hfence.
destruct Hfence as [-> | Hfence].
  right; right; eauto; done.
so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
  left.
  eapply so_pre; eauto; done.

  right; left; eauto.
Qed.


Lemma xo_extend_cases :
  forall H h i i',
    trco (H; h)
    -> xo (H; h) i i'
    -> xo H i i'
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i1 i2 Htrco Hxo.
invert Hxo; [].
intros b Hso.
#3 so (so_extend_cases _#5 Htrco Hso) as [Hso' | [Hinit | Hfence]].
  left; eapply xo_specified; eassumption.
  
  right; left; assumption.

  right; right; assumption.
Qed.


Lemma xos_extend_cases :
  forall H h i i',
    trco (H; h)
    -> xos (H; h) i i'
    -> xos H i i'
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i i' Htrco Hxos.
induct Hxos.

(* refl *)
intros i.
left; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hxo12 Hxos23 IH.
destruct IH as [Hxos23' | IH].
  so (xo_extend_cases _#4 Htrco Hxo12) as [Hxo12' | HH].
    left; eapply star_step; eauto; done.

    so (star_plus _#4 Hxos23) as [Heq | Hxop].
      subst; right; assumption.

      exfalso.
      so (xop_po _#3 Htrco Hxop) as Hpo.
      destruct HH as [(p & ->) | (f & ->)].
        eapply extend_init_npo; eauto; done.

        eapply extend_fence_npo; eauto; done.

  right; assumption.
Qed.


Lemma to_extend_cases :
  forall H h i i',
    trco (H; h)
    -> to (H; h) i i'
    -> to H i i' \/ (executes h i /\ i <> i' /\ ~ exec H i').
Proof.
intros H h i i' Htrco Hto.
destruct Hto as (H1 & H2 & Heq & Hexec & Hnexec).
so (app_eq_cons _#5 Heq) as [(H2' & _ & Heq') | (_ & Heq')].
  left.
  exists H1, H2'; auto; done.

  subst H1.
  so (exec_cons_invert _#3 Hexec) as [Hexecutes | Hexec'].
    right.
    do2 2 split; auto; [|].
      contradict Hnexec.
      subst; assumption.

      contradict Hnexec.
      apply exec_cons_incl; assumption.

    left.
    exists H, nil.
    do2 2 split; auto; [].
    contradict Hnexec.
    apply exec_cons_incl; assumption.
Qed.


Lemma vo_extend_cases :
  forall H h i i',
    trco (H; h)
    -> vo (H; h) i i'
    -> vo H i i'
       \/ h = ev_rf i i'
       \/ (h = ev_exec i /\ H {{ ev_is i ac_push }} /\ i <> i' /\ ~ exec H i')
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i i' Htrco Hvo.
destruct Hvo as [Hsvo | [Hrf | Hpio]].
(* svo *)
so (so_extend_cases _#5 Htrco Hsvo) as [Hsvo' | Hcond].
  left.
  left; assumption.

  right; right; right; assumption.

(* rf *)
destruct Hrf as [Heq | Hrf].
  right; left; assumption.

  left; right; left; assumption.

(* pio *)
destruct Hpio as (Hisp & Hto).
so (to_extend_cases _#4 Htrco Hto) as [Hto' | (Hexecutes & Hneq & Hnexec)].
  left; right; right.
  split; auto; [].
  destruct Hisp as [Heq | ?]; auto; [].
  exfalso.
  subst h.
  destruct (trco_is_invert_nexec _#3 Htrco); [].
  eapply to_execed_fst; eauto; done.

  right; right; left.
  discriminate_executes Hexecutes; [|].
    intro Heq.
    repeat2 split; auto; [].
    discriminate_in Hisp; intros; subst; auto; discriminate.

    intros iw ->.
    exfalso.
    eapply isr_isp_contra; eauto; [].
    exact (rf_isw_isr _#3 Htrco (or_introl (eq_refl _)) ander).
Qed.


Lemma vos_extend_cases :
  forall H h i i',
    trco (H; h)
    -> vos (H; h) i i'
    -> vos H i i'
       \/ (exists iw ir, h = ev_rf iw ir /\ vos H i iw /\ xos H ir i')
       \/ (exists ip, h = ev_exec ip /\ H {{ ev_is ip ac_push }} /\ vos H i ip /\ ip <> i' /\ ~ exec H i')
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i i' Htrco Hvos.
induct Hvos.

(* refl *)
intros.
left; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hvo12 Hvos23 IH.
so (vo_extend_cases _#4 Htrco Hvo12) as [Hvo12' | [Heqh | [(Heqh & Hisp & Hneq12 & Hnexec2) | Hcond]]].
  destruct IH as [Hvos23' | [Hcond | [Hcond | Hcond]]].
    left; eapply star_step; eauto; done.

    right; left.
    destruct_all Hcond.
    repeat eexists; repeat2 split; eauto; [].
    eapply star_step; eauto; done.

    right; right; left.
    destruct_all Hcond.
    repeat eexists; repeat2 split; eauto; [].
    eapply star_step; eauto; done.

    right; right; right; assumption.

  right; left.
  destruct IH as [Hvos23' | [(iw & ir & Heqh' & Hvos2w & Hxosr3) | Hcond]].
    exists i1, i2.
    repeat2 split; auto; [|].
      apply star_refl; auto; done.

      so (star_plus _#4 Hvos23') as [Heq23 | Hvop23].
        subst; apply star_refl; done.
      subst h.
      apply plus_star; [].
      eapply vop_extend_exec_nisp_xop; eauto using vop_cons_incl; [].
      intro Hisp.
      so (rf_isw_isr _#3 Htrco (or_introl eq_refl)) as (_ & Hisr).
      eapply isr_isp_contra; eauto; [].
      right; assumption.

    rewrite -> Heqh in Heqh'.
    injection Heqh'; intros; subst iw ir.
    exists i1, i2.
    repeat2 split; auto; [].
    apply star_refl; auto; done.

    exfalso.
    do2 2 (destruct Hcond as [Hcond | Hcond]);
    destruct_all Hcond;
    subst; discriminate.

  right; right; left.
  subst h.
  exists i1.
  repeat2 split; auto; [| |].
    apply star_refl; done.

    intro; subst i3.
    destruct (vop_irreflex _ i1 Htrco).
    exists i2; auto; done.

    contradict Hnexec2.
    so (vos_execed_2 _#3 Htrco Hvos23 (exec_cons_incl _#3 Hnexec2)) as Hexec2.
    so (exec_cons_invert _#3 Hexec2) as [Hexecutes | Hexec2'].
      discriminate_executes Hexecutes; [].
      intro Heq12.
      injection Heq12; intros; contradiction.

      assumption.

  right; right; right.
  so (star_plus _#4 Hvos23) as [? | Hvop23].
    subst i3; assumption.

    exfalso.
    so (vop_inert _#3 Htrco Hvop23) as [Hexec | (i4 & Hpo)];
    destruct Hcond as [(? & ?) | (? & ?)]; subst; eauto using extend_fence_nexec, extend_init_nexec, extend_fence_npo, extend_init_npo.
Qed.


Lemma vop_extend_cases :
  forall H h i i',
    trco (H; h)
    -> vop (H; h) i i'
    -> vop H i i'
       \/ (exists iw ir, h = ev_rf iw ir /\ vos H i iw /\ xos H ir i')
       \/ (exists ip, h = ev_exec ip /\ H {{ ev_is ip ac_push }} /\ vos H i ip /\ ip <> i' /\ ~ exec H i')
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i i' Htrco Hvop.
so (vos_extend_cases _#4 Htrco (plus_star _#4 Hvop)) as [Hvos' | Hcond].
  left.
  apply star_neq_plus; auto; [].
  intro; subst i'.
  edestruct vop_irreflex; eauto; done.

  right; assumption.
Qed.


Lemma vt_extend_cases :
  forall H h i i',
    trco (H; h)
    -> vt (H; h) i i'
    -> vt H i i'
       \/ (exists iw ir, h = ev_rf iw ir /\ vos H i iw /\ xos H ir i')
       \/ (exists ip, h = ev_exec ip /\ H {{ ev_is ip ac_push }} /\ vos H i ip /\ ip <> i' /\ ~ exec H i')
       \/ (exists p, h = ev_init i' p)
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i1 i3 Htrco Hvt.
destruct Hvt as (i2 & Hvop & Hxos).
so (xos_extend_cases _#4 Htrco Hxos) as [Hxos' | [Hcond | Hcond]]; auto; [].
so (trco_tail _ _ Htrco) as Htrco'.
so (vop_extend_cases _#4 Htrco Hvop) as [Hvop' | [Hcond | [Hcond | [Hcond | Hcond]]]].
  left.
  eexists; eauto; done.

  right; left.
  destruct_all Hcond; repeat eexists; repeat2 split; eauto.
  eapply star_trans; eauto; done.

  right; right; left.
  destruct Hcond as (ip & -> & Hisp & Hvos & Hneq & Hnexec).
  repeat eexists; repeat2 split; eauto; [|].
    so (star_plus _#4 Hxos) as [? | Hxop].
      subst; assumption.
    intro; subst i3.
    so (xop_execed_2 _#3 Htrco Hxop (@exec_exec (H; ev_exec ip) _ (or_introl eq_refl))) as Hexec.
    discriminate_exec Hexec; [|].
      intro Heq; injection Heq; intro; contradiction.

      intro; contradiction.

    contradict Hnexec.
    eapply xos_execed_2; eauto; done.

  destruct Hcond as (p & ->).
  right; right; right; left.
  exists p; f_equal; [].
  so (star_plus _#4 Hxos) as [? | Hxop].
    assumption.
  exfalso.
  so (xop_po _#3 Htrco Hxop) as Hpo.
  eapply extend_init_npo; eauto; done.

  destruct Hcond as (f & ->).
  right; right; right; right.
  exists f; f_equal; [].
  so (star_plus _#4 Hxos) as [? | Hxop].
    assumption.
  exfalso.
  so (xop_po _#3 Htrco Hxop) as Hpo.
  eapply extend_fence_npo; eauto; done.
Qed.


Lemma wrp_extend_cases :
  forall H h i i',
    trco (H; h)
    -> wrp (H; h) i i'
    -> wrp H i i'
       \/ (exists iw ir, h = ev_rf iw ir /\ vos H i iw /\ xos H ir i')
       \/ (exists ip, h = ev_exec ip /\ H {{ ev_is ip ac_push }} /\ vos H i ip /\ ip <> i' /\ ~ exec H i')
       \/ (exists l, h = ev_is i' (ac_read l))
       \/ (exists l v, h = ev_is i' (ac_rw l v))
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i i' Htrco Hwrp.
invert Hwrp; clear Hwrp.

(* vt *)
intros l Hvt (v & Hwrite) Hread.
#3 so (reads_cons_invert _#4 Hread) as [Hread' | [Heq | (v' & Heq)]].
  #3 so (writes_cons_invert _#5 Hwrite) as [Hwrite' | [Heq | Heq]].
    so (vt_extend_cases _#4 Htrco Hvt) as HH.
    #3 repeat2 (splithyp HH); auto.
      left; eapply wrp_vt; eauto; [].
      eexists; eassumption.

      destruct HH as (p & ->).
      exfalso.
      edestruct trco_init_invert_nis; eauto; [].
      invert Hread'; eauto; done.

      do 5 right; assumption.

    exfalso.
    subst h.
    so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop].
      eapply extend_is_nexec; eauto; done.
  
      eapply (extend_is_npo _ _ _ i'); eauto; [].
      apply xop_po; auto; done.

    exfalso.
    subst h.
    so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop].
      eapply extend_is_nexec; eauto; done.
  
      eapply (extend_is_npo _ _ _ i'); eauto; [].
      apply xop_po; auto; done.

  subst; do 3 right; left; eauto; done.

  subst; do 4 right; left; eauto; done.

(* po *)
intros l Hpo (v & Hwrite) Hread.
#3 so (reads_cons_invert _#4 Hread) as [Hread' | [Heq | (v' & Heq)]].
  #3 so (writes_cons_invert _#5 Hwrite) as [Hwrite' | [Heq | Heq]].
    #2 so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
      left; eapply wrp_po; eauto; [].
      eexists; eassumption.
    
      exfalso.
      edestruct trco_init_invert_nis; eauto; [].
      invert Hread'; eauto; done.

    exfalso.
    subst h.
    eapply (extend_is_npo _ _ _ i'); eauto; done.
    
    exfalso.
    subst h.
    eapply (extend_is_npo _ _ _ i'); eauto; done.
    
  subst; do 3 right; left; eauto; done.

  subst; do 4 right; left; eauto; done.
Qed.


Lemma rwp_extend_cases :
  forall H h i i',
    trco (H; h)
    -> rwp (H; h) i i'
    -> rwp H i i'
       \/ (exists l v, h = ev_is i' (ac_write l v))
       \/ (exists l v, h = ev_is i' (ac_rw l v)).
Proof.
intros H h i i' Htrco Hrwp.
destruct Hrwp as (l & Hpo & Hread & (v & Hwrite)).
#3 so (writes_cons_invert _#5 Hwrite) as [Hwrite' | [Heq | Heq]].
  #3 so (reads_cons_invert _#4 Hread) as [Hread' | [Heq | (v' & Heq)]].
    #2 so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
      left; eexists; repeat2 split; eauto; [].
      eexists; eassumption.
    
      exfalso.
      eapply trco_init_invert_nis; eauto; [].
      invert Hwrite'; eauto; done.

    exfalso.
    subst h.
    eapply extend_is_npo; eauto; done.

    exfalso.
    subst h.
    eapply extend_is_npo; eauto; done.

  right; left; eauto; done.

  right; right; eauto; done.
Qed.


Lemma wwp_extend_cases :
  forall H h i i',
    trco (H; h)
    -> wwp (H; h) i i'
    -> wwp H i i'
       \/ (exists iw ir i'', h = ev_rf iw ir /\ vos H i iw /\ xos H ir i'' /\ rwpq H i'' i')
       \/ (exists ip i'', h = ev_exec ip /\ H {{ ev_is ip ac_push }} /\ vos H i ip /\ ip <> i'' /\ ~ exec H i'' /\ rwpq H i'' i')
       \/ (exists l v, h = ev_is i' (ac_write l v))
       \/ (exists l v, h = ev_is i' (ac_rw l v))
       \/ (exists f, h = ev_fence f i').
Proof.
intros H h i1 i2 Htrco Hwwp.
invert Hwwp; clear Hwwp.

(* vt *)
intros l Hvt (v1 & Hwrite1) (v2 & Hwrite2).
#3 so (writes_cons_invert _#5 Hwrite2) as [Hwrite2' | [Heq | Heq]].
  #3 so (writes_cons_invert _#5 Hwrite1) as [Hwrite1' | [Heq | Heq]].
    so (vt_extend_cases _#4 Htrco Hvt) as HH.
    #5 repeat2 (splithyp HH).
      left; eapply wwp_vt; eauto;
      eexists; eassumption.
    
      destruct HH as (iw & ir & Heq & Hvos & Hxos).
      right; left.
      repeat eexists; repeat2 split; eauto; [].
      left; reflexivity.
    
      destruct HH as (ip & Heq & Hisp & Hvos & Hneq & Hnexec).
      do 2 right; left.
      repeat eexists; repeat2 split; eauto; [].
      left; reflexivity.
    
      destruct HH as (p & ->).
      exfalso.
      edestruct trco_init_invert_nis; eauto; [].
      invert Hwrite2'; eauto; done.
    
      do 5 right; assumption.

    exfalso.
    subst h.
    so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop].
      eapply extend_is_nexec; eauto; done.
  
      eapply (extend_is_npo _ _ _ i2); eauto; [].
      apply xop_po; auto; done.

    exfalso.
    subst h.
    so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop].
      eapply extend_is_nexec; eauto; done.
  
      eapply (extend_is_npo _ _ _ i2); eauto; [].
      apply xop_po; auto; done.

  subst; do 3 right; left; eauto; done.

  subst; do 4 right; left; eauto; done.

(* po *)
intros l Hpo (v1 & Hwrite1) (v2 & Hwrite2).
#3 so (writes_cons_invert _#5 Hwrite2) as [Hwrite2' | [Heq | Heq]].
  #3 so (writes_cons_invert _#5 Hwrite1) as [Hwrite1' | [Heq | Heq]].
    #2 so (po_extend_cases _#4 Htrco Hpo) as [Hpo' | (p & ->)].
      left; eapply wwp_po; eauto; eexists; eassumption.
    
      exfalso.
      edestruct trco_init_invert_nis; eauto; [].
      invert Hwrite2'; eauto; done.

    exfalso.
    subst h.
    eapply (extend_is_npo _ _ _ i2); eauto; done.
    
    exfalso.
    subst h.
    eapply (extend_is_npo _ _ _ i2); eauto; done.
    
  subst; do 3 right; left; eauto; done.

  subst; do 4 right; left; eauto; done.

(* read *)
intros ir Hwrp Hrwp.
#3 so (rwp_extend_cases _#4 Htrco Hrwp) as [Hrwp' | [(l & v & His) | (l & v & His)]].
  clear Hrwp.
  so (wrp_extend_cases _#4 Htrco Hwrp) as HH.
  clear Hwrp.
  #6 repeat2 (splithyp HH).
    left; eapply wwp_read; eauto; done.

    destruct_all HH.
    right; left.
    repeat eexists; repeat2 split; eauto; [].
    right; assumption.

    destruct_all HH.
    right; right; left.
    repeat eexists; repeat2 split; eauto; [].
    right; assumption.

    destruct HH as (l & ->).
    exfalso.
    eapply extend_is_npo; eauto; [].
    apply rwp_po; auto; [].
    eapply rwp_cons_incl; eauto; done.

    destruct HH as (l & v & ->).
    exfalso.
    eapply extend_is_npo; eauto; [].
    apply rwp_po; auto; [].
    eapply rwp_cons_incl; eauto; done.

    destruct HH as (f & ->).
    exfalso.
    eapply extend_fence_npo; eauto; [].
    apply rwp_po; auto; [].
    eapply rwp_cons_incl; eauto; done.

  do 3 right; left; eauto; done.

  do 4 right; left; eauto; done.
Qed.


Lemma rwf_extend_cases :
  forall H h i i',
    trco (H; h)
    -> rwf (H; h) i i'
    -> rwf H i i'
       \/ (h = ev_rf i i' /\ H {{ ev_is i' (ac_rw l v) | l v }}).
Proof.
intros H h i i' Htrco Hrwf.
destruct Hrwf as (Hrf & (l & v & His)).
destruct Hrf as [Heq | Hrf].
  subst h.
  right.
  split; auto; [].
  exists l, v.
  discriminate_in His; auto; done.

  left.
  split; auto; [].
  exists l, v.
  destruct His as [Heq | His].
    exfalso.
    subst.
    eapply extend_is_nexec; eauto; [].
    eapply exec_rf; [].
    right; eassumption.

    assumption.
Qed.


Lemma co_extend_init :
  forall H i p i1 i2,
    trco (H; ev_init i p)
    -> co (H; ev_init i p) i1 i2
    -> co H i1 i2.
Proof.
intros H i p i1 i2 Htrco Hco.
remember (H; ev_init i p) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hco.

(* read *)
intros H0 i1 i2 ir Hwrp Hrf Hneq -> Htrco.
unfold rf in Hrf; discriminate_in Hrf; intro Hrf; [].
so (wrp_extend_cases _#4 Htrco Hwrp) as Hcond.
repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate); [].
eapply co_read; eauto; done.

(* write *)
intros H0 i1 i2 Hwwp -> Htrco.
so (wwp_extend_cases _#4 Htrco Hwwp) as Hcond.
repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate); [].
eapply co_write; eauto; done.

(* rw *)
intros H0 i1 i2 iw Hrwf _ IH Hneq -> Htrco.
so (IH (eq_refl _) Htrco) as Hco'.
  so (rwf_extend_cases _#4 Htrco Hrwf) as [Hrwf' | (Heqh & _)].
    eapply co_rw; eauto; done.

    discriminate Heqh.

(* fiat *)
intros H' i1 i2 Hin -> Htrco.
discriminate_in Hin.
intro Hin.
apply co_fiat; assumption.
Qed.


Lemma co_extend_is :
  forall H i a i1 i2,
    trco (H; ev_is i a)
    -> co (H; ev_is i a) i1 i2
    -> co H i1 i2 \/ i = i2.
Proof.
intros H i a i1 i2 Htrco Hco.
remember (H; ev_is i a) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hco.

(* read *)
intros H0 i1 i2 ir Hwrp Hrf Hneq -> Htrco.
left.
unfold rf in Hrf; discriminate_in Hrf; intro Hrf; [].
assert (~ i = ir) as Hneqi.
  intro; subst i.
  eapply trco_is_invert_nexec; eauto; [].
  eapply exec_rf; eauto; done.
so (wrp_extend_cases _#4 Htrco Hwrp) as Hcond.
#3 repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate).
  eapply co_read; eauto; done.

  exfalso.
  destruct Hcond as (l & Heq).
  injection Heq; intros; subst; destruct Hneqi; auto; done.

  exfalso.
  destruct Hcond as (l & v & Heq).
  injection Heq; intros; subst; destruct Hneqi; auto; done.

(* write *)
intros H0 i1 i2 Hwwp -> Htrco.
so (wwp_extend_cases _#4 Htrco Hwwp) as Hcond.
#3 repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate).
  left; apply co_write; eauto; done.

  right.
  destruct Hcond as (l & v & Heq).
  injection Heq; eauto; done.

  right.
  destruct Hcond as (l & v & Heq).
  injection Heq; eauto; done.

(* rw *)
intros H0 i1 i2 iw Hrwf _ IH Hneq -> Htrco.
so (IH (eq_refl _) Htrco) as [Hco' | Heq].
  so (rwf_extend_cases _#4 Htrco Hrwf) as [Hrwf' | (Heqh & _)].
    left.
    eapply co_rw; eauto; done.

    discriminate Heqh.

  right; assumption.

(* fiat *)
intros H0 i1 i2 Hin -> Htrco.
left.
discriminate_in Hin; intro; apply co_fiat; assumption.
Qed.


Lemma co_extend_fence :
  forall H i f i1 i2,
    trco (H; ev_fence f i)
    -> co (H; ev_fence f i) i1 i2
    -> co H i1 i2 \/ i = i2.
Proof.
intros H i f i1 i2 Htrco Hco.
remember (H; ev_fence f i) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hco.

(* read *)
intros H0 i1 i2 ir Hwrp Hrf Hneq -> Htrco.
left.
unfold rf in Hrf; discriminate_in Hrf; intro Hrf; [].
assert (~ i = ir) as Hneqi.
  intro; subst i.
  eapply trco_fence_invert_nexec; eauto; [].
  eapply exec_rf; eauto; done.
so (wrp_extend_cases _#4 Htrco Hwrp) as Hcond.
#2 repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate).
  eapply co_read; eauto; done.

  exfalso.
  destruct Hcond as (f' & Heq).
  injection Heq; intros; subst; destruct Hneqi; auto; done.

(* write *)
intros H0 i1 i2 Hwwp -> Htrco.
so (wwp_extend_cases _#4 Htrco Hwwp) as Hcond.
#2 repeat2 (splithyp Hcond); try (destruct_all Hcond; discriminate).
  left; apply co_write; eauto; done.

  right.
  destruct Hcond as (f' & Heq).
  injection Heq; eauto; done.

(* rw *)
intros H0 i1 i2 iw Hrwf _ IH Hneq -> Htrco.
so (IH (eq_refl _) Htrco) as [Hco' | Heq].
  so (rwf_extend_cases _#4 Htrco Hrwf) as [Hrwf' | (Heqh & _)].
    left.
    eapply co_rw; eauto; done.

    discriminate Heqh.

  right; assumption.

(* fiat *)
intros H0 i1 i2 Hin -> Htrco.
discriminate_in Hin; intro.
left; apply co_fiat; assumption.
Qed.


Lemma cos_extend_init :
  forall H i p i1 i2,
    trco (H; ev_init i p)
    -> cos (H; ev_init i p) i1 i2
    -> cos H i1 i2.
Proof.
intros H i p i1 i2 Htrco Hcos.
remember (H; ev_init i p) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hcos.

(* refl *)
intros; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hco _ IH -> Htrco.
#2 apply (@star_step _ _ _ i2).
  eapply co_extend_init; eauto; done.

  apply IH; auto; done.
Qed.


Lemma cop_extend_init :
  forall H i p i1 i2,
    trco (H; ev_init i p)
    -> cop (H; ev_init i p) i1 i2
    -> cop H i1 i2.
Proof.
intros H i p i1 i3 Htrco Hcop.
destruct Hcop as (i2 & Hco & Hcos).
exists i2.
split.
  eapply co_extend_init; eauto; done.

  eapply cos_extend_init; eauto; done.
Qed.  


Lemma cos_extend_is :
  forall H i a i1 i2,
    trco (H; ev_is i a)
    -> cos (H; ev_is i a) i1 i2
    -> cos H i1 i2 \/ cos H i i2.
Proof.
intros H i a i1 i2 Htrco Hcos.
remember (H; ev_is i a) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hcos.

(* refl *)
intros; left; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hco _ IH -> Htrco.
so (IH (eq_refl _) Htrco) as [Hcos' | Hcos'].
  so (co_extend_is _#5 Htrco Hco) as [Hco' | Heq].
    left; eapply star_step; eauto; done.

    right.
    subst i2; assumption.

  right; assumption.
Qed.


Lemma cop_extend_is :
  forall H i a i1 i2,
    trco (H; ev_is i a)
    -> cop (H; ev_is i a) i1 i2
    -> cop H i1 i2 \/ cos H i i2.
Proof.
intros H i a i1 i3 Htrco Hcop.
destruct Hcop as (i2 & Hco & Hcos).
so (cos_extend_is _#5 Htrco Hcos) as [Hcos' | Hcos'].
  so (co_extend_is _#5 Htrco Hco) as [Hco' | Heq].
    left.
    exists i2; eauto; done.

    right; subst; assumption.

  right; assumption.
Qed.


Lemma cos_extend_fence :
  forall H i f i1 i2,
    trco (H; ev_fence f i)
    -> cos (H; ev_fence f i) i1 i2
    -> cos H i1 i2 \/ cos H i i2.
Proof.
intros H i f i1 i2 Htrco Hcos.
remember (H; ev_fence f i) as H' eqn:HeqH'.
revert HeqH' Htrco.
induct Hcos.

(* refl *)
intros; left; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hco _ IH -> Htrco.
so (IH (eq_refl _) Htrco) as [Hcos' | Hcos'].
  so (co_extend_fence _#5 Htrco Hco) as [Hco' | Heq].
    left; eapply star_step; eauto; done.

    right.
    subst i2; assumption.

  right; assumption.
Qed.


Lemma cop_extend_fence :
  forall H i f i1 i2,
    trco (H; ev_fence f i)
    -> cop (H; ev_fence f i) i1 i2
    -> cop H i1 i2 \/ cos H i i2.
Proof.
intros H i f i1 i3 Htrco Hcop.
destruct Hcop as (i2 & Hco & Hcos).
so (cos_extend_fence _#5 Htrco Hcos) as [Hcos' | Hcos'].
  so (co_extend_fence _#5 Htrco Hco) as [Hco' | Heq].
    left.
    exists i2; eauto; done.

    right; subst; assumption.

  right; assumption.
Qed.


(* Coherence acyclicity *)

Lemma init_acyclic :
  forall H i p,
    trco (H; ev_init i p)
    -> acyclic (co H) 
    -> acyclic (co (H; ev_init i p)).
Proof.
intros H i p Htrco Hcoacy.
intros (i' & Hcop).
destruct Hcoacy.
exists i'.
eapply cop_extend_init; eauto; done.
Qed.


Lemma is_acyclic :
  forall H i a,
    trco (H; ev_is i a)
    -> acyclic (co H) 
    -> acyclic (co (H; ev_is i a)).
Proof.
intros H i a Htrco Hcoacy.
intros (i1 & Hcop11).
so (cop_extend_is _#5 Htrco Hcop11) as [Hcop11' | Hcos1].
  destruct Hcoacy; eauto; done.

  destruct Hcop11 as (i2 & Hco12 & _).
  assert (cop (H; ev_is i a) i i2) as Hcop2.
    apply plusr_plus; [].
    exists i1.
    split; auto; [].
    apply cos_cons_incl; auto; done.
  so (cop_inert _#3 Htrco Hcop2) as [(i3 & Hin) | [Hexec | (i3 & Hpo)]].
    discriminate_in Hin.
    intro Hin.
    edestruct trco_is_invert_nco; eauto; done.

    eapply trco_is_invert_nexec; eauto; [].
    eapply exec_tail; eauto; discform; done.

    eapply trco_is_invert_npo; eauto; done.
Qed.


Lemma edge_acyclic :
  forall H b T T',
    trco (H; ev_edge b T T')
    -> acyclic (co H)
    -> acyclic (co (H; ev_edge b T T')).
Proof.
intros H b T T' Htrco Hcoacy.
intros (i & Hcop).
destruct Hcoacy.
exists i.
eapply cop_tail; eauto; discform; done.
Qed.


Lemma fence_acyclic :
  forall H i f,
    trco (H; ev_fence f i)
    -> acyclic (co H) 
    -> acyclic (co (H; ev_fence f i)).
Proof.
intros H i f Htrco Hcoacy.
intros (i1 & Hcop11).
so (cop_extend_fence _#5 Htrco Hcop11) as [Hcop11' | Hcos1].
  destruct Hcoacy; eauto; done.

  destruct Hcop11 as (i2 & Hco12 & _).
  assert (cop (H; ev_fence f i) i i2) as Hcop2.
    apply plusr_plus; [].
    exists i1.
    split; auto; [].
    apply cos_cons_incl; auto; done.
  so (cop_inert _#3 Htrco Hcop2) as [(i3 & Hin) | [Hexec | (i3 & Hpo)]].
    discriminate_in Hin.
    intro Hin.
    edestruct trco_fence_invert_nco; eauto; done.

    eapply trco_fence_invert_nexec; eauto; [].
    eapply exec_tail; eauto; discform; done.

    eapply trco_fence_invert_npo; eauto; done.
Qed.


(* Induction principles *)

Lemma well_founded_po :
  forall H,
    trco H
    -> well_founded (po H).
Proof.
intros H Htrco.
apply acyclic_well_founded; [|].
  apply contained_downfin; [].
  exists (ids H).
  intros i i' Hpo.
  unfold transpose in Hpo.
  refine (po_ids _#4 andel); eauto; done.

  intros (i & Hpop).
  apply (po_irreflex H i); auto; [].
  apply plus_of_transitive; auto; [].
  unfold transitive.
  intros.
  eapply po_trans; eauto; done.
Qed.


Lemma well_founded_pot :
  forall H,
    trco H
    -> well_founded (transpose (po H)).
Proof.
intros H Htrco.
apply acyclic_well_founded; [|].
  apply contained_downfin; [].
  exists (ids H).
  intros i i' Hpo.
  unfold transpose in Hpo.
  eapply po_ids; eauto; done.

  apply acyclic_transpose; [].
  intros (i & Hpop).
  apply (po_irreflex H i); auto; [].
  apply plus_of_transitive; auto; [].
  unfold transitive.
  intros.
  eapply po_trans; eauto; done.
Qed.


Lemma well_founded_to :
  forall H,
    trco H
    -> well_founded (to H).
Proof.
intros H Htrco.
apply acyclic_well_founded; [|].
  apply contained_downfin; [].
  exists (ids H).
  intros i i' Hto.
  unfold transpose in Hto.
  refine (to_ids _#3 Htrco _); eauto; done.

  intros (i & Hto).
  apply (to_irreflex H i); auto; [].
  apply plus_of_transitive; auto; [].
  unfold transitive.
  intros.
  eapply to_trans; eauto; done.
Qed.


Lemma well_founded_tot :
  forall H,
    trco H
    -> well_founded (transpose (to H)).
Proof.
intros H Htrco.
assert (well_founded (fun i i' => to H i' i /\ inited H i)) as Hwf.
  apply acyclic_well_founded; [|].
    apply contained_downfin; [].
    exists (ids H).
    intros i i' (Hto & Hinited).
    apply init_ids; assumption.
  
    intros (i & Htotr).
    apply (to_irreflex H i); auto; [].
    apply plus_of_transitive; auto.
      unfold transitive.
      intros.
      eapply to_trans; eauto; done.

      apply plus_transpose; [].
      refine (plus_mono _#6 Htotr).
      intros ? ? (? & _).
      assumption.
intros i.
wfinduct i using Hwf.
unfold transpose.
intros i IH.
apply Acc_intro.
intros i' Hto.
so (present_1_decidable _ _ (match_init_i_decidable i') H) as [Hinit | Hninit].
  apply IH; auto; done.

  apply Acc_intro.
  intros i'' Hto'.
  destruct Hninit; [].
  apply exec_inited; auto; [].
  eapply to_execed_fst; eauto; done.
Qed.


Lemma well_founded_co :
  forall H,
    trco H
    -> acyclic (co H)
    -> well_founded (co H).
Proof.
intros H Htrco Hcoacy.
apply acyclic_well_founded; [|].
  intro i'.
  exists (ids H).
  intros i Hcop.
  destruct Hcop as (i'' & Hco & _).
  exact (co_ids _#3 Htrco Hco andel).

  apply Hcoacy; auto; done.
Qed.


Lemma well_founded_cot :
  forall H,
    trco H
    -> acyclic (co H)
    -> well_founded (transpose (co H)).
Proof.
intros H Htrco Hcoacy.
apply acyclic_well_founded.
  toshow (downfin (transpose (co H))).
  intro i'.
  exists (ids H).
  intros i Hcotp.
  destruct Hcotp as (i'' & Hco & _).
  unfold transpose in Hco.
  exact (co_ids _#3 Htrco Hco ander).

  apply acyclic_transpose; [].
  apply Hcoacy; auto; done.
Qed.


Lemma acyclic_vo :
  forall H,
    trco H
    -> acyclic (vo H).
Proof.
intros H Htrco (i & Hvop).
so (vop_execed_fst_or_svop _#3 Htrco Hvop) as [Hexec | Hsvop].
  so (vop_to _#3 Htrco Hvop Hexec) as Hto.
  eapply to_irreflex; eauto; done.

  apply (po_irreflex H i); auto; [].
  eapply sop_po; eauto; done.
Qed.


Lemma well_founded_vo :
  forall H,
    trco H
    -> well_founded (vo H).
Proof.
intros H Htrco.
apply acyclic_well_founded.
  intro.
  exists (ids H).
  intros i Hvotp.
  destruct Hvotp as (? & Hvo & _).
  exact (vo_ids _#3 Htrco Hvo).

  apply acyclic_vo; auto; done.
Qed.


Lemma well_founded_vop :
  forall H,
    trco H
    -> well_founded (vop H).
Proof.
intros H Htrco.
apply acyclic_well_founded.
  intro.
  exists (ids H).
  intros i Hvotp.
  destruct Hvotp as (? & (? & Hvo & _) & _).
  exact (vo_ids _#3 Htrco Hvo).

  apply acyclic_plus.
  apply acyclic_vo; auto; done.
Qed.


Lemma well_founded_xo :
  forall H,
    trco H
    -> well_founded (xo H).
Proof.
intros H Htrco.
apply (wf_incl _ _ (po H)); [|].
  unfold inclusion; auto using xo_po; done.

  apply well_founded_po; auto; done.
Qed.
