
Require Import Tactics.
Require Import Static.
Require Import Dynamic.
Require Import Relation.
Require Import History.
Require Import Truncate.
Require Import Sequence.
Require Import Hwf.
Require Export Itemizable.
Require Export Decidable.
Require Import Match.
Require Import Acyclic.


(* Inited *)

Lemma inited_itemizable :
  forall H, itemizable (inited H).
Proof.
intros M.
induction M.
(* nil *)
apply (iff_itemizable _ (fun _ => False)).
  intro x; split; intro H; repeat (destruct H).

  apply itemizable_empty.

(* cons *)
rename a into h.
destruct h;
  [| destruct IHM as (L & HL);
     exists L;
     intro i';
     split;
       [ intro H; apply HL;
         destruct H as (ac & H);
         simpl in H;
         destruct H; [discriminate |];
         eexists; eauto
       | intro H;
         destruct (HL i' ander H) as (ac & H');
         exists ac;
         right;
         assumption ] .. ].
rename t into p.
destruct IHM as (L & HL).
exists (L; i).
intro i'.
split.
  intro H.
  destruct H as (a' & H).
  invert H.
    intro H'.
    inversion H'.
    left; auto.

    right.
    apply HL.
    exists a'.
    auto.

  intro H.
  destruct H as [H | H].
    subst i'.
    exists p.
    left; auto.

    pose proof (HL i' ander H) as (p' & H').
    exists p'.
    right; auto.
Qed.


(* Is *)

Lemma reads_decidable :
  forall H i l,
    decidable (reads H i l).
Proof.
intros H i l.
#2 so (present_decidable (ev_is i (ac_read l)) H) as [Hread | Hnr].
  left.
  apply reads_read; assumption.

  #2 so (present_1_decidable _ _ (match_is_rw_il_decidable i l) H) as [(v & Hrw) | Hnrw].
    left.
    eapply reads_rw; eassumption.

    right.
    intros Hread.
    #2 invert Hread.
      intro His.
      apply Hnr; eauto; done.

      intros v His.
      apply Hnrw; eauto; done.
Qed.


Lemma reads_decidable_i :
  forall H i,
    decidable (exists l, reads H i l).
Proof.
intros H i.
#2 so (present_1_decidable _ _ (match_is_read_i_decidable i) H) as [(l & Hread) | Hnr].
  left.
  eexists.
  eapply reads_read; eassumption.

  #2 so (present_2_decidable _ _ _ (match_is_rw_i_decidable i) H) as [(l & v & Hrw) | Hnrw].
    left.
    eexists.
    eapply reads_rw; eassumption.

    right.
    intros (l & Hread).
    #2 invert Hread.
      intro His.
      apply Hnr; eauto; done.

      intros v His.
      apply Hnrw; eauto; done.
Qed.


Lemma writesto_decidable :
  forall H i l,
    decidable (writesto H i l).
Proof.
intros H i l.
#2 so (present_1_decidable _ _ (match_is_write_il_decidable i l) H) as [(v & Hwrite) | Hnw].
  left.
  exists v.
  apply writes_write; assumption.

  #2 so (present_1_decidable _ _ (match_is_rw_il_decidable i l) H) as [(v & Hrw) | Hnrw].
    left.
    exists v.
    apply writes_rw; assumption.

    right.
    intros (v & Hwrite).
    #2 invert Hwrite.
      intro His.
      apply Hnw; eauto; done.

      intro His.
      apply Hnrw; eauto; done.
Qed.


Lemma writesto_decidable_i :
  forall H i,
    decidable (exists l, writesto H i l).
Proof.
intros H i.
#2 so (present_2_decidable _ _ _ (match_is_write_i_decidable i) H) as [(l & v & Hwrite) | Hnw].
  left.
  exists l, v.
  apply writes_write; assumption.

  #2 so (present_2_decidable _ _ _ (match_is_rw_i_decidable i) H) as [(l & v & Hrw) | Hnrw].
    left.
    exists l, v.
    apply writes_rw; assumption.

    right.
    intros (l & v & Hwrite).
    #2 invert Hwrite.
      intro His.
      apply Hnw; eauto; done.

      intro His.
      apply Hnrw; eauto; done.
Qed.


Lemma writesto_itemizable :
  forall H l,
    trco H
    -> itemizable (fun i => writesto H i l).
Proof.
intros H l Htrco.
#2 apply (itemizable_from_univ _ _ (ids H)).
  intros; apply writesto_decidable; done.

  intros ? (? & ?).
  apply init_ids; [].
  eapply writes_inited; eauto; done.
Qed.


Lemma locations_itemizable :
  forall H, itemizable (fun l => exists i, writesto H i l).
Proof.
intros H.
apply (iff_itemizable _ (fun l => (exists iv, In (ev_is (car iv) (ac_write l (cdr iv))) H) \/ (exists iv, In (ev_is (car iv) (ac_rw l (cdr iv))) H))).
  intros l.
  split.
    intros [((i, v) & His) | ((i, v) & His)]; exists i, v; auto with dynamic; done.
    
    intros (i & v & Hwrite).
    #2 invert Hwrite.
      intro His; left; exists (i, v); assumption.

      intro His; right; exists (i, v); assumption.

  #2 apply itemizable_or.
    #2 apply (itemizable_in_match_1 _ _ _ (fun l iv => ev_is (car iv) (ac_write l (cdr iv)))).
      intro h.
      destruct h; try (right; intro HH; destruct_all HH; discriminate HH); [].
      destruct t; try (right; intro HH; destruct_all HH; discriminate HH); [].
      left.
      exists l, (i, t).
      reflexivity.

      intros l l' (i, v) (i', v') Heq.
      injection Heq.
      intros; subst; auto; done.

    #2 apply (itemizable_in_match_1 _ _ _ (fun l iv => ev_is (car iv) (ac_rw l (cdr iv)))).
      intro h.
      destruct h; try (right; intro HH; destruct_all HH; discriminate HH); [].
      destruct t; try (right; intro HH; destruct_all HH; discriminate HH); [].
      left.
      exists l, (i, t).
      reflexivity.

      intros l l' (i, v) (i', v') Heq.
      injection Heq.
      intros; subst; auto; done.
Qed.


(* Reads-from *)

Lemma rf_itemizable_left :
  forall H i,
    itemizable (fun i' => rf H i' i).
Proof.
intros H i.
apply (itemizable_in_match _ _ (fun i' => ev_rf i' i)); [|].
  intro h.
  apply match_rf_r_decidable; done.

  intros y y' Heq.
  injection Heq.
  auto; done.
Qed.


Lemma rf_itemizable_right :
  forall H i,
    itemizable (fun i' => rf H i i').
Proof.
intros H i.
apply (itemizable_in_match _ _ (fun i' => ev_rf i i')); [|].
  intro h.
  apply match_rf_w_decidable; done.

  intros y y' Heq.
  injection Heq.
  auto; done.
Qed.


Lemma rwf_itemizable_left :
  forall H i,
    itemizable (fun i' => rwf H i' i).
Proof.
intros H i.
#2 apply itemizable_and.
  apply rf_itemizable_left; auto; done.

  intros _.
  exact (present_2_decidable _#3 (match_is_rw_i_decidable i) H).
Qed.


Lemma rwf_itemizable_right :
  forall H i,
    itemizable (fun i' => rwf H i i').
Proof.
intros H i.
#2 apply itemizable_and.
  apply rf_itemizable_right; auto; done.

  intros i'.
  exact (present_2_decidable _#3 (match_is_rw_i_decidable i') H).
Qed.


(* Program order *)

Lemma Exists_inorder_decidable :
  forall (T : Set) (P : T -> T -> Prop),
    (forall x y, decidable (P x y))
    -> forall l,
         decidable (exists l1 l2 l3 x y,
                      l = l1; x;; l2; y;; l3
                      /\ P x y).
Proof.
intros T P HdecP l.
induction l as [| z].

(* nil *)
right.
intro H.
destruct H as (l1 & l2 & l3 & x & y & Heq & Hxy).
eapply app_cons_not_nil.
eexact Heq.

(* cons *)
destruct IHl as [Hl | Hl].
  (* holds in tail *)
  left.
  destruct Hl as (l1 & l2 & l3 & x & y & Heq & Hxy).
  exists l1, l2, (l3; z), x, y.
  split; [| assumption].
  simpl; f_equal; auto.

  (* does not hold in tail *)
  destruct (Exists_decidable _ _ (fun x => HdecP x z) l) as [Hlx | Hlx].
    (* Exists x in l st. P x z *)
    left.
    thus (exists x, In x l /\ P x z) as Hlx' by Exists_exists.
    destruct Hlx' as (x & Hx & Hxz).
    exploit (in_split x l) as Happ; auto.
    destruct Happ as (l2 & l1 & Heq).
    exists l1, l2, nil, x, z.
    split; [| auto].
    simpl; f_equal; auto.

    (* ~ Exist x in l st P x z *)
    right.
    intro Hfull.  (* assume for contradiction that it does hold somewhere *)
    destruct Hfull as (l1 & l2 & l3 & x & y & Heq & Hxy).
    have (~
          (exists (l1 l2 l3 : list T) (x y : T), l = l1; x;; l2; y;; l3 /\ P x y)) as Hl.
    have (~ Exists (fun x : T => P x z) l) as Hlx.
    have (l; z = l1; x;; l2; y;; l3) as Heq.
    have (P x y) as Hxy.
    toshow (False).
    destruct (app_eq_cons _#5 Heq) as [(l1' & Heql1 & _) | (Heql1 & _)].
      (* contradictory assumption does not use z as y *)
      have (l3 = l1'; z) as Heql1.
      subst l3.
      simpl in Heq.
      injection Heq; intro Heql.
      have (l = l1; x;; l2; y;; l1') as Heql.
      contradict Hl.
      exists l1, l2, l1', x, y.
      auto.

      (* contradictory assumption uses z as y *)
      have (l3 = nil) as Heql1.
      subst l3.
      simpl in Heq.
      injection Heq; intros Heql Heqzy.
      have (l = l1; x;; l2) as Heql.
      have (z = y) as Heqzy.
      subst z.
      contradict Hlx.
      apply Exists_exists.
      exists x.
      split; [| auto].
      subst l.
      apply in_or_app.
      simpl.
      right; left; reflexivity.
Qed.


Lemma po_decidable :
  forall H i i', decidable (po H i i').
Proof.
intros M i i'.
unfold po.
exploit (Exists_inorder_decidable _ 
           (fun h h' =>
              exists p,
                h = ev_init i p /\ h' = ev_init i' p)) as Hdec.

(* show the underlying predicate is decidable *)
intros h h'.
toshow (decidable (exists p : thread, h = ev_init i p /\ h' = ev_init i' p)).
destruct (match_init_i_decidable i h) as [Hi_yes | Hi_no].
  destruct Hi_yes as (p & Hh).
  destruct (eq_event_dec h' (ev_init i' p)) as [Hi'_yes | Hi'_no].
    left; eexists; eauto.

    right.
    intros (p' & Hh'' & Hh').
    subst h h'.
    contradict Hi'_no.
    f_equal.
    injection Hh''; auto.

  right.
  intros (p & Hh & Hh').
  contradict Hi_no.
  subst h h'.
  exists p; auto.

(* finish up *)
simpl in Hdec.
pose proof (Hdec M) as Hdec'.
apply (fun H => iff_decidable _ _ H Hdec').
clear Hdec Hdec'.
split.
  intros (l1 & l2 & l3 & x & y & Heq & p & Hx & Hy).
  subst x y.
  repeat eexists; eauto.

  intros (l1 & l2 & l3 & p & Heq).
  subst M.
  exists l1, l2, l3, (ev_init i p), (ev_init i' p).
  split; auto.
  exists p; auto.
Qed.


Lemma poq_decidable :
  forall H i i', decidable (poq H i i').
Proof.
intros H i i'.
so (eq_dec_decidable _ eq_ident_dec i i') as [Heq | Hneq].
  left; left; assumption.

  so (po_decidable H i i') as [Hpo | Hnpo].
    left; right; assumption.

    right.
    intros [|]; contradiction.
Qed.


Lemma po_itemizable :
  forall H, itemizable (fun p => po H (car p) (cdr p)).
Proof.
intros M.
apply (itemizable_subset _ (fun x => inited M (car x) /\ inited M (cdr x))).
  toshow (forall x : identifier * identifier,
          po M (car x) (cdr x) -> inited M (car x) /\ inited M (cdr x)).
  intros (i, i') Hpo.
  simpl in Hpo.
  pose proof (po_inited _#3 Hpo) as (Hi, Hi').
  split; simpl; auto.

  toshow (forall x : identifier * identifier,
          inited M (car x) /\ inited M (cdr x) ->
          decidable (po M (car x) (cdr x))).
  intros (i, i') _.
  simpl.
  apply po_decidable.

  apply itemizable_pair; apply inited_itemizable.
Qed.


Lemma po_itemizable_left :
  forall H i,
    itemizable (fun i' => po H i' i).
Proof.
intros H i.
apply (itemizable_car_fixed _ _ (fun x => po H (car x) (cdr x))).
  intro i'.
  exact (eq_dec_decidable _ eq_ident_dec _ _).

  apply po_itemizable; done.
Qed.


Lemma po_itemizable_right :
  forall H i,
    itemizable (fun i' => po H i i').
Proof.
intros H i.
apply (itemizable_cdr_fixed _ _ (fun x => po H (car x) (cdr x))).
  intro i'.
  exact (eq_dec_decidable _ eq_ident_dec _ _).

  apply po_itemizable; done.
Qed.


(* Trace order *)

Lemma exec_decidable :
  forall H i,
    decidable (exec H i).
Proof.
intros H i.
so (present_decidable (ev_exec i) H) as [Hexec | Hnexec].
  left; apply exec_exec; auto; done.

  so (present_1_decidable _ _ (match_rf_r_decidable i) H) as [(iw & Hrf) | Hnrf].
    left; eapply exec_rf; eauto; done.

    right.
    intro Hexec.
    invert Hexec; [|].
      exact Hnexec.

      intros iw Hrf.
      destruct Hnrf; eauto; done.
Qed.


Lemma to_decidable :
  forall H i i', trco H -> decidable (to H i i').
Proof.
intros H i1 i2 Htrco.
so (eq_ident_dec i1 i2) as [Heq | Hneq].
  right.
  subst i2.
  apply to_irreflex; auto; done.
so (exec_decidable H i1) as [Hexec1 | Hnexec1].
  so (exec_invert _ _ Hexec1) as (h & Hin & Hexecutes).
  so (in_split _#3 Hin) as (H2 & H1 & ->).
  so (exec_decidable H1 i2) as [Hexec2 | Hnexec2].
    right.
    intro Hto12.
    destruct (to_irreflex (H1; h;; H2) i1); auto; [].
    eapply to_trans; eauto; [].
    eapply to_from_before; eauto; done.

    left.
    exists (H1; h), H2.
    splitall; auto; [|].
      eapply executes_exec; eauto; [].
      left; reflexivity.

      contradict Hnexec2.
      so (exec_cons_invert _#3 Hnexec2) as [Hexecutes' | ?]; auto; [].
      so (executes_fun_1 _#3 Hexecutes Hexecutes') as Heq.
      contradiction.

  right.
  contradict Hnexec1.
  eapply to_execed_fst; eauto; done.
Qed.


(* Specified order *)

Lemma fence_itemizable :
  forall H i,
    itemizable (fun f => H {{ ev_fence f i }}).
Proof.
intros H i.
apply itemizable_in_match.
  apply match_fence_i_decidable; done.

  intros ? ? Heq; injection Heq; auto; done.
Qed.


Lemma fence_tag_itemizable :
  forall H i,
    itemizable (fun T => H {{ ev_fence (fc_tag (tg_lit T)) i }}).
Proof.
intros H i.
apply (itemizable_preimage _#3 (fun f => H {{ ev_fence f i }})).
  intros f Hfence.
  apply match_fc_tag_lit_decidable; done.

  intros T T' Heq.
  injection Heq; auto; done.
  
  apply fence_itemizable; done.
Qed.


Lemma so_decidable :
  forall b H i i',
    decidable (so b H i i').
Proof.
intros b H i i'.
so (po_decidable H i i') as [Hpo | Hnpo].
  have (po H i i') as Hpo.
  so (present_decidable (ev_fence (fc_post b) i) H) as [Hfence1 | Hnfence1].
    have (H {{ ev_fence (fc_post b) i }}) as Hfence1.
    left.
    apply so_post; auto; done.

    have (~ H {{ ev_fence (fc_post b) i }}) as Hnfence1.
    so (present_decidable (ev_fence (fc_pre b) i') H) as [Hfence2 | Hnfence2].
      have (H {{ ev_fence (fc_pre b) i' }}) as Hfence2.
      left.
      apply so_pre; auto; done.

      have (~ H {{ ev_fence (fc_pre b) i' }}) as Hnfence2.
      apply (iff_decidable (exists (T2 : tag * tag),
                              (H {{ ev_fence (fc_tag (tg_lit (car T2))) i }}
                               /\ H {{ ev_fence (fc_tag (tg_lit (cdr T2))) i' }})
                              /\ H {{ ev_edge b (car T2) (cdr T2) }})).
        split.
          toshow ((exists T2 : tag * tag,
                     (H {{ ev_fence (fc_tag (tg_lit (car T2))) i }} /\
                      H {{ ev_fence (fc_tag (tg_lit (cdr T2))) i' }}) /\
                     H {{ ev_edge b (car T2) (cdr T2) }}) -> so b H i i').
          intros ((T, T') & (Hfence & Hfence') & Hedge).
          simpl in Hfence, Hfence', Hedge.
          eapply so_tag; eauto; done.

          toshow (so b H i i' ->
                  exists T2 : tag * tag,
                    (H {{ ev_fence (fc_tag (tg_lit (car T2))) i }} /\
                     H {{ ev_fence (fc_tag (tg_lit (cdr T2))) i' }}) /\
                    H {{ ev_edge b (car T2) (cdr T2) }}).
          intro Hso.
          invert Hso; try (intros; findcontra); [].
          intros T T' _ Hfence Hfence' Hedge.
          exists (T, T').
          simpl.
          split; [split |]; auto; done.

        apply itemizable_decidable_exists; [].
        apply itemizable_and; [|].
          toshow (itemizable
                    (fun x : tag * tag =>
                     H {{ ev_fence (fc_tag (tg_lit (car x))) i }} /\
                     H {{ ev_fence (fc_tag (tg_lit (cdr x))) i' }})).
          apply (itemizable_pair _ _
                   (fun T => H {{ ev_fence (fc_tag (tg_lit T)) i }})
                   (fun T => H {{ ev_fence (fc_tag (tg_lit T)) i' }})); [|].
            apply fence_tag_itemizable; done.

            apply fence_tag_itemizable; done.

          toshow (forall x : tag * tag, decidable (H {{ ev_edge b (car x) (cdr x) }})).
          intros (T & T'); simpl.
          apply present_decidable; done.

  have (~ po H i i') as Hnpo.
  right.
  intro Hso.
  destruct Hnpo.
  invert Hso; intros; assumption.
Qed.


Lemma bso_decidable :
  forall H i i',
    decidable (exists b, so b H i i').
Proof.
intros H i i'.
so (so_decidable at_v H i i') as [Hsvo | Hnsvo].
  left; eexists; eassumption.
so (so_decidable at_x H i i') as [Hsxo | Hnsxo].
  left; eexists; eassumption.
right.
intros (b & Hso).
destruct b; auto; done.
Qed.


(* Visibility order *)

Lemma pio_decidable :
  forall H i i',
    trco H
    -> decidable (pio H i i').
Proof.
intros H i i' Htrco.
so (present_decidable (ev_is i ac_push) H) as [His | Hnis].
  so (to_decidable H i i' Htrco) as [Hto | Hnto].
    left; split; auto; done.

    right; intros (? & ?).
    findcontra.

  right; intros (? & ?).
  findcontra.
Qed.


Lemma vo_decidable :
  forall H i i',
    trco H
    -> decidable (vo H i i').
Proof.
intros H i i' Htrco.
so (so_decidable at_v H i i') as [Hsvo | Hnsvo].
  left; left; assumption.

  so (pio_decidable H i i' Htrco) as [Hpio | Hnpio].
    left; right; right; assumption.
    
    so (present_decidable (ev_rf i i') H) as [Hrf | Hnrf].
      left; right; left; assumption.

      right.
      intro Hvo.
      destruct Hvo as [|[|]]; findcontra.
Qed.


Lemma vo_itemizable :
  forall H U, trco H -> itemizable (fun x => In (cdr x) U /\ vo H (car x) (cdr x)).
Proof.
intros H U Htrco.
set (U' := list_product (ids H) U).
apply (itemizable_from_univ _ _ U').
  intros (i, i') _.
  so (In_decidable _ i' U eq_ident_dec) as [Hin | Hnin].
    so (vo_decidable _ i i' Htrco) as [Hvo | Hnvo].
      left; auto; done.

      right.
      intros (_ & Hvo); contradiction.

    right.
    intros (Hin & _); contradiction.

  intros (i, i') (Hin & Hvo).
  so (vo_ids _#3 Htrco Hvo) as Hin'.
  apply In_product; auto; done.
Qed.


Lemma vo_left_itemizable :
  forall H, trco H -> forall i, itemizable (fun i' => vo H i' i).
Proof.
intros H Htrco.
intro i2.
apply (iff_itemizable _ (fun i1 => In i2 (i2 :: nil) /\ vo H i1 i2)); [|].
  intro i.
  split.
    intros (_ & ?); assumption.

    intro Hvo; split; auto; [].
    left; reflexivity.

  generalize i2 at 1 3.
  apply itemizable_forall; [|].
    apply (itemizable_swap _ _ (fun i2' i1 => In i2' (i2 :: nil) /\ vo H i1 i2')); [].
    eapply vo_itemizable; eauto; done.
  
    apply (eq_dec_decidable _ eq_ident_dec); done.
Qed.


Lemma vo_right_itemizable :
  forall H U, trco H -> forall i, itemizable (fun i' => In i' U /\ vo H i i').
Proof.
intros H U Htrco.
apply itemizable_forall.
  apply vo_itemizable; auto; done.

  apply (eq_dec_decidable _ eq_ident_dec); done.
Qed.


Lemma vop_left_itemizable :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => vop H i' i).
Proof.
intros H Htrco i.
apply itemizable_plus_left; [|].
  apply well_founded_vo; auto; done.
  
  apply vo_left_itemizable; auto; done.
Qed.


Lemma vos_left_itemizable :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => vos H i' i).
Proof.
intros H Htrco i.
apply itemizable_star_left; [|].
  apply well_founded_vo; auto; done.
  
  apply vo_left_itemizable; auto; done.
Qed.


Lemma vos_decidable :
  forall H i i',
    trco H
    -> decidable (vos H i i').
Proof.
intros H i i' Htrco.
revert i.
apply itemizable_decidable; [|].
  apply vos_left_itemizable; auto; done.

  exact (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma vop_decidable :
  forall H i i',
    trco H
    -> decidable (vop H i i').
Proof.
intros H i i' Htrco.
revert i.
apply itemizable_decidable; [|].
  apply vop_left_itemizable; auto; done.

  exact (eq_dec_decidable _ eq_ident_dec).
Qed.


(* Execution order *)

Lemma xo_decidable :
  forall H i i',
    trco H
    -> decidable (xo H i i').
Proof.
intros H i i' Htrco.
so (so_decidable at_v H i i') as [Hsov | Hnsov].
  left; eapply xo_specified; eauto; done.
so (so_decidable at_x H i i') as [Hsox | Hnsox].
  left; eapply xo_specified; eauto; done.
right.
intro Hxo.
invert Hxo.
intros b Hso.
destruct b; contradiction.
Qed.


Lemma xo_itemizable_left :
  forall H i,
    trco H
    -> itemizable (fun i' => xo H i' i).
Proof.
intros H i Htrco.
apply (itemizable_from_univ _ _ (ids H)); [|].
  intros i' _.
  apply xo_decidable; auto; done.

  intros i' Hxo.
  so (xo_ids _#3 Htrco Hxo) as (? & ?); auto; done.
Qed.


Lemma xos_itemizable_left :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => xos H i' i).
Proof.
intros H Htrco i.
apply itemizable_star_left; [|].
  apply well_founded_xo; auto; done.
  
  intro i'.
  apply xo_itemizable_left; auto; done.
Qed.


Lemma xo_decidable_left :
  forall H i,
    trco H
    -> decidable (exists i', xo H i' i).
Proof.
intros.
apply itemizable_decidable_exists.
apply xo_itemizable_left; auto.
Qed.


Lemma xop_itemizable_left :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => xop H i' i).
Proof.
intros H Htrco i.
apply itemizable_plus_left; [|].
  apply well_founded_xo; auto; done.
  
  intro i'.
  apply xo_itemizable_left; auto; done.
Qed.


(* Visible-to *)

Lemma vt_decidable :
  forall H i i',
    trco H
    -> decidable (vt H i i').
Proof.
intros H i1 i3 Htrco.
apply itemizable_decidable_exists; [].
apply itemizable_and_right; [|].
  intro.
  apply vop_decidable; auto; done.

  apply xos_itemizable_left; auto; done.
Qed.



(* Coherence order *)

Lemma wrp_decidable :
  forall H i i', trco H -> decidable (wrp H i i').
Proof.
intros H i1 i2 Htrco.
#2 so (writesto_decidable_i H i1) as [(l & Hwrite1) | Hnisw1].
  #2 so (reads_decidable H i2 l) as [Hread | Hnread].
    #2 so (vt_decidable _ i1 i2 Htrco) as [Hvt | Hnvt].
      left.
      eapply wrp_vt; eauto; done.

      so (po_decidable H i1 i2) as [Hpo | Hnpo].
        left.
        eapply wrp_po; eauto; done.

        right.
        intro Hwrp.
        invert Hwrp; intros; contradiction.

  right.
  intro Hwrp.
  destruct Hnread.
  eapply wrp_loc_1; eauto; done.

right.
intro Hwrp.
so (wrp_wf_prop _#3 Htrco Hwrp) as (l & Hwrite1 & _).
destruct Hnisw1; eauto; done.
Qed.


Lemma wrp_itemizable_left :
  forall H i, trco H -> itemizable (fun i' => wrp H i' i).
Proof.
intros H i Htrco.
apply (itemizable_from_univ _ _ (ids H)); [|].

  intros; apply wrp_decidable; auto; done.

  intros i' Hwrp.
  so (wrp_wf_prop _#3 Htrco Hwrp) as (l & (v & His) & _).
  eapply writes_ids; eauto; done.
Qed.


Lemma rwp_itemizable_left :
  forall H i, trco H -> itemizable (fun i' => rwp H i' i).
Proof.
intros H iw Htrco.
#2 so (writesto_decidable_i H iw) as [(l & Hwrite1) | Hnisw1].
  apply (iff_itemizable _ (fun ir => po H ir iw /\ reads H ir l)); [|].
    intro ir.
    split.
      intros (Hpo & Hisr).
      eexists; eauto; done.

      intro Hrwp.
      split.
        unfold rwp in Hrwp; destruct_all Hrwp; assumption.

        eapply rwp_loc_2; eauto; done.

    apply itemizable_and; [|].
      apply po_itemizable_left; auto; done.

      intro i.
      apply reads_decidable; done.

  apply itemizable_none; [].
  intros (ir & Hrwp).
  destruct Hrwp as (l & _ & _ & His).
  destruct Hnisw1.
  eexists; eassumption.
Qed.


Lemma rwp_decidable :
  forall H i i', trco H -> decidable (rwp H i i').
Proof.
intros H i i' Htrco.
apply (itemizable_decidable _ (fun i => rwp H i i')); [|].
  apply rwp_itemizable_left; auto; done.

  exact (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma rwpq_decidable :
  forall H i i', trco H -> decidable (rwpq H i i').
Proof.
intros H i i' Htrco.
so (eq_ident_dec i i') as [Heq | Hneq].
  subst i'.
  left; left; reflexivity.

  so (rwp_decidable _ i i' Htrco) as [Hrwp | Hnrwp].
    left; right; assumption.

    right.
    intro Hrwpq.
    destruct Hrwpq as [Heq | Hrwp]; contradiction.
Qed.


Lemma wwp_decidable :
  forall H i i', trco H -> decidable (wwp H i i').
Proof.
intros H i1 i2 Htrco.
#2 so (writesto_decidable_i H i1) as [(l & Hwrite1) | Hnisw1].
  #2 so (writesto_decidable H i2 l) as [Hwrite2 | Hnwrite2].
    so (vt_decidable _ i1 i2 Htrco) as [Hvt | Hnvt].
      left.
      eapply wwp_vt; eauto; done.

      so (po_decidable H i1 i2) as [Hpo | Hnpo].
        left.
        eapply wwp_po; eauto; done.

        assert (decidable (exists ir, wrp H i1 ir /\ rwp H ir i2)) as [(ir & Hwrp & Hrwp) | Hnprop].
          apply itemizable_decidable_exists; [].
          apply itemizable_and_right; [|].
            intro; apply wrp_decidable; auto; done.
            
            apply rwp_itemizable_left; auto; done.

          left.
          eapply wwp_read; eauto; done.

          right.
          intro Hwwp.
          invert Hwwp; try (intros; contradiction); [].
          intros; destruct Hnprop; eauto; done.

  right.
  intro Hwwp.
  destruct Hnwrite2.
  eapply wwp_loc_1; eauto; done.

right.
intro Hwwp.
so (wwp_wf_prop _#3 Htrco Hwwp) as (l & Hwrite1 & _).
destruct Hnisw1; eauto; done.
Qed.


Lemma co_decidable :
  forall H i i', trco H -> decidable (co H i i').
Proof.
intros H i1 i2 Htrco.
revert i2.
wfinduct i1 using (well_founded_to _ Htrco).
intros i1 IH i2.
so (present_decidable (ev_co i1 i2) H) as [Hin | Hnin].
  left.
  apply co_fiat; auto; done.

  so (wwp_decidable H i1 i2 Htrco) as [Hwwp | Hnwwp].
    left.
    apply co_write; auto; done.
  
    so (eq_ident_dec i1 i2) as [Heq | Hneq].
      right.
      intro Hco.
      subst i2.
      invert Hco; auto; done.
  
      assert (decidable (exists ir, wrp H i1 ir /\ rf H i2 ir)) as [(ir & Hwrp & Hrf) | Hnprop].
        apply itemizable_decidable_exists; [].
        apply itemizable_and_right; [|].
          intro ir.
          apply wrp_decidable; auto; done.
  
          apply rf_itemizable_right; auto; done.
  
        left.
        eapply co_read; eauto; done.
  
        assert (decidable (exists iw, rwf H iw i1)) as [(iw & Hrwf) | Hnrwf].
          apply itemizable_decidable_exists; [].
          apply rwf_itemizable_left; auto; done.
  
          assert (decidable (co H iw i2)) as [Hco | Hnco].
            apply IH; [].
            apply rf_to; auto; [].
            apply rwf_rf; auto; done.
  
            left.
            eapply co_rw; eauto; done.
  
            right.
            intro Hco.
            #4 invert Hco.
              intros ir Hwrp Hrf _.
              destruct Hnprop; eauto; done.
      
              intro; contradiction Hnwwp.
  
              intros iw' Hrwf' Hco' _.
              replace iw' with iw in Hco'.
                contradiction.
              eapply rf_fun; eauto using rwf_rf; done.

              intro; contradiction.
  
          right.
          intro Hco.
          #4 invert Hco.
            intros ir Hwrp Hrf _.
            destruct Hnprop; eauto; done.
  
            intro; contradiction Hnwwp.
  
            intros iw Hrwf _ _.
            destruct Hnrwf; eauto; done.

            intro; contradiction.
Qed.


Lemma co_itemizable :
  forall H, trco H -> itemizable (fun x => co H (car x) (cdr x)).
Proof.
intros H Htrco.
set (U := list_product (ids H) (ids H)).
apply (itemizable_from_univ _ _ U); [|].
  intros (i, i') _.
  apply co_decidable; auto; done.

  intros (i, i') Hco.
  simpl in Hco.
  pose proof (co_ids _#3 Htrco Hco) as (H1, H2).
  apply In_product; auto.
Qed.


Lemma co_left_itemizable :
  forall H, trco H -> forall i, itemizable (fun i' => co H i' i).
Proof.
intros H Htrco.
apply itemizable_forall; [|].
  apply (itemizable_swap _ _ (fun x y => co H y x)); [].
  apply co_itemizable; auto; done.

apply (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma co_right_itemizable :
  forall H, trco H -> forall i, itemizable (co H i).
Proof.
intros H Htrco.
apply itemizable_forall.
  apply co_itemizable; auto.

apply (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma cos_left_itemizable :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => cos H i' i).
Proof.
intros H Htrco i.
#3 apply (itemizable_star_left_finite _ _ (ids H)).
  exact (eq_dec_decidable _ eq_ident_dec).

  intros i1 i2 Hco.
  eapply co_ids; eauto; done.
  
  apply co_left_itemizable; auto; done.
Qed.


Lemma cos_right_itemizable :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => cos H i i').
Proof.
intros H Htrco i.
#3 apply (itemizable_star_right_finite _ _ (ids H)).
  exact (eq_dec_decidable _ eq_ident_dec).

  intros i1 i2 Hco.
  refine (co_ids _#5 andel); eauto; done.

  apply co_right_itemizable; auto; done.
Qed.


Lemma cop_right_itemizable :
  forall H, 
    trco H
    -> forall i,
         itemizable (fun i' => cop H i i').
Proof.
intros H Htrco i.
#3 apply (itemizable_plus_right_finite _ _ (ids H)).
  exact (eq_dec_decidable _ eq_ident_dec).

  intros i1 i2 Hco.
  refine (co_ids _#5 andel); eauto; done.

  apply co_right_itemizable; auto; done.
Qed.


Lemma cos_decidable :
  forall H i i',
    trco H
    -> decidable (cos H i i').
Proof.
intros H i i' Htrco.
apply itemizable_decidable.
  apply cos_right_itemizable; auto; done.

  exact (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma cop_decidable :
  forall H i i',
    trco H
    -> decidable (cop H i i').
Proof.
intros H i i' Htrco.
apply itemizable_decidable.
  apply cop_right_itemizable; auto; done.

  exact (eq_dec_decidable _ eq_ident_dec).
Qed.


Lemma acyclic_co_decidable :
  forall H,
    trco H
    -> decidable (acyclic (co H)).
Proof.
intros H Htrco.
apply dec_not; [].
apply itemizable_decidable_exists; [].
#2 apply (itemizable_from_univ _ _ (ids H)).
  intros i _.
  apply cop_decidable; auto; done.

  intros i Hcop.
  destruct Hcop as (i' & Hco & _).
  refine (co_ids _#5 andel); eauto; done.
Qed.


Lemma init_decidable :
  forall e, decidable (init e).
Proof.
intro e.
#2 induct e;
try (intros; right; intro Hinit; invert Hinit; done);
try (intros; left; auto with static; done).

(* bind *)
intros t _ e1 IH1 e2 IH2.
destruct IH1 as [Hinit1 | Hninit1].
  destruct IH2 as [Hinit2 | Hninit2].
    left; auto with static; done.

    right.
    contradict Hninit2.
    invert Hninit2; auto; done.

  right.
  contradict Hninit1.
  invert Hninit1; auto; done.

(* spec *)
intros v _ v' _ e IH.
destruct IH as [Hinit | Hninit].
  left; auto with static; done.

  right.
  contradict Hninit.
  invert Hninit; auto; done.
Qed.


