
Require Import Tactics.
Require Import Relation.
Require Import Subst.
Export Syntax.
Require Import Sequence.
Require Import Static.


Inductive tstep : context -> (* tm *) term -> (* tm *) term -> Prop :=
| tstep_prim1 {G m1 m2 m1'}
    : tstep G m1 m1'
      -> tstep G (tm_prim m1 m2) (tm_prim m1' m2)

| tstep_prim2 {G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> tstep G (tm_prim v1 m2) (tm_prim v1 m2')

| tstep_prim3 {G c1 c2}
    : tstep G (tm_prim (tm_lit c1) (tm_lit c2)) (tm_lit (prim c1 c2))

| tstep_if1 {G m m1 m2 m'}
    : tstep G m m'
      -> tstep G (tm_if m m1 m2) (tm_if m' m1 m2)

| tstep_if2 {G m1 m2}
    : tstep G (tm_if (tm_lit base0) m1 m2) m1

| tstep_if3 {G c m1 m2}
    : c <> base0
      -> tstep G (tm_if (tm_lit c) m1 m2) m2

| tstep_app1 {G m1 m2 m1'}
    : tstep G m1 m1'
      -> tstep G (tm_app m1 m2) (tm_app m1' m2)

| tstep_app2 {G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> tstep G (tm_app v1 m2) (tm_app v1 m2')

| tstep_app3 {G t m1 v2}
    : value G v2
      -> tstep G (tm_app (tm_lam t m1) v2) (subst1 v2 m1)

| tstep_app3' {G t1 t2 m1 v2}
    : value G v2
      -> tstep G (tm_app (tm_fun t1 t2 m1) v2) (subst (dot v2 (dot (tm_fun t1 t2 m1) id)) m1)
.


Definition addfence (f : (* fc *) term) (d : (* tr *) term) : (* tr *) term :=
  (match d with
   | tr_init fl i a => tr_init (fl_cons f fl) i a
   | _ => d
   end).


Inductive xstep : tgtp -> lctp -> context -> (* xp *) term -> (* xp *) term -> (* tr *) term -> Prop :=
| xstep_ret {U P G m m'}
    : tstep G m m'
      -> xstep U P G (xp_ret m) (xp_ret m') tr_nothing

| xstep_force1 {U P G m m'}
    : tstep G m m'
      -> xstep U P G (xp_force m) (xp_force m') tr_nothing

| xstep_force2 {U P G e}
    : xstep U P G (xp_force (tm_susp e)) e tr_nothing

| xstep_read1 {U P G m m'}
    : tstep G m m'
      -> xstep U P G (xp_read m) (xp_read m') tr_nothing

| xstep_read2 {U P G l i}
    : xstep U P G (xp_read (tm_loc l)) (xp_action i) (tr_init fl_nil i (ac_read l))

| xstep_write1 {U P G m1 m2 m1'}
    : tstep G m1 m1'
      -> xstep U P G (xp_write m1 m2) (xp_write m1' m2) tr_nothing

| xstep_write2 {U P G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> xstep U P G (xp_write v1 m2) (xp_write v1 m2') tr_nothing

| xstep_write3 {U P G l v i}
    : value G v
      -> xstep U P G (xp_write (tm_loc l) v) (xp_action i) (tr_init fl_nil i (ac_write l v))

| xstep_rw1 {U P G m1 m2 m1'}
    : tstep G m1 m1'
      -> xstep U P G (xp_rw m1 m2) (xp_rw m1' m2) tr_nothing

| xstep_rw2 {U P G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> xstep U P G (xp_rw v1 m2) (xp_rw v1 m2') tr_nothing

| xstep_rw3 {U P G l v i}
    : value G v
      -> xstep U P G (xp_rw (tm_loc l) v) (xp_action i) (tr_init fl_nil i (ac_rw l v))

| xstep_rmw1 {U P G t m1 m2 m1'}
    : tstep G m1 m1'
      -> xstep U P G (xp_rmw t m1 m2) (xp_rmw t m1' m2) tr_nothing

| xstep_rmw2 {U P G t v1 m2 v}
    : value G v1
      -> tmof U P G v t
      -> value G v
      -> xstep U P G (xp_rmw t v1 m2) (xp_bind t (xp_rw v1 (subst1 v m2)) (xp_spec (var 0) (shift1 v) (xp_ret (shift1 v)))) tr_nothing

| xstep_push {U P G i}
    : xstep U P G xp_push (xp_action i) (tr_init fl_nil i ac_push)

| xstep_nop {U P G i}
    : xstep U P G xp_nop (xp_action i) (tr_init fl_nil i ac_nop)

| xstep_exec {U P G i v}
    : xstep U P G (xp_action i) (xp_ret v) (tr_exec i v)

| xstep_bind1 {U P G t e1 e1' e2 d}
    : xstep U P G e1 e1' d
      -> xstep U P G (xp_bind t e1 e2) (xp_bind t e1' e2) d

| xstep_bind2 {U P G t e1 e2 e2' d}
    : init e1
      -> xstep U P (G; cl_vl t) e2 e2' (shift1 d)
      -> xstep U P G (xp_bind t e1 e2) (xp_bind t e1 e2') d

| xstep_bind3 {U P G t v e}
    : value G v
      -> xstep U P G (xp_bind t (xp_ret v) e) (subst1 v e) tr_nothing

| xstep_spec1 {U P G v1 v2 t e}
    : tmof U P G v1 t
      -> tmof U P G v2 t
      -> value G v1
      -> value G v2
      -> xstep U P G (subst1 v1 e) (xp_spec v1 v2 (subst1 v2 e)) tr_nothing

| xstep_spec2 {U P G e e' d v1 v2}
    : xstep U P G e e' d
      -> xstep U P G (xp_spec v1 v2 e) (xp_spec v1 v2 e') d

| xstep_spec3 {U P G v e}
    : xstep U P G (xp_spec v v e) e tr_nothing

| xstep_new {U P G e b T1 T2}
    : xstep U P G (xp_new b e) (subst (dot (tg_lit T2) (dot (tg_lit T1) id)) e) (tr_new b T1 T2)

| xstep_fence1 {U P G f e e' d}
    : xstep U P G e e' d
      -> xstep U P G (xp_fence f e) (xp_fence f e') (addfence f d)

| xstep_fence2 {U P G f e}
    : init e
      -> xstep U P G (xp_fence f e) e tr_nothing
.


Inductive exstep : tgtp -> lctp -> xstate -> xstate -> (* tr *) term -> thread -> Prop :=
| exstep_hit {U P ex p e e' d}
    : xstep U P nil e e' d
      -> exstep U P (ex; (p, e)) (ex; (p, e')) d p

| exstep_miss {U P ex ex' p e d p'}
    : exstep U P ex ex' d p'
      -> exstep U P (ex; (p, e)) (ex'; (p, e)) d p'
.


Hint Constructors tstep init xstep exstep evolve : dynamic.


(* Reads and write *)
Inductive reads : history -> identifier -> location -> Prop :=
| reads_read {H i l}
    : H {{ ev_is i (ac_read l) }}
      -> reads H i l

| reads_rw {H i l v}
    : H {{ ev_is i (ac_rw l v) }}
      -> reads H i l
.


Inductive writes : history -> identifier -> location -> (* tm *) term -> Prop :=
| writes_write {H i l v}
    : H {{ ev_is i (ac_write l v) }}
      -> writes H i l v

| writes_rw {H i l v}
    : H {{ ev_is i (ac_rw l v) }}
      -> writes H i l v
.


Definition writesto H i l := exists v, writes H i l v.


Hint Constructors reads writes : dynamic.


(* program order *)
(* Should we define this as for trace order? *)
Definition po (H : history) (i i' : identifier) : Prop :=
  exists H1 H2 H3 p,
    H = H1; ev_init i p;; H2; ev_init i' p;; H3.

Definition poq H i i' := i = i' \/ po H i i'.


(* exec *)
Inductive exec : history -> identifier -> Prop :=
| exec_exec {H i}
    : H {{ ev_exec i }}
      -> exec H i

| exec_rf {H i' i}
    : H {{ ev_rf i' i }}
      -> exec H i.


(* trace order *)
Definition to (H : history) (i i' : identifier) : Prop :=
  exists H1 H2,
    H = H1;; H2
    /\ exec H1 i
    /\ ~ exec H1 i'.
    
Definition toq H i i' := i = i' \/ to H i i'.


(* specified order *)
Inductive so : attr -> history -> identifier -> identifier -> Prop :=
| so_tag {b H i i' T T'}
    : po H i i'
      -> H {{ ev_fence (fc_tag (tg_lit T)) i }}
      -> H {{ ev_fence (fc_tag (tg_lit T')) i' }}
      -> H {{ ev_edge b T T' }}
      -> so b H i i'

| so_post {b H i i'}
    : po H i i'
      -> H {{ ev_fence (fc_post b) i }}
      -> so b H i i'

| so_pre {b H i i'}
    : po H i i'
      -> H {{ ev_fence (fc_pre b) i' }}
      -> so b H i i'
.

Definition sop b H := plus (so b H).


(* reads-from *)
Definition rf H i i' := H {{ ev_rf i i' }}.
Definition rwf H i i' := H {{ ev_rf i i' }} /\ H {{ ev_is i' (ac_rw le ve) | le ve }}.

Definition rwfs H := star (rwf H).
Definition rwfp H := plus (rwf H).


(* push order *)
Definition pio H i i' :=
  H {{ ev_is i ac_push }} /\ to H i i'.


(* visibility order *)
Definition vo H i i' :=
  so at_v H i i' \/ rf H i i' \/ pio H i i'.

Definition vos H := star (vo H).
Definition vop H := plus (vo H).


(* execution order *)
Inductive xo : history -> identifier -> identifier -> Prop :=
| xo_specified {b H i i'}
    : so b H i i'
      -> xo H i i'
.

Definition xos H := star (xo H).
Definition xop H := plus (xo H).


Hint Constructors xo so : xo.


(* visible-to *)
(* Should we revise this to include xop?  Then scrap wrp_po? *)
Definition vt H i i' :=
  exists i'',
    vop H i i''
    /\ xos H i'' i'.


(* coherence order *)
Inductive wrp : history -> identifier -> identifier -> Prop :=
| wrp_vt {H i i' l}
    : vt H i i'
      -> writesto H i l
      -> reads H i' l
      -> wrp H i i'

| wrp_po {H i i' l}
    : po H i i'
      -> writesto H i l
      -> reads H i' l
      -> wrp H i i'
.

Definition rwp H i i' :=
  exists l,
    po H i i'
    /\ reads H i l
    /\ writesto H i' l.

(* Don't need a visible-to case in read-write priority, as
   it would not add any coherence edges (Lemma wrp_vt_co).
*)

Definition rwpq H i i' := i = i' \/ rwp H i i'.

Inductive wwp : history -> identifier -> identifier -> Prop :=
| wwp_vt {H i i' l}
    : vt H i i'
      -> writesto H i l
      -> writesto H i' l
      -> wwp H i i'

| wwp_po {H i i' l}
    : po H i i'
      -> writesto H i l
      -> writesto H i' l
      -> wwp H i i'

| wwp_read {H i i' ir}
    : wrp H i ir
      -> rwp H ir i'
      -> wwp H i i'
.

Inductive co : history -> identifier -> identifier -> Prop :=
| co_read {H i i' ir}
    : wrp H i ir
      -> rf H i' ir
      -> i <> i'
      -> co H i i'

| co_write {H i i'}
    : wwp H i i'
      -> co H i i'

| co_rw {H i i' iw}
    : rwf H iw i
      -> co H iw i'
      -> i <> i'
      -> co H i i'

| co_fiat {H i i'}
    : H {{ ev_co i i' }}
      -> co H i i'
.


Definition cos H := star (co H).
Definition cop H := plus (co H).


Definition executable H i :=
  ~ exec H i
  /\ forall i', xo H i' i -> exec H i'.


Fixpoint fc_list (fl : (* fl *) term) {struct fl} : list term :=
  (match fl with
   | fl_cons f t => f :: fc_list t
   | _ => nil
   end).


Inductive tag_declared : history -> tag -> Prop :=
| tag_declared_1 {H b T T'}
    : H {{ ev_edge b T T' }}
      -> tag_declared H T

| tag_declared_2 {H b T T'}
    : H {{ ev_edge b T T' }}
      -> tag_declared H T'
.


Inductive fence_declared : history -> (* fc *) term -> Prop :=
| fence_declared_tag {H T}
    : tag_declared H T
      -> fence_declared H (fc_tag (tg_lit T))

| fence_declared_pre {H b}
    : fence_declared H (fc_pre b)

| fence_declared_post {H b}
    : fence_declared H (fc_post b)
.


Inductive trans : history -> (* tr *) term -> thread -> history -> Prop :=
| trans_nothing {H p}
    : trans H tr_nothing p H

| trans_init {H fl i a p}
    : ~ H {{ ev_init i p | p }}
      -> trans H (tr_init fl i a) p 
         (H; ev_init i p; ev_is i a;; map (fun f => ev_fence f i) (fc_list fl))

| trans_read {H ir l p iw v}
    : H {{ ev_init ir p }}
      -> H {{ ev_is ir (ac_read l) }}
      -> executable H ir
      -> writes H iw l v
      -> exec H iw
      -> acyclic (co (H; ev_rf iw ir))
      -> trans H (tr_exec ir v) p
         (H; ev_rf iw ir)

| trans_write {H i l v p}
    : H {{ ev_init i p }}
      -> H {{ ev_is i (ac_write l v) }}
      -> executable H i
      -> acyclic (co (H; ev_exec i))
      -> trans H (tr_exec i tm_unit) p
         (H; ev_exec i)

| trans_rw {H ir vw l p iw v}
    : H {{ ev_init ir p }}
      -> H {{ ev_is ir (ac_rw l vw) }}
      -> executable H ir
      -> writes H iw l v
      -> exec H iw
      -> acyclic (co (H; ev_rf iw ir))
      -> trans H (tr_exec ir v) p
         (H; ev_rf iw ir)

| trans_push {H i p}
    : H {{ ev_init i p }}
      -> H {{ ev_is i ac_push }} 
      -> executable H i
      -> acyclic (co (H; ev_exec i))
      -> trans H (tr_exec i tm_unit) p
         (H; ev_exec i)

| trans_nop {H i p}
    : H {{ ev_init i p }}
      -> H {{ ev_is i ac_nop }}
      -> executable H i
      -> acyclic (co (H; ev_exec i))
      -> trans H (tr_exec i tm_unit) p
         (H; ev_exec i)

| trans_new {H b T T' p}
    : ~ tag_declared H T
      -> ~ tag_declared H T'
      -> T <> T'
      -> trans H (tr_new b T T') p
         (H; ev_edge b T T')
.


Inductive step : state -> state -> Prop :=
| step_sync {U P I U' P' I' H H' ex ex' d p}
    : exstep U P ex ex' d p
      -> trans H d p H'
      -> evolve U P I d p U' P' I'
      -> step (U, P, I, H, ex) (U', P', I', H', ex')
.
