
Require Import Tactics.
Require Import Relation.
Require Import Sequence.
Require Import Static.
Require Import Dynamic.
Require Import Subst.


(* trace coherence *)

Definition head (H : list event) (h : event) :=
  exists H', H = H'; h.


Inductive isac : term -> Prop :=
| isac_read {l}
    : isac (ac_read l)
| isac_write {l v}
    : isac (ac_write l v)
| isac_rw {l v}
    : isac (ac_rw l v)
| isac_push
    : isac ac_push
| iac_nop
    : isac ac_nop
.


Hint Constructors isac : isac.


Inductive trco : history -> Prop :=
| trco_nil
    : trco nil

| trco_init {H i p}
    : trco H
      -> ~ H {{ ev_init i pe | pe }}
      -> trco (H; ev_init i p)

| trco_is {H i a}
    : trco H
      -> isac a
      -> closed a
      -> (exists p, head H (ev_init i p))
      -> trco (H; ev_is i a)

| trco_exec {H i a}
    : trco H
      -> H {{ ev_is i a }}
      -> ~ exec H i
      -> (forall i', xo H i' i -> exec H i')
      -> (~ exists l, a = ac_read l)
      -> (~ exists l v, a = ac_rw l v)
      -> trco (H; ev_exec i)

| trco_rf {H iw ir l}
    : trco H
      -> writesto H iw l
      -> reads H ir l
      -> exec H iw
      -> ~ exec H ir
      -> (forall i, xo H i ir -> exec H i)
      -> trco (H; ev_rf iw ir)

| trco_edge {H b T T'}
    : trco H
      -> ~ tag_declared H T
      -> ~ tag_declared H T'
      -> T <> T'
      -> trco (H; ev_edge b T T')

| trco_fence {H f i}
    : trco H
      -> (exists a, head H (ev_is i a)) \/ (exists f', head H (ev_fence f' i))
      -> fence_declared H f
      -> trco (H; ev_fence f i)

| trco_co {H i i' l}
    : trco H
      -> writesto H i l
      -> writesto H i' l
      -> trco (H; ev_co i i')
.


(* Ambient type satisfaction *)

Definition tgtp_sat (H : history) (U : tgtp) : Prop :=
  tgtpof U
  /\ (forall T, In T U -> tag_declared H T)
  /\ (forall T, tag_declared H T -> In T U).


Definition initialized (H : history) (l : location) : Prop :=
  exists iw ip,
    writesto H iw l
    /\ H {{ ev_is ip ac_push }}
    /\ vop H iw ip
    /\ exec H ip
    /\ forall ir, reads H ir l -> to H ip ir.
  

Definition lctp_sat (U : tgtp) (Pl : lctp) (H : history) (P : lctp) : Prop :=
  lctpof P
  /\ (forall l t,
        In (l, t) P
        -> initialized H l)
  /\ (forall l t i v,
        In (l, t) P
        -> writes H i l v
        -> value nil v /\ tmof U Pl nil v t)
  /\ (forall i l v,
        writes H i l v
        -> indom l P).


Definition idtp_sat (U : tgtp) (P : lctp) (H : history) (I : idtp) : Prop :=
  idtpof I
  /\ (forall i t p,
        In (i, (t, p)) I
        -> exists a, H {{ ev_init i p }} /\ H {{ ev_is i a }} /\ acof U P nil a t /\ ~ exec H i)
  /\ (forall i p,
        H {{ ev_init i p }}
        -> ~ exec H i
        -> indom i I)
  /\ (forall I1 I2 I3 i t i' t' p,
        I = I1; (i, (t, p));; I2; (i', (t', p));; I3
        -> po H i i').


Definition hok (H : history) (U : tgtp) (P : lctp) (I : idtp) : Prop :=
  tgtp_sat H U
  /\ lctp_sat U P H P
  /\ idtp_sat U P H I
  /\ trco H
  /\ acyclic (co H).


Inductive ok : state -> Prop :=
| okay {U P I H ex}
    : hok H U P I
      -> exof U P I ex
      -> ok (U, P, I, H, ex)
.
