
Require Import Tactics.
Require Import TSubst.
Require Import Static.
Require Import Dynamic.
Require Import Hwf.
Require Import Relation.
Require Import History.
Require Import Truncate.
Require Import Sequence.
Require Import Ambients.
Require Import Strengthen.
Require Import Unique.
Require Import Match.
Require Import Decide.
Require Import Acyclic.


(* Term preservation *)

Lemma preservation_tm :
  forall U P G m t m',
    tmof U P G m t
    -> tstep G m m'
    -> tmof U P G m' t.
Proof.
intros U P G m t m' Hofm Hstep.
revert U P t Hofm.
induct Hstep;
try (intros;
     repeat (match goal with
             | H : tmof _ _ _ ?m _ |- _ => isapp m; invert H; []; clear H; intros; subst
             end);
     eauto with static;
     done).

(* app3 *)
intros G t m1 v2 H U P t' Hofm.
have (tmof U P G (tm_app (tm_lam t m1) v2) t') as Hofm.
toshow (tmof U P G (subst1 v2 m1) t').
invert Hofm; intros t1 Hoflam Hofv2.
invert Hoflam; intros Hoft' Hofm1 Heqt1.
subst t1.
have (tmof U P (G; cl_vl t) m1 (shift1 t')) as Hofm1.
have (tmof U P G v2 t) as Hofv2.
thus (tmof U P G (subst1 v2 m1) (subst1 v2 (shift1 t'))) as Hof by subst1_tm_tm.
simp_sub in Hof.
exact Hof.

(* app3' *)
intros G t1 t2 m1 v2 H U P t Hofm.
invert Hofm.
intros t1' Hoffun Hofv2.
invert Hoffun.
intros Hoft1 Hoft2 Hofm1 <- <-.
assert (subof (ofd U P) G (dot v2 (dot (tm_fun t1 t2 m1) id)) (G; cl_vl (tp_arrow t1 t2); cl_vl (shift1 t1))) as Hsubof.
  apply subof_dot; [|].
    apply subof_dot; [|].
      apply subof_id; done.

      simp_sub.
      auto with static; done.

    simp_sub.
    auto with static; done.
so (subst_tm _#7 Hsubof Hofm1) as Hofm1'.
simp_sub in Hofm1'.
exact Hofm1'.
Qed.


(* Restriction *)

Definition eq_thread_bool (p p' : thread) := if eq_thread_dec p p' then true else false.

Lemma eq_thread_bool_true :
  forall p p',
    p = p' <-> eq_thread_bool p p' = true.
Proof.
intros p p'.
unfold eq_thread_bool.
remember (eq_thread_dec p p') as b.
destruct b as [Heq | Hneq].
  split; auto; done.

  split.
    intro Heq.
    subst; destruct Hneq; auto; done.

    intro HH.
    discriminate HH.
Qed.


Lemma eq_thread_bool_false :
  forall p p',
    p <> p' <-> eq_thread_bool p p' = false.
Proof.
intros p p'.
unfold eq_thread_bool.
remember (eq_thread_dec p p') as b.
destruct b as [Heq | Hneq].
  split.
    intro Hneq.
    findcontra.

    intro HH.
    discriminate HH.

  split; auto; done.
Qed.


Lemma restricteq_filter :
  forall I p,
    restricteq I p = filter (fun itp : (identifier * (term * thread)) => let (_, tp) := itp in let (_, p') := tp in eq_thread_bool p p') I.
Proof.
intros I p.
induction I as [| (i, (t, p')) I IH].

(* nil *)
reflexivity.

(* cons *)
simpl.
unfold eq_thread_bool.
generalize (eq_thread_dec p p').
intros [Heq | Hneq].
  f_equal; [].
  exact IH.

  exact IH.
Qed.


Lemma restrictneq_filter :
  forall I p,
    restrictneq I p = filter (fun itp : (identifier * (term * thread)) => let (_, tp) := itp in let (_, p') := tp in negb (eq_thread_bool p p')) I.
Proof.
intros I p.
induction I as [| (i, (t, p')) I IH].

(* nil *)
reflexivity.

(* cons *)
simpl.
unfold eq_thread_bool.
generalize (eq_thread_dec p p').
intros [Heq | Hneq].
  simpl.
  exact IH.

  simpl.
  f_equal; [].
  exact IH.
Qed.


Lemma restricteq_app :
  forall I1 I2 p,
    restricteq (I1;; I2) p = restricteq I1 p;; restricteq I2 p.
Proof.
intros I1 I2 p.
rewrite -> !restricteq_filter.
apply filter_app.
Qed.


Lemma restrictneq_app :
  forall I1 I2 p,
    restrictneq (I1;; I2) p = restrictneq I1 p;; restrictneq I2 p.
Proof.
intros I1 I2 p.
rewrite -> !restrictneq_filter.
apply filter_app.
Qed.


Lemma restricteq_cons_in :
  forall I i t p,
    restricteq (I; (i, (t, p))) p = restricteq I p; (i, (t, p)).
Proof.
intros I i t p.
rewrite -> !restricteq_filter.
rewrite -> filter_cons_in; auto.
apply eq_thread_bool_true; auto.
Qed.


Lemma restricteq_cons_out :
  forall I i t p p',
    p <> p'
    -> restricteq (I; (i, (t, p'))) p = restricteq I p.
Proof.
intros I i t p p' Hneq.
rewrite -> !restricteq_filter.
rewrite -> filter_cons_out; auto.
apply eq_thread_bool_false; auto.
Qed.


Lemma restrictneq_cons_out :
  forall I i t p,
    restrictneq (I; (i, (t, p))) p = restrictneq I p.
Proof.
intros I i t p.
rewrite -> !restrictneq_filter.
rewrite -> filter_cons_out; auto.
apply Bool.negb_false_iff.
apply eq_thread_bool_true; auto.
Qed.


Lemma restrictneq_cons_in :
  forall I i t p p',
    p <> p'
    -> restrictneq (I; (i, (t, p'))) p = restrictneq I p; (i, (t, p')).
Proof.
intros I i t p p' Hneq.
rewrite -> !restrictneq_filter.
rewrite -> filter_cons_in; auto.
apply Bool.negb_true_iff.
apply eq_thread_bool_false; auto.
Qed.


Lemma lookup_restricteq :
  forall I i t p p',
    lookup i (restricteq I p) (t, p')
    -> p = p'.
Proof.
intros I i t p p' Hlookup.
rewrite -> restricteq_filter in Hlookup.
apply eq_thread_bool_true.
exact (lookup_filter _#6 Hlookup).
Qed.


Lemma restricteq_contain :
  forall I p i x,
    lookup i (restricteq I p) x
    -> lookup i I x.
Proof.
intros I p i x Hlookup.
rewrite -> restricteq_filter in Hlookup.
eapply filter_contain; eauto.
Qed.


Lemma restrictneq_contain :
  forall I p i x,
    lookup i (restrictneq I p) x
    -> lookup i I x.
Proof.
intros I p i x Hlookup.
rewrite -> restrictneq_filter in Hlookup.
eapply filter_contain; eauto.
Qed.


Lemma restricteq_ddistinct :
  forall I p,
    ddistinct I
    -> ddistinct (restricteq I p).
Proof.
intros I p Hdist.
rewrite -> restricteq_filter.
apply filter_ddistinct; auto.
Qed.


Lemma restrictneq_ddistinct :
  forall I p,
    ddistinct I
    -> ddistinct (restrictneq I p).
Proof.
intros I p Hdist.
rewrite -> restrictneq_filter.
apply filter_ddistinct; auto.
Qed.


Lemma indom_restricteq :
  forall i I p,
    indom i (restricteq I p) -> indom i I.
Proof.
intros i I p Hindom.
rewrite -> restricteq_filter in Hindom.
eapply indom_filter; eauto.
Qed.


Lemma indom_restrictneq :
  forall i I p,
    indom i (restrictneq I p) -> indom i I.
Proof.
intros i I p Hindom.
rewrite -> restrictneq_filter in Hindom.
eapply indom_filter; eauto.
Qed.


Lemma in_restricteq :
  forall I i t p,
    In (i, (t, p)) I
    -> In (i, (t, p)) (restricteq I p).
Proof.
intros I i t p Hin.
rewrite -> restricteq_filter.
apply filter_In.
split; auto; [].
apply eq_thread_bool_true; auto.
Qed.


Lemma in_restrictneq :
  forall I i t p p',
    In (i, (t, p)) I
    -> p <> p'
    -> In (i, (t, p)) (restrictneq I p').
Proof.
intros I i t p p' Hin Hneq.
rewrite -> restrictneq_filter.
apply filter_In.
split; auto; [].
apply Bool.negb_true_iff; [].
apply eq_thread_bool_false; auto; done.
Qed.


Lemma restricteq_restrictneq :
  forall I p p',
    p <> p'
    -> restricteq (restrictneq I p') p = restricteq I p.
Proof.
intros I p1 p2 Hneq.
rewrite -> restrictneq_filter.
rewrite -> !restricteq_filter.
rewrite -> filter_and.
apply filter_iff.
intros (i, (t, p)).
remember (eq_thread_bool p1 p) as b.
destruct b.
  thus (p1 = p) by eq_thread_bool_true.
  subst p.
  rewrite -> (eq_thread_bool_false p2 p1 andel); auto; done.

  reflexivity.
Qed.


(* Monotonicity *)

Lemma tcof_mono :
  forall U U' G T,
    tgof U G T
    -> tgof (U;; U') G T.
Proof.
intros U U' G T H.
destruct H; eauto using in_or_app with static.
Qed.


Lemma fcof_mono :
  forall U U' G f,
    fcof U G f
    -> fcof (U;; U') G f.
Proof.
intros U U' G f H.
destruct H; eauto using tcof_mono with static.
Qed.


Lemma tmof_xpof_mono :
  (forall U U' P P' G m t,
     tmof U P G m t
     -> tmof (U;; U') (P;; P') G m t)
  /\
  (forall U U' P P' I G e t,
     xpof U P I G e t
     -> xpof (U;; U') (P;; P') I G e t).
Proof.
assert (forall U P G e t, xpof U P (nil;; nil) G e t -> xpof U P nil G e t) as Hnil by auto.
exploit
  (tmof_xpof_ind
     (fun U P G m t =>
        forall U' P',
          tmof U P G m t
          -> tmof (U;; U') (P;; P') G m t)
     (fun U P I G e t =>
        forall U' P',
          xpof U P I G e t
          -> xpof (U;; U') (P;; P') I G e t)) as Hind;
[.. | destruct Hind as (Htm & Hxp); repeat split; eauto];
eauto using lookup_app_right, in_or_app, fcof_mono with static.
Qed.


Lemma tmof_mono :
  forall U U' P P' G m t,
    tmof U P G m t
    -> tmof (U;; U') (P;; P') G m t.
Proof.
exact (carp tmof_xpof_mono).
Qed.


Lemma xpof_mono :
  forall U U' P P' I G e t,
    xpof U P I G e t
    -> xpof (U;; U') (P;; P') I G e t.
Proof.
exact (cdrp tmof_xpof_mono).
Qed.


Lemma acof_mono :
  forall U U' P P' G a t,
    acof U P G a t
    -> acof (U;; U') (P;; P') G a t.
Proof.
intros U1 U2 P1 P2 G a t Hofa.
induct Hofa;
eauto using lookup_app_right, tmof_mono with static.
Qed.


Lemma exof_mono :
  forall U U' P P' I ex,
    exof U P I ex
    -> exof (U;; U') (P;; P') I ex.
Proof.
intros U U' P P' I ex Hofex.
induct Hofex.
(* nil *)
intros U P.
apply exof_nil.

(* cons *)
intros U P I ex p e t Hnindom Hofex IH Hofe.
eapply exof_cons; eauto.
apply xpof_mono; eauto.
Qed.


Lemma lctp_sat_mono :
  forall U U' P P' H P'',
    lctp_sat U P H P''
    -> lctp_sat (U;; U') (P;; P') H P''.
Proof.
intros U1 U2 P1 P2 H P Hsat.
destruct Hsat as (HofP & Hallsees & Hallof & Hisin).
do2 3 split; auto.
intros l t i v Hin His.
so (Hallof _#4 Hin His) as (Hvalue & Hofv).
split; auto.
apply tmof_mono; auto.
Qed.


Lemma idtp_sat_mono :
  forall U U' P P' H I,
    idtp_sat U P H I
    -> idtp_sat (U;; U') (P;; P') H I.
Proof.
intros U1 U2 P1 P2 H I Hsat.
destruct Hsat as (HofI & Hinis & Hinitin & Horder).
do2 3 split; auto; [].
intros i t p Hin.
so (Hinis _#3 Hin) as (a & Hinit & His & Hofa & Hexec).
exists a.
do2 3 split; auto; [].
apply acof_mono; auto; done.
Qed.


Lemma xstep_mono :
  forall U U' P P' G e e' d,
    xstep U P G e e' d
    -> xstep (U;; U') (P;; P') G e e' d.
Proof.
intros U U' P P' G e e' d Hstep.
induct Hstep;
eauto using tmof_mono with dynamic.
Qed.


Lemma exstep_mono :
  forall U U' P P' ex ex' d p,
    exstep U P ex ex' d p
    -> exstep (U;; U') (P;; P') ex ex' d p.
Proof.
intros U U' P P' ex ex' d p Hstepex.
induct Hstepex.
(* hit *)
intros U P ex p e e' d Hstep.
apply exstep_hit; [].
apply xstep_mono; [].
assumption.

(* miss *)
intros U P ex ex' p e d p' Hstepex IH.
apply exstep_miss.
assumption.
Qed.


(* Ambient resolution *)

Lemma lctp_sat_resolve :
  forall U P P' H l i v t,
    lctp_sat U P H P'
    -> writes H i l v
    -> lookup l P' t
    -> value nil v /\ tmof U P nil v t.
Proof.
intros U P P' H l i v t HsatP Hwrite Hlookup.
destruct HsatP as (_ & _ & Hallof & _).
eapply Hallof; eauto.
apply lookup_iff_in; eassumption.
Qed.


Lemma idtp_sat_resolve :
  forall U P I H i a t p,
    trco H
    -> idtp_sat U P H I
    -> H {{ ev_is i a }}
    -> lookup i I (t, p)
    -> acof U P nil a t.
Proof.
intros U P I H i a t p Htrco HsatI His Hlookup.
destruct HsatI as (_ & Hallof & _).
thus (In (i, (t, p)) I) as Hin by lookup_iff_in.
so (Hallof _#3 Hin) as (a' & _ & His' & Hofa & _).
thus (a = a') as Heq by is_fun.
subst a'.
assumption.
Qed.


Lemma idtp_sat_init :
  forall U P I H i t p p',
    trco H
    -> idtp_sat U P H I
    -> lookup i (restricteq I p) (t, p')
    -> H {{ ev_init i p }}.
Proof.
intros U P I H i t p p' Htrco Hsat Hlookup.
thus (p = p') by lookup_restricteq.
subst p'.
thus (lookup i I (t, p)) as Hlookup' by restricteq_contain.
thus (In (i, (t, p)) I) as Hin by lookup_iff_in.
destruct Hsat as (_ & Hininit & _).
so (Hininit _#3 Hin) as (_ & Hinit & _).
exact Hinit.
Qed.


(* Trofo *)

Lemma trofo_mono_idtp :
  forall U P I I' G d,
    (forall x, In x I -> In x I')
    -> trofo U P I G d
    -> trofo U P I' G d.
Proof.
intros U P I I' G d Hincl Htrofo.
assert (forall i tp, lookup i I tp -> lookup i I' tp).
  intros.
  apply lookup_iff_in; [].
  apply Hincl; [].
  apply lookup_iff_in; assumption.
destruct Htrofo; eauto with static; done.
Qed.


Lemma trofo_addfence :
  forall U P I G f d,
    fcof U G f
    -> trofo U P I G d
    -> trofo U P I G (addfence f d).
Proof.
intros U P I G f d Hf Hd.
destruct d; auto.
rename d1 into fl, d2 into a, Hd into Hofinit.
simpl.
invert Hofinit; intros t Hoffl Hofa.
eauto with static.
Qed.


Lemma xstep_trofo :
  forall U P I G e t d e',
    xpof U P I G e t
    -> xstep U P G e e' d
    -> trofo U P I G d.
Proof.
intros U P I G e t d e' Hof Hstep.
revert I t Hof.
#6 induction Hstep; eauto with static;
try (intros;
     (match goal with
       H : (xpof _ _ _ _ _ _) |- _ => complete (inversion H; subst; eauto using lookup_hit with static)
     end)).

(* read *)
intros I t Hof.
invert Hof; intros Hofloc.
invert Hofloc; intros t' Hlookup Heqt.
eauto with static; done.

(* write *)
intros I t Hof.
invert Hof; intros t' Hofloc Hofv Heqt.
invert Hofloc; intros t'' Hlookup Heqt'.
subst t'.
eauto with static; done.

(* rw *)
intros I t Hof.
invert Hof; intros Hofloc Hofv Heqt.
invert Hofloc; intros t'' Hlookup Heqt'.
subst t.
eauto with static; done.

(* bind1 *)
intros I t' Hof.
invert Hof; [|].
  intros Hoft Hofe1 Hofe2.
  eapply IHHstep; eauto; done.

  intros I1 I2 _ Hoft Hofe1 Hofe2 <-.
  eapply (trofo_mono_idtp _ _ I1); [|].
    intros; apply in_or_app; right; auto; done.

    eapply IHHstep; eauto; done.

(* bind2 *)
intros I t' Hof.
have (xstep U P (G; cl_vl t) e2 e2' (shift1 d)) as Hstep.
have (xpof U P I G (xp_bind t e1 e2) t') as Hof.
have (forall (I : idtp) (t0 : term),
      xpof U P I (G; cl_vl t) e2 t0 ->
      trofo U P I (G; cl_vl t) (shift1 d)) as IHHstep.
toshow (trofo U P I G d).
invert Hof; [|].
  intros Hoft Hofe1 Hofe2.
  thus (trofo U P nil (G; cl_vl t) (shift1 d)) as Hd by IHHstep.
  eapply (trofo_mono_idtp _ _ nil); [|].
    intros ? Hin; destruct Hin; done.

    eapply strengthen_trofo; eauto; done.

  intros I1 I2 _ Hoft Hofe1 Hofe2 <-.
  thus (trofo U P I2 (G; cl_vl t) (shift1 d)) as Hd by IHHstep.
  eapply (trofo_mono_idtp _ _ I2); [|].
    intros; apply in_or_app; left; auto; done.
  
    eapply strengthen_trofo; eauto; done.

(* fence *)
intros I t Hoffence.
have (xstep U P G e e' d) as Hstep.
have (xpof U P I G (xp_fence f e) t) as Hoffence.
have (forall (I : idtp) (t : term), xpof U P I G e t -> trofo U P I G d) as IHHstep.
toshow (trofo U P I G (addfence f d)).
invert Hoffence; intros Hoff Hofe.
thus (trofo U P I G d) as Hd by IHHstep.
apply trofo_addfence; auto.
Qed.


Lemma exstep_indom :
  forall U P ex ex' d p,
    exstep U P ex ex' d p
    -> indom p ex.
Proof.
intros U P ex ex' d p Hstep.
induction Hstep.
  apply indom_hit; auto; done.

  apply indom_miss; auto; done.
Qed.


Lemma exstep_trofo :
  forall U P I ex d p ex',
    exof U P I ex
    -> exstep U P ex ex' d p
    -> trofo U P (restricteq I p) nil d.
Proof.
intros U P I ex d p ex' Hofex Hstepex.
revert I Hofex.
induct Hstepex.
(* hit *)
intros U P ex p e e' d Hstepe I Hofex.
invert Hofex; intros t _ Hofex' Hofe.
eapply xstep_trofo; eauto; done.

(* miss *)
intros U P ex ex' p e d p' Hstep IH I Hofex.
invert Hofex; intros t _ Hofex' Hofe.
so (eq_thread_dec p p') as [Heqp | Hneq].
  subst p'.
  invert Hofex.
  intros _ Hnindom _ _.
  destruct Hnindom.
  eapply exstep_indom; eauto; done.
  
  rewrite <- (restricteq_restrictneq I p' p); auto; done.
Qed.


(* Evolution *)

Lemma evolve_extends :
  forall U1 U2 P1 P2 I1 I2 d p,
    evolve U1 P1 I1 d p U2 P2 I2
    -> (exists U, U2 = U1;; U)
       /\ (exists P, P2 = P1;; P).
Proof.
intros U1 U2 P1 P2 I1 I2 d p Hevolve.
induct Hevolve;
intros; split; try (exists nil; auto; done); [].
exists (nil; T; T'); reflexivity.
Qed.


Lemma exstep_evolved :
  forall U U' P P' I I' d ex ex' d' p,
    evolve U P I d p U' P' I'
    -> exstep U P ex ex' d' p
    -> exstep U' P' ex ex' d' p.
Proof.
intros U U' P P' I I' d ex ex' d' p Hevolve H.
pose proof (evolve_extends _#8 Hevolve) as ((U'' & HeqU) & (P'' & HeqP)).
subst U' P'.
apply exstep_mono; auto.
Qed.


Lemma evolve_addfence :
  forall U P I f d p U' P' I',
    evolve U P I (addfence f d) p U' P' I'
    -> evolve U P I d p U' P' I'.
Proof.
intros U P I f d p U' P' I' Hevolve.
induction d; invert Hevolve; try (intros; subst; auto; done); [].
intros ? ? Hoffl ? ? ? ?.
subst U' P' I'.
apply evolve_init; auto; [].
invert Hoffl ; auto; done.
Qed.


Lemma evolve_restricteq_eq :
  forall U P I U' P' I' d p p',
    evolve U P I d p U' P' I'
    -> p <> p'
    -> restricteq I p' = restricteq I' p'.
Proof.
intros U P I U' P' I' d p p' Hevolve Hneq.
destruct Hevolve; auto;
try (rewrite -> ?restricteq_app;
     simpl;
     generalize (eq_thread_dec p' p);
     (intros [Heq | _]; [findcontra | reflexivity]);
     done).
Qed.


Lemma restricteq_evolve :
  forall U P I U' P' I' d p,
    evolve U P I d p U' P' I'
    -> evolve U P (restricteq I p) d p U' P' (restricteq I' p).
Proof.
intros U P I U' P' I' d p Hevolve.
destruct Hevolve;
auto with static;
rewrite -> ?restricteq_app;
simpl;
(generalize (eq_thread_dec p p); intros [Heq | Hneq]; [| destruct Hneq; auto; done]);
auto with static;
[].
apply evolve_init; auto; [].
contradict H.
eapply indom_restricteq; eauto; done.
Qed.


Lemma restrictneq_evolve :
  forall U P I U' P' I' d p p',
    p <> p'
    -> evolve U P I d p U' P' I'
    -> evolve U P (restrictneq I p') d p U' P' (restrictneq I' p').
Proof.
intros U P I U' P' I' d p p' Hneq Hevolve.
destruct Hevolve; auto with static; [|].
  rewrite -> restrictneq_cons_in; auto; [].
  apply evolve_init; auto; [].
  contradict H.
  eapply indom_restrictneq; eauto; done.

  rewrite -> !restrictneq_app.
  rewrite -> restrictneq_cons_in; auto; [].
  apply evolve_exec; eauto; done.
Qed.


Lemma evolve_restrictneq_eq :
  forall U P I U' P' I' d p,
    evolve U P I d p U' P' I'
    -> restrictneq I p = restrictneq I' p.
Proof.
intros U P I U' P' I' d p Hevolve.
destruct Hevolve; auto; [|].
  rewrite -> restrictneq_cons_out.
  reflexivity.
  
  rewrite -> !restrictneq_app.
  f_equal.
  apply restrictneq_cons_out; done.
Qed.


(* Expression preservation *)

Inductive acorr : (* ax *) term -> (* ac *) term -> Prop :=
| acorr_read {l}
    : acorr (xp_read (tm_loc l)) (ac_read l)

| acorr_write {l v}
    : acorr (xp_write (tm_loc l) v) (ac_write l v)

| acorr_rw {l v}
    : acorr (xp_rw (tm_loc l) v) (ac_rw l v)

| acorr_push
    : acorr xp_push ac_push

| acorr_nop
    : acorr xp_nop ac_nop
.


Hint Constructors acorr : acorr.


Lemma acorr_xpof_acof :
  forall U P I G e t a t',
    ddistinct P
    -> acorr e a
    -> xpof U P I G e t
    -> acof U P nil a t'
    -> t = subst (sh (length G)) t'.
Proof.
intros U P I G e t1 a t2 HdistP Hcorr Hofe Hofa.
destruct Hcorr; try (complete (inversion Hofe; inversion Hofa; auto)).

(* read *)
invert Hofe; intros Hofloc.
invert Hofloc; intros t1' Hlookup1 <- <-.
have (lookup l P t1') as Hlookup1.
f_equal.
toshow (t1' = t2).
invert Hofa; intros t2' Hlookup2 Heqt2; rewrite <- Heqt2.
simp_sub.
eapply ddistinct_lookup_unique; eauto; done.

(* rw *)
invert Hofe.
intros Hofloc Hofv HeqI.
invert Hofloc.
intros t1' Hlookup1 <-.
f_equal.
invert Hofa.
intros t2' Hlookup2 Hvalue Hofv' Heqt2.
rewrite <- Heqt2.
simp_sub.
eapply ddistinct_lookup_unique; eauto; done.
Qed.


Lemma init_xstep_tr :
  forall U P G e e' fl i a,
    init e
    -> xstep U P G e e' (tr_init fl i a)
    -> False.
Proof.
intros U P G e e' fl i a Hinit Hstep.
remember (tr_init fl i a) as d eqn:Heqd in Hstep.
revert fl i a Heqd Hinit.
induct Hstep;
try (intros; first [discriminate Heqd | invert Hinit; intros; eapplyall; eauto; done]);
[].

(* bind2 *)
intros U P G t e1 e2 e2' d _ Hstep IH fl i a Heqd Hinit.
invert Hinit; intros.
eapply IH; eauto; [].
rewrite -> Heqd.
simp_sub.
reflexivity.
Qed.


Lemma preservation_init :
  forall U P G e e' d,
    init e
    -> xstep U P G e e' d
    -> init e'.
Proof.
intros U P G e e' d Hinit Hstep.
revert Hinit.
induct Hstep; try (intros; invert Hinit; auto with static; done).

(* bind3 *)
intros U P G t v e Hval Hinit.
apply subst_init.
invert Hinit; auto; done.

(* spec1 *)
intros U P G v1 v2 t e Hofv1 Hofv2 Hval1 Hval2 Hinit.
apply init_spec; [].
apply subst_init; [].
replace (subst1 v1 e) with (subst (under 0 (dot v1 id)) e) in Hinit by (simp_sub; auto).
eapply unsubst_init; eauto; [].
intro Hinit'.
exact (tmof_init_disjoint _#5 Hofv1 Hinit').
Qed.


Lemma addfence_closed :
  forall f d,
    closed (addfence f d)
    -> closed d.
Proof.
intros f d Hclosed.
destruct d;
simpl in Hclosed; try (reduce_closed in Hclosed; intros); reduce_closed; auto.
Qed.


Lemma preservation_xp :
  forall U U' P P' I1 I2 I3 I' G e t e' d p,
    ddistinct P
    -> ddistinct (I1;; I2;; I3)
    -> (forall fl i a, d = tr_init fl i a -> I3 = nil)
    -> closed d
    -> evolve U P (I1;; I2;; I3) d p U' P' I'
    -> xpof U P I2 G e t
    -> xstep U P G e e' d
    -> exists I2', I' = I1;; I2';; I3 /\ xpof U' P' I2' G e' t.
Proof.
intros U U' P P' I1 I2 I3 I' G e t e' d p HdistP HdistI Hzone Hclosed Hevolve Hofe Hstep.
revert I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
#9 induct Hstep;
try (solve [
           (* simple cases *)
           intros;
           simp_sub in Hevolve; invert Hevolve; intros; subst;
           repeat (match goal with
                   | H : xpof _ _ _ _ ?e _ |- _ => isapp e; invert H; []; clear H; intros; subst
                   | H : tmof _ _ _ ?m _ |- _ => isapp m; invert H; []; clear H; intros; subst
                   end);
           eexists; eauto using preservation_tm with static
           |
           (* init cases *)
           intros;
           (eexploit Hzone; eauto; []);
           subst I3;
           simp_sub in Hevolve; invert Hevolve; intros t' _ _ Hofa <- <- <-;
           assert (t = subst (sh (length G)) t') by (eapply acorr_xpof_acof; eauto; eauto with acorr);
           subst t;
           invert Hofe; intros; subst I2;
           evargen ((identifier * (term * thread))%type) ltac:(fun x => exists (x :: nil));
           split; [reflexivity | apply xpof_action]
           ]).

(* rmw2 *)
intros U P G t v1 m2 v Hvaluev1 Hofv Hvaluev I1 I2 I3 t' HdistP HdistI Hzone Hclosed Hevolve Hofe.
invert Hevolve; intros; subst U' P' I'.
invert Hofe.
intros Hoft Hofv1 Hofm2 <- <-.
exists nil.
split; [reflexivity |].
#2 apply xpof_bind; auto.
  apply xpof_rw; auto; [].
  so (subst1_tm_tm _#7 Hofv Hvaluev Hofm2) as Hofm2'.
  simp_sub in Hofm2'.
  exact Hofm2'.

  #5 apply (@xpof_spec U P nil (G; cl_vl t) (var 0) (shift1 v) (shift1 t)).
    apply tmof_var; [].
    apply index_0; done.

    so (weaken_tmof _ _ _ (nil; cl_vl t) _ _ Hofv) as Hof.
    simpl in Hof; simp_sub in Hof.
    exact Hof.

    apply (@value_var _ _ t).
    apply index_0; done.

    eapply (subst_value U P); eauto; [].
    apply subof_sh; auto with sequence; done.

    apply xpof_ret; [].
    so (weaken_tmof _ _ _ (nil; cl_vl t) _ _ Hofv) as Hof.
    simpl in Hof; simp_sub in Hof.
    exact Hof.

(* exec *)
intros U P G i v I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
simp_sub in Hevolve.
invert Hevolve; intros I1' I3' t' Hofv HeqI <- <- <-.
invert Hofe; intros t'' p' <- Heqt.
simpl in HeqI, HdistI.
symmetry in HeqI.
so (ddistinct_partition _#9 HdistI HeqI) as (Heq1 & Heq2 & Heq3).
injection Heq2; clear Heq2; intros.
subst I1' I3' p' t''.
exists nil.
split; auto; [].
apply xpof_ret; [].
have (closed (tr_exec i v)) as Hclosed.
have (tmof U P nil v t') as Hofv.
toshow (tmof U P G v t).
reduce_closed in Hclosed; intro Hclosedv.
rewrite <- Heqt.
rewrite <- (closed_subst v (sh (length G))); [| assumption].
eapply subst_tm; eauto; [].
apply subof_sh; [].
apply truncate_all; done.

(* bind1 *)
intros U P G t1 e1 e1' e2 d Hstep IH I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
invert Hofe.
  (* e1 not inited *)
  intros Hoft Hofe1 Hofe2.
  eexploit IH as (I2' & HeqI' & Hofe1'); eauto; [].
  exists I2'.
  split; auto; [].
  apply xpof_bind; auto; [].
  so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
  apply xpof_mono; auto; done.

  (* e1 inited *)
  intros I2a I2b Hinit Hoft1 Hofe1 Hofe2 <-.
  repl (I1;; (I2a;; I2b);; I3) with (I1;; I2a;; (I2b;; I3)) in HdistI by (rewrite <- !app_assoc; auto).
  repl (I1;; (I2a;; I2b);; I3) with (I1;; I2a;; (I2b;; I3)) in Hevolve by (rewrite <- !app_assoc; auto).
  eexploit (IH I1 I2a (I2b;; I3)) as (I2a' & -> & Hofe1'); eauto; [|].
    toshow (forall (fl : term) (i : identifier) (a : term),
            d = tr_init fl i a -> I2b;; I3 = nil).
    intros fl i a Heqd'.
    exfalso.
    subst d.
    eapply init_xstep_tr; eauto; done.
  exists (I2a';; I2b).
  split.
    rewrite <- !app_assoc; auto; done.

    apply xpof_bind_init; auto; [|].
      eapply preservation_init; eauto; done.

      so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
      apply xpof_mono; auto; done.

(* bind2 *)
intros U P G t1 e1 e2 e2' d Hinit Hstep IH I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
assert (exists I2a I2b, I2 = I2a;; I2b /\ tpof G t1 /\ xpof U P I2a G e1 t1 /\ xpof U P I2b (G; cl_vl t1) e2 (shift1 t)) as Hprop.
  invert Hofe; [|].
    intros.
    exists I2, nil.
    do2 3 split; auto; done.

    intros; do 2 eexists; do2 3 split; eauto; done.
destruct Hprop as (I2a & I2b & -> & Hoft1 & Hofe1 & Hofe2).
repl (I1;; (I2a;; I2b);; I3) with ((I1;; I2a);; I2b;; I3) in HdistI by (rewrite <- !app_assoc; auto).
repl (I1;; (I2a;; I2b);; I3) with ((I1;; I2a);; I2b;; I3) in Hevolve by (rewrite <- !app_assoc; auto).
eexploit (IH (I1;; I2a) I2b I3) as (I2b' & -> & Hofe2'); eauto; [| | |].
  intros fl i a Heq.
  f_apply (fun d' => subst (dot nonsense id) d') in Heq as Heq'.
  simp_sub in Heq'.
  eapply Hzone; eauto; done.

  apply subst_closed; auto; done.

  unfold shift1; rewrite -> closed_subst; auto; done.
exists (I2a;; I2b').
  split.
    rewrite <- !app_assoc; auto; done.

    apply xpof_bind_init; auto.
    so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
    apply xpof_mono; auto; done.

(* bind3 *)
intros U P G t1 v e1 Hval I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
simp_sub in Hevolve.
invert Hevolve; intros <- <- <-.
exists I2.
split; auto; [].
assert (exists I2a I2b, I2 = I2a;; I2b /\ tpof G t1 /\ xpof U P I2a G (xp_ret v) t1 /\ xpof U P I2b (G; cl_vl t1) e1 (shift1 t)) as Hprop; [|].
  invert Hofe; [|].
    intros.
    exists I2, nil.
    do2 3 split; auto; done.

    intros; do 2 eexists; do2 3 split; eauto; done.
destruct Hprop as (I2a & I2b & -> & Hoft1 & Hofret & Hofe1).
invert Hofret; intros Hofv <-.
repl t with (subst1 v (shift1 t)) by (simp_sub; auto).
rewrite -> app_nil_r.
eapply subst1_tm_xp; eauto; done.

(* spec1 *)
intros U P G v1 v2 t1 e1 Hofv1 Hofv2 Hval1 Hval2 I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
simp_sub in Hevolve; invert Hevolve; intros <- <- <-.
exists I2.
split; auto; [].
eapply xpof_spec; eauto; [].
so (ddistinct_app _#4 HdistI) as (_ & HdistI12).
so (ddistinct_app _#4 HdistI12) as (HdistI2 & _).
so (xpof_factor _#8 HdistP HdistI2 Hofv1 Hofe) as Hofe'.
replace t with (subst1 v2 (shift1 t)) by (simp_sub; auto).
eapply subst1_tm_xp; eauto; done.

(* spec2 *)
intros U P G e e' d v1 v2 Hstep IH I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
invert Hofe; intros t' Hofv1 Hofv2 Hval1 Hval2 Hofe'.
eexploit (IH I1 I2 I3) as (I2' & -> & Hofe''); eauto; [].
exists I2'.
split; auto; [].
so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
eapply xpof_spec; eauto using tmof_mono; done.

(* new *)
intros U P G e b T1 T2 I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
simp_sub in Hevolve; invert Hevolve; intros Hfresh1 Hfresh2 Hneq <- <- <-.
invert Hofe; intros Hofe' <-.
exists nil.
split; auto; [].
replace t with (subst (dot (tg_lit T2) (dot (tg_lit T1) id)) (subst (sh 2) t)) by (simp_sub; auto).
apply (subst_xp _#4 (G; cl_tg; cl_tg)); auto; [|].
  apply subof_dot; [|].
    apply subof_dot; [|].
      apply subof_id; done.

      apply ofd_tg; [].
      apply tgof_lit; [].
      right; left; reflexivity.

    apply ofd_tg; [].
    apply tgof_lit; [].
    left; reflexivity.

  so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
  apply xpof_mono; auto; done.

(* fence *)
intros U P G f e e' d Hstep IH I1 I2 I3 t HdistP HdistI Hzone Hclosed Hevolve Hofe.
so (evolve_addfence _#9 Hevolve) as Hevolve'.
invert Hofe; intros Hoff Hofe'.
eexploit (IH I1 I2 I3) as (I2' & -> & Hofe''); eauto; [| |].
  intros fl i a Heq.
  subst d.
  eapply Hzone; eauto.
  simpl.
  reflexivity.

  eapply addfence_closed; eauto; done.

  exists I2'.
  split; auto; [].
  apply xpof_fence; auto; [].
  so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
  apply fcof_mono; auto; done.
Qed.


(* Execution State preservation *)

Lemma exstep_impl_indom :
  forall U P ex ex' d p,
    exstep U P ex ex' d p
    -> indom p ex.
Proof.
intros U P ex ex' d p Hstep.
induction Hstep; auto using indom_hit, indom_miss.
Qed.


Lemma exstep_preserves_domain :
  forall U P ex ex' d p,
    exstep U P ex ex' d p
    -> forall p', indom p' ex <-> indom p' ex'.
Proof.
intros U P ex ex' d p Hstep.
induct Hstep.

(* hit *)
intros U P ex p e e' d H p'.
split;
intro Hin;
(invert Hin; [intros; subst; apply indom_hit | intros; apply indom_miss; assumption]).

(* miss *)
intros U P ex ex' p e d p' H IH p''.  
split;
intro Hin;
(invert Hin; [intros; subst; apply indom_hit | intros; apply indom_miss; apply IH; assumption]).
Qed.


Lemma preservation_ex :
  forall U P I U' P' I' ex ex' d p,
    ddistinct P
    -> ddistinct I
    -> exof U P I ex
    -> closed d
    -> evolve U P I d p U' P' I'
    -> exstep U P ex ex' d p
    -> exof U' P' I' ex'.
Proof.
intros U P I U' P' I' ex ex' d p HdistP HdistI Hofex Hclosed Hevolve Hstep.
revert I I' HdistP HdistI Hofex Hclosed Hevolve.
induct Hstep.

(* hit *)
intros U P ex p e e' d Hstep I I' HdistP HdistI Hof Hclosed Hevolve.
invert Hof; intros t Hnindom Hofex Hofe.
apply (@exof_cons _#6 t); eauto; [|].
  erewrite <- evolve_restrictneq_eq; eauto; [].
  so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
  apply exof_mono; auto; done.

  eexploit (preservation_xp U U' P P' nil (restricteq I p) nil (restricteq I' p) nil e t e' d p) as (I2' & Heq & Hofe'); eauto.
    rewrite -> app_nil_l, app_nil_r.
    apply restricteq_ddistinct; auto; done.

    rewrite -> app_nil_l, app_nil_r.
    apply restricteq_evolve; auto; done.

    rewrite -> app_nil_l, app_nil_r in Heq.
    subst I2'.
    assumption.

(* miss *)
intros U P ex ex' p e d p' Hstep IH I I' HdistP HdistI Hof Hclosed Hevolve.
invert Hof; intros t Hnindom Hofex Hofe.
eapply exof_cons; eauto; [| |].
  contradict Hnindom.
  eapply exstep_preserves_domain; eauto; done.

  apply (IH (restrictneq I p) (restrictneq I' p)); auto.
    apply restrictneq_ddistinct; auto; done.

    apply restrictneq_evolve; auto; [].
    intro; subst p'.
    destruct Hnindom.
    eapply exstep_indom; eauto; done.

  so (evolve_extends _#8 Hevolve) as ((U'' & ->) & (P'' & ->)).
  apply xpof_mono; [].
  erewrite <- evolve_restricteq_eq; eauto; [].
  intros ->.
  destruct Hnindom.
  eapply exstep_impl_indom; eauto; done.
Qed.


(* Trace coherence preservation *)

Lemma preservation_trco_ninit :
  forall H d p H',
    trans H d p H'
    -> ~ (exists fl i a, d = tr_init fl i a)
    -> trco H
    -> trco H'.
Proof.
intros H d p H' Htrans Hninit Htrco.
revert Hninit Htrco.
induct Htrans.

(* nothing *)
auto; done.

(* init *)
intros H fl i a p _ Hninit _.
destruct Hninit.
repeat eexists; reflexivity.

(* read *)
intros H ir l p iw v Hinit His Hexecutable Hwrite Hexec Hacyclic _ Htrco.
destruct Hexecutable as (Hnexec & Hpriors).
#2 apply (@trco_rf _#3 l); eauto.
  exists v; assumption.

  apply reads_read; auto; done.

(* write *)
intros H i l v p Hinit His Hexecutable Hacyclic _ Htrco.
destruct Hexecutable as (Hnexec & Hdep).
#2 eapply trco_exec; eauto.
  intro Heq; destruct_all Heq; discriminate Heq.

  intro Heq; destruct_all Heq; discriminate Heq.

(* rw *)
intros H ir vw l p iw v Hinit His Hexecutable Hwrite Hexec Hacyclic _ Htrco.
destruct Hexecutable as (Hnexec & Hdep).
#2 eapply trco_rf; eauto.
  exists v; eassumption.

  eapply reads_rw; eauto; done.

(* push *)
intros H i p Hinit His Hexecutable Hacyclic _ Htrco.
destruct Hexecutable as (Hnexec & Hdep).
#2 eapply trco_exec; eauto.
  intro Heq; destruct_all Heq; discriminate Heq.

  intro Heq; destruct_all Heq; discriminate Heq.

(* nop *)
intros H i p Hinit His Hexecutable Hacyclic _ Htrco.
destruct Hexecutable as (Hnexec & Hdep).
#2 eapply trco_exec; eauto.
  intro Heq; destruct_all Heq; discriminate Heq.

  intro Heq; destruct_all Heq; discriminate Heq.

(* new *)
intros H b T T' p Hndecl Hndecl' Hneq _ Htrco.
apply trco_edge; auto; done.
Qed.


Lemma acof_isac :
  forall U P G a t,
    acof U P G a t
    -> isac a.
Proof.
intros U P G a t Hacof.
invert Hacof; intros; subst; auto with isac.
Qed.


Lemma preservation_trco :
  forall H d p H' U P I,
    trans H d p H'
    -> tgtp_sat H U
    -> idtp_sat U P H I
    -> trofo U P (restricteq I p) nil d
    -> trco H
    -> trco H'.
Proof.
intros H d p H' U P I Htrans HsatU HsatI Htrofo Htrco.
so (match_tr_init_decidable d) as [(fl & i & a & ->) | Hninit].
  invert Htrans; clear Htrans.
  intros Hndecl <-.
  invert Htrofo; clear Htrofo.
  intros t Hoffl Hofa.
  remember (@nil term) as G.
  revert HeqG.
  induction Hoffl as [| U G f fl Hoffl IH Hoff].
    (* nil *)
    intros ->.
    simpl.
    assert (trco (H; ev_init i p)) as Htrcoinit by (apply trco_init; auto).
    #3 apply trco_is; auto.
      eapply acof_isac; eauto; done.

      eapply acof_context_closed; eauto; done.

      exists p, H.
      reflexivity.
  
    (* cons *)
    intro; subst G.
    simpl.
    apply trco_fence; [| |].
      apply IH; auto; done.
  
      invert Hoffl; [|].
        intros <-.
        left.
        simpl.
        exists a, (H; ev_init i p).
        reflexivity.
  
        intros f' fl' _ _ <-.
        right.
        simpl.
        exists f'.
        eexists.
        reflexivity.
  
      apply fence_declared_app.
      apply fence_declared_cons.
      apply fence_declared_cons.
      eapply fcof_declared; eauto; done.

  eapply preservation_trco_ninit; eauto; done.
Qed.


Lemma preservation_acyclic :
  forall H d p H',
    trans H d p H'
    -> trco H'
    -> acyclic (co H)
    -> acyclic (co H').
Proof.
intros H d p H' Htrans Htrco Hcoacy.
revert Htrco Hcoacy.
#2 induct Htrans; auto.

(* init *)
intros H fl i a p _ Htrco Hcoacy.
remember (fc_list fl) as L eqn:HeqL.
clear HeqL.
revert Htrco.
induction L as [| f L IH].
  simpl.
  intro Htrco.
  apply is_acyclic; auto; [].
  apply init_acyclic; auto; [].
  eapply trco_tail; eauto; done.

  simpl.
  intro Htrco.
  apply fence_acyclic; auto; [].
  apply IH; [].
  eapply trco_tail; eauto; done.

(* edge *)
intros H b T T' p _ _ _ Htrco Hcoacy.
apply edge_acyclic; auto; done.
Qed.


(* Store preservation wrt ambient types *)

Lemma preservation_lctpof :
  forall U U' P P' I I' d p,
    lctpof P
    -> evolve U P I d p U' P' I'
    -> lctpof P'.
Proof.
intros U U' P P' I I' d p HofP Hevolve.
revert HofP.
elim Hevolve; auto.
Qed.


Local Ltac prove_spec :=
  repeat2 split;
  [ rewrite <- ?app_assoc;
    reflexivity
  | intros;
    (match goal with
     | |- ~ _ => intro
     | |- _ => idtac
     end);
    last_hyp ltac:(fun H => discriminate_in H);
    intros;
    subst;
    repeat eexists;
    reflexivity ..].


Lemma trans_extends_spec :
  forall H1 d p H2,
    trans H1 d p H2
    -> exists H,
         H2 = H1;; H
         /\ (forall i p, H {{ ev_init i p }} -> exists fl a, d = tr_init fl i a)
         /\ (forall i, H {{ ev_exec i }} -> exists v, d = tr_exec i v)
         /\ (forall i' i, H {{ ev_rf i' i }} -> exists v, d = tr_exec i v)
         /\ (forall i a, H {{ ev_is i a }} -> exists fl, d = tr_init fl i a)
         /\ (forall b T T', H {{ ev_edge b T T' }} -> d = tr_new b T T').
Proof.
intros H1 d p H2 Htrans.
invert Htrans.

(* nothing *)
intros <- <-.
exists nil.
prove_spec.

(* init *)
intros fl i a _ <- <-.
exists (nil; ev_init i p; ev_is i a;; map (fun f : term => ev_fence f i) (fc_list fl)).
prove_spec.

(* read *)
intros ir l iw v _ _ _ _ _ _ <- <-.
exists (nil; ev_rf iw ir).
prove_spec.

(* write *)
intros i l v _ _ _ _ <- <-.
exists (nil; ev_exec i).
prove_spec.

(* rw *)
intros ir vw l iw v _ _ _ _ _ _ <- <-.
exists (nil; ev_rf iw ir).
prove_spec.

(* push *)
intros i _ _ _ _ <- <-.
exists (nil; ev_exec i).
prove_spec.

(* nop *)
intros i _ _ _ _ <- <-.
exists (nil; ev_exec i).
prove_spec.

(* new *)
intros b T T' _ _ _ <- <-.
exists (nil; ev_edge b T T').
prove_spec.
Qed.


Lemma trans_extends :
  forall H1 d p H2,
    trans H1 d p H2
    -> exists H, H2 = H1;; H.
Proof.
intros H1 d p H2 Htrans.
so (trans_extends_spec _#4 Htrans) as (H & Heq & _).
exists H.
exact Heq.
Qed.


Lemma evolve_tgtp :
  forall U P I U' P' I' d p,
    evolve U P I d p U' P' I'
    -> U = U' \/ exists b T T' , d = tr_new b T T'.
Proof.
intros U P I U' P' I' d p Hevolve.
induct Hevolve; intros;
try (left; reflexivity).
right; repeat eexists; reflexivity.
Qed.


Lemma preservation_tgtp_sat :
  forall H U P I H' U' P' I' d p,
    trans H d p H'
    -> evolve U P I d p U' P' I'
    -> tgtp_sat H U
    -> tgtp_sat H' U'.
Proof.
intros H1 U1 P1 I1 H2 U2 P2 I2 d p Htrans Hevolve Hsat.
so (match_tr_new_decidable d) as [(b & T & T' & Heq) | Hnmatch].
  (* new *)
  have (d = tr_new b T T') as Heq.
  subst d.
  invert Hevolve; intros HninT HninT' HneqT ? ? ?.
  subst U2 P2 I2.
  invert Htrans; intros _ _ _ ?.
  subst H2.
  toshow (tgtp_sat (H1; ev_edge b T T') (U1; T; T')).
  apply tgtp_sat_bind; auto; done.

  (* not new *)
  have (~ (exists (b : attr) (T T' : tag), d = tr_new b T T')) as Hnmatch.
  so (evolve_tgtp _#8 Hevolve) as [HeqU |]; [| findcontra].
  have (U1 = U2) as HeqU.
  subst U2.
  so (trans_extends_spec _#4 Htrans) as (H & HeqH2 & _ & _ & _ & _ & Hnedge).
  subst H2.
  apply tgtp_sat_app; auto; [].
  intros (b & T & T' & Hedge).
  destruct Hnmatch.
  repeat eexists; apply Hnedge; eauto; done.
Qed.


Lemma evolve_lctp :
  forall U P I U' P' I' d p,
    evolve U P I d p U' P' I'
    -> P = P'.
Proof.
intros U P I U' P' I' d p Hevolve.
induct Hevolve; intros; reflexivity.
Qed.


Lemma preservation_lctp_sat :
  forall H U P I H' U' P' I' d p,
    trco H'
    -> trans H d p H'
    -> evolve U P I d p U' P' I'
    -> lctp_sat U P H P
    -> lctp_sat U' P' H' P'.
Proof.
intros H1 U1 P1 I1 H2 U2 P2 I2 d p Htrco Htrans Hevolve Hsat.
so (match_tr_init_decidable d) as [(fl & i & a & Heq) | Hnmatch].
  (* init *)
  have (d = tr_init fl i a) as Heq.
  subst d.
  invert Hevolve; intros t Hnindom _ Hofa ? ? ?.
  subst U2 P2 I2.
  invert Htrans; intros _ HeqH2; rewrite <- HeqH2.
  toshow (lctp_sat U1 P1
            (H1; ev_init i p; ev_is i a;;
             map (fun f : term => ev_fence f i) (fc_list fl)) P1).
  apply lctp_sat_app; [prove_trco | | |].
    toshow (lctp_sat U1 P1 (H1; ev_init i p; ev_is i a) P1).
    assert (lctp_sat U1 P1 (H1; ev_init i p) P1).
      apply lctp_sat_cons; auto; [prove_trco | |].
        intros; discriminate.

        intros; discriminate.
    so (match_ac_write_decidable a) as [(l & v & Heq) | Hnmatch].
      (* write *)
      subst a.
      invert Hofa; intros t' Hlookup Hvalue Hofv _.
      simp_sub in Hofv.
      eapply lctp_sat_write; eauto; prove_trco.

      (* not write *)
      so (match_ac_rw_decidable a) as [(l & v & Heq) | Hnmatch'].
        (* rw *)
        subst a.
        invert Hofa; intros t' Hlookup Hvalue Hofv _.
        simp_sub in Hofv.
        eapply lctp_sat_write; eauto; prove_trco.

        (* neither write nor rw *)
        apply lctp_sat_cons; auto; [prove_trco | |].
          intros; injection; intros; subst a.
          destruct Hnmatch; repeat eexists; reflexivity.

          intros; injection; intros; subst a.
          destruct Hnmatch'; repeat eexists; reflexivity.

    intro Hin.
    discriminate_in Hin; done.

    intro Hin.
    discriminate_in Hin; done.

  (* not init *)
  so (trans_extends_spec _#4 Htrans) as (H & HeqH2 & _ & _ & _ & Hnois & _).
  subst H2.
  assert (~ H {{ ev_is ie ae | ie ae }}) as Hnois'.
    intro Hin; destruct_all Hin.
    destruct Hnmatch.
    eexploit Hnois as Hmatch; eauto; [].
    destruct_all Hmatch.
    subst d.
    repeat eexists; reflexivity.
  apply lctp_sat_app; [prove_trco | | |].
    so (evolve_extends _#8 Hevolve) as ((U & HeqU2) & _).
    thus (P1 = P2) by evolve_lctp.
    subst U2 P2.
    repl P1 with (P1;; nil) by reflexivity.
    apply lctp_sat_mono; auto; done.

    contradict Hnois'; destruct_all Hnois'; eauto; done.

    contradict Hnois'; destruct_all Hnois'; eauto; done.
Qed.


Lemma evolve_idtp :
  forall U U' P P' I I' d p,
    evolve U P I d p U' P' I'
    -> I = I'
       \/ (exists fl i a, d = tr_init fl i a)
       \/ (exists i v, d = tr_exec i v).
Proof.
intros U U' P P' I I' d p Hevolve.
induct Hevolve; intros;
try (left; reflexivity).
  right; left; repeat eexists; reflexivity.

  right; right; repeat eexists; reflexivity.
Qed.


Lemma trans_exec_form :
  forall H1 i v p H2,
    trans H1 (tr_exec i v) p H2
    -> exists h, H2 = H1; h /\ executes h i.
Proof.
intros H1 i v p H2 Htrans.
invert Htrans.

(* read *)
intros _ iw _ _ _ _ _ _ <-.
exists (ev_rf iw i).
split; auto; [].
right; eauto; done.

(* write *)
intros l _ _ _ _ _ <-.
exists (ev_exec i).
split; auto; done.

(* rw *)
intros vw l iw _ _ _ _ _ _ <-.
exists (ev_rf iw i).
split; auto; [].
right; eauto; done.

(* push *)
intros _ _ _ _ _ <-.
exists (ev_exec i).
split; auto; done.

(* nop *)
intros _ _ _ _ <-.
exists (ev_exec i).
split; auto; done.
Qed.


Lemma preservation_idtp_sat :
  forall H U P I H' U' P' I' d p,
    trco H
    -> trans H d p H'
    -> evolve U P I d p U' P' I'
    -> idtp_sat U P H I
    -> idtp_sat U' P' H' I'.
Proof.
intros H1 U1 P1 I1 H2 U2 P2 I2 d p Htrco Htrans Hevolve Hsat.
so (match_tr_init_decidable d) as [(fl & i & a & Heq) | Hnmatchinit].
  (* init *)
  have (d = tr_init fl i a) as Heq.
  subst d.
  invert Hevolve; intros t Hnindom _ Hofa ? ? ?.
  subst U2 P2 I2.
  invert Htrans; intros Hninit HeqH2; rewrite <- HeqH2. 
  apply idtp_sat_app; [| | |].
    apply idtp_sat_bind; auto; done.

    intro Hin; discriminate_in Hin; done.

    intro Hin; discriminate_in Hin; done.

    intro Hin; discriminate_in Hin; done.

  (* not init *)
  so (match_tr_exec_decidable d) as [(i & v & Heq) | Hnmatchexec].
    have (d = tr_exec i v) as Heq.
    subst d.
    invert Hevolve; intros I1a I1b ? ? <- <- <- <-.
    so (trans_exec_form _#5 Htrans) as (h & Heq & Hexecutes).
    subst H2.
    eapply idtp_sat_exec; eauto; done.

    so (trans_extends_spec _#4 Htrans) as (H & HeqH2 & Hnoinit & Hnoexec & Hnorf & _).
    subst H2.
    apply idtp_sat_app; [| | |].
      so (evolve_extends _#8 Hevolve) as ((U & HeqU2) & (P & HeqP2)).
      so (evolve_idtp _#8 Hevolve) as [HeqI2 |[ Heqinit | Heqexec]].
        subst U2 P2 I2.
        apply idtp_sat_mono; auto; done.
  
        destruct_all Heqinit.
        subst d.
        destruct Hnmatchinit; repeat eexists; reflexivity.

        destruct_all Heqexec.
        subst d.
        destruct Hnmatchexec; repeat eexists; reflexivity.

      intro Hinit.
      destruct_all Hinit.
      eexploit Hnoinit as Heqinit; eauto; [].
      destruct_all Heqinit.
      subst d.
      destruct Hnmatchinit; repeat eexists; reflexivity.

      intro Hexec.
      destruct_all Hexec.
      eexploit Hnoexec; eauto; done.

      intro Hrf.
      destruct_all Hrf.
      eexploit Hnorf; eauto; done.
Qed.


(* Preservation *)

Lemma preservation_hok :
  forall H H' U U' P P' I I' d p,
    trans H d p H'
    -> evolve U P I d p U' P' I'
    -> trofo U P (restricteq I p) nil d
    -> hok H U P I
    -> hok H' U' P' I'.
Proof.
intros H H' U U' P P' I I' d p Htrans Hevolve Htrofo Hok.
so Hok as (HsatU & HsatP & HsatI & Htrco & Hcoacy).
thus (trco H') as Htrco' by preservation_trco.
do2 4 split.
  eapply preservation_tgtp_sat; eauto; done.

  thus (ddistinct P) as HdistP by lctp_sat_ddistinct.
  eapply preservation_lctp_sat; eauto; done.
  
  clear Htrco'.
  eapply preservation_idtp_sat; eauto; done.
  
  assumption.

  eapply preservation_acyclic; eauto; done.
Qed.


Lemma evolve_closed :
  forall U P I d p U' P' I',
    evolve U P I d p U' P' I'
    -> closed d.
Proof.
intros U P I d p U' P' I' Hevolve.
destruct Hevolve; reduce_closed; eauto using tmof_context_closed, acof_context_closed; [].
eapply flof_context_closed; eauto; done.
Qed.


Lemma preservation_sync :
  forall U P I U' P' I' H H' ex ex' d p,
    exstep U P ex ex' d p
    -> trans H d p H'
    -> evolve U P I d p U' P' I'
    -> ok (U, P, I, H, ex)
    -> ok (U', P', I', H', ex').
Proof.
intros U P I U' P' I' H H' ex ex' d p Hstep Htrans Hevolve Hok.
invert Hok; intros Hhok Hofex; clear Hok.
have (hok H U P I) as Hhok.
have (exof U P I ex) as Hofex.
thus (trofo U P (restricteq I p) nil d) as Htrofo by exstep_trofo.
thus (hok H' U' P' I') as Hhok' by preservation_hok.
assert (lctp_sat U' P' H' P') as HsatP' by (destruct Hhok' as (? & ? & ? & ?); auto).
assert (idtp_sat U' P' H' I') as HsatI' by (destruct Hhok' as (? & ? & ? & ?); auto).
apply okay; [|].
  toshow (hok H' U' P' I').
  eapply preservation_hok; eauto; done.

  eapply preservation_ex; eauto; [| |].
    eapply lctp_sat_ddistinct; eauto; [].
    eexact (Hhok ander andel).

    eapply idtp_sat_ddistinct; eauto; [].
    eexact (Hhok ander ander andel).

    eapply evolve_closed; eauto; done.
Qed.


Theorem preservation :
  forall S S',
    step S S'
    -> ok S
    -> ok S'.
Proof.
intros S S' Hstep Hok.
destruct S as ((A & H) & ex).
destruct S' as ((A' & H') & ex').
revert Hok.
induct Hstep.

(* sync *)
intros.
eapply preservation_sync; eauto; done.
Qed.
