
Require Import Tactics.
Require Import Static.
Require Import Dynamic.
Require Import Subst.
Require Import Hwf.
Require Import Ambients.
Require Import Match.
Require Import Decide.
Require Import Relation.
Require Import Sequence.
Require Import History.
Require Import Truncate.
Require Import Acyclic.
Require Import Preservation.
Require Import TSubst.
Require Import Unique.
Require Wellfounded.Inclusion.


(* Rightness *)

Inductive right_tm : context -> term -> Prop :=
| right_tm_var {G i t}
    : index i G (cl_vl t)
      -> right_tm G (var i)

| right_tm_unit {G}
    : right_tm G tm_unit

| right_tm_lit {G c}
    : right_tm G (tm_lit c)

| right_tm_prim {G m1 m2}
    : right_tm G m1
      -> right_tm G m2
      -> (value nil m1 -> exists c, m1 = tm_lit c)
      -> (value nil m2 -> exists c, m2 = tm_lit c)
      -> right_tm G (tm_prim m1 m2)

| right_tm_if {G m m1 m2}
    : right_tm G m
      -> right_tm G m1
      -> right_tm G m2
      -> (value nil m -> exists c, m = tm_lit c)
      -> right_tm G (tm_if m m1 m2)

| right_tm_loc {G l}
    : right_tm G (tm_loc l)

| right_tm_susp {G e}
    : closedn (length G) e
      -> right_tm G (tm_susp e)

| right_tm_lam {G t m}
    : closedn (length G) t
      -> closedn (S (length G)) m
      -> right_tm G (tm_lam t m)

| right_tm_fun {G t1 t2 m}
    : closedn (length G) t1
      -> closedn (length G) t2
      -> closedn (2 + length G) m
      -> right_tm G (tm_fun t1 t2 m)

| right_tm_app {G m1 m2}
    : right_tm G m1
      -> right_tm G m2
      -> (value nil m1
          -> (exists t m1', m1 = tm_lam t m1')
             \/ (exists t1 t2 m1', m1 = tm_fun t1 t2 m1'))
      -> right_tm G (tm_app m1 m2)
.


Inductive right_tg : history -> context -> term -> Prop :=
| right_tg_var {H G i}
    : index i G cl_tg
      -> right_tg H G (var i)

| right_tg_lit {H G T}
    : tag_declared H T
      -> right_tg H G (tg_lit T)
.


Inductive right_fc : history -> context -> term -> Prop :=
| right_fc_tag {H G T}
    : right_tg H G T
      -> right_fc H G (fc_tag T)

| right_fc_pre {H G b}
    : right_fc H G (fc_pre b)

| right_fc_post {H G b}
    : right_fc H G (fc_post b)
.


Inductive right_ident : history -> identifier -> Prop :=
| right_ident_read {H i l iw}
    : H {{ ev_is i (ac_read l) }}
      -> wrp H iw i
      -> right_ident H i

| right_ident_write {H i l v}
    : H {{ ev_is i (ac_write l v) }}
      -> right_ident H i

| right_ident_rw {H i l v iw}
    : H {{ ev_is i (ac_rw l v) }}
      -> wrp H iw i
      -> right_ident H i

| right_ident_push {H i}
    : H {{ ev_is i ac_push }}
    -> right_ident H i

| right_ident_nop {H i}
    : H {{ ev_is i ac_nop }}
    -> right_ident H i
.


Inductive right_xp : history -> thread -> list identifier -> tgtp -> lctp -> context -> (* xp *) term -> Prop :=
| right_xp_ret {H p L U P G m}
    : right_tm G m
      -> right_xp H p L U P G (xp_ret m)

| right_xp_force {H p L U P G m}
    : right_tm G m
      -> (value nil m -> exists e, m = tm_susp e)
      -> right_xp H p L U P G (xp_force m)

| right_xp_bind {H p L U P G t e1 e2}
    : closedn (length G) t
      -> right_xp H p L U P G e1
      -> right_xp H p (L;; xp_actions e1) U P (G; cl_vl t) e2
      -> (xp_actions e2 <> nil -> init e1)
      -> right_xp H p L U P G (xp_bind t e1 e2)

| right_xp_spec {H p L U P G m1 m2 e}
    : closedn (length G) m1
      -> closedn (length G) m2
      -> value G m1
      -> value G m2
      -> right_xp H p L U P G e
      -> right_xp H p L U P G (xp_spec m1 m2 e)

| right_xp_read {H p L U P G m}
    : right_tm G m
      -> (value nil m -> exists l, m = tm_loc l)
      -> right_xp H p L U P G (xp_read m)

| right_xp_write {H p L U P G m1 m2}
    : right_tm G m1
      -> right_tm G m2
      -> (value nil m1 -> exists l, m1 = tm_loc l)
      -> right_xp H p L U P G (xp_write m1 m2)

| right_xp_rw {H p L U P G m1 m2}
    : right_tm G m1
      -> right_tm G m2
      -> (value nil m1 -> exists l, m1 = tm_loc l)
      -> right_xp H p L U P G (xp_rw m1 m2)

| right_xp_rmw {H p L U P G t m1 m2}
    : closedn (length G) t
      -> right_tm G m1
      -> right_tm (G; cl_vl t) m2
      -> (value nil m1 -> exists v, tmof U P G v t /\ value G v)
      -> right_xp H p L U P G (xp_rmw t m1 m2)

| right_xp_push {H p L U P G}
    : right_xp H p L U P G xp_push

| right_xp_nop {H p L U P G}
    : right_xp H p L U P G xp_nop

| right_xp_action {H p L U P G i}
    : H {{ ev_init i p }}
      -> ~ exec H i
      -> (forall i', (po H i' i /\ ~ exec H i') <-> In i' L)
      -> right_ident H i
      -> right_xp H p L U P G (xp_action i)

| right_xp_new {H p L U P G b e}
    : right_xp H p L U P (G; cl_tg; cl_tg) e
      -> right_xp H p L U P G (xp_new b e)

| right_xp_fence {H p L U P G f e}
    : right_fc H G f
      -> right_xp H p L U P G e
      -> right_xp H p L U P G (xp_fence f e)
.


Inductive right_ex : history -> tgtp -> lctp -> xstate -> Prop :=
| right_ex_nil {H U P}
    : right_ex H U P nil

| right_ex_cons {H U P ex p e}
    : right_ex H U P ex
      -> right_xp H p nil U P nil e
      -> (forall i, H {{ ev_init i p }} -> ~ exec H i -> In i (xp_actions e))
      -> right_ex H U P (ex; (p, e)).


Inductive right : state -> Prop :=
| right_i {U P I H ex}
    : trco H
      -> right_ex H U P ex
      -> (forall i p, H {{ ev_init i p }} -> ~ exec H i -> indom p ex)
      -> right (U, P, I, H, ex)
.


Hint Constructors right_tm right_tg right_fc right_ident right_xp right_ex right : progress.


(* Stronger dynamic semantics *)

(* Stronger form of xstep: no speculation. *)
Inductive xstepp : tgtp -> lctp -> context -> (* xp *) term -> (* xp *) term -> (* tr *) term -> Prop :=
| xstepp_ret {U P G m m'}
    : tstep G m m'
      -> xstepp U P G (xp_ret m) (xp_ret m') tr_nothing

| xstepp_force1 {U P G m m'}
    : tstep G m m'
      -> xstepp U P G (xp_force m) (xp_force m') tr_nothing

| xstepp_force2 {U P G e}
    : xstepp U P G (xp_force (tm_susp e)) e tr_nothing

| xstepp_read1 {U P G m m'}
    : tstep G m m'
      -> xstepp U P G (xp_read m) (xp_read m') tr_nothing

| xstepp_read2 {U P G l i}
    : xstepp U P G (xp_read (tm_loc l)) (xp_action i) (tr_init fl_nil i (ac_read l))

| xstepp_write1 {U P G m1 m2 m1'}
    : tstep G m1 m1'
      -> xstepp U P G (xp_write m1 m2) (xp_write m1' m2) tr_nothing

| xstepp_write2 {U P G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> xstepp U P G (xp_write v1 m2) (xp_write v1 m2') tr_nothing

| xstepp_write3 {U P G l v i}
    : value G v
      -> xstepp U P G (xp_write (tm_loc l) v) (xp_action i) (tr_init fl_nil i (ac_write l v))

| xstepp_rw1 {U P G m1 m2 m1'}
    : tstep G m1 m1'
      -> xstepp U P G (xp_rw m1 m2) (xp_rw m1' m2) tr_nothing

| xstepp_rw2 {U P G v1 m2 m2'}
    : value G v1
      -> tstep G m2 m2'
      -> xstepp U P G (xp_rw v1 m2) (xp_rw v1 m2') tr_nothing

| xstepp_rw3 {U P G l v i}
    : value G v
      -> xstepp U P G (xp_rw (tm_loc l) v) (xp_action i) (tr_init fl_nil i (ac_rw l v))

| xstepp_rmw1 {U P G t m1 m2 m1'}
    : tstep G m1 m1'
      -> xstepp U P G (xp_rmw t m1 m2) (xp_rmw t m1' m2) tr_nothing

| xstepp_rmw2 {U P G t v1 m2 v}
    : value G v1
      -> tmof U P G v t
      -> value G v
      -> xstepp U P G (xp_rmw t v1 m2) (xp_bind t (xp_rw v1 (subst1 v m2)) (xp_spec (var 0) (shift1 v) (xp_ret (shift1 v)))) tr_nothing

| xstepp_push {U P G i}
    : xstepp U P G xp_push (xp_action i) (tr_init fl_nil i ac_push)

| xstepp_nop {U P G i}
    : xstepp U P G xp_nop (xp_action i) (tr_init fl_nil i ac_nop)

| xstepp_exec {U P G i v}
    : xstepp U P G (xp_action i) (xp_ret v) (tr_exec i v)

| xstepp_bind1 {U P G t e1 e1' e2 d}
    : xstepp U P G e1 e1' d
      -> xstepp U P G (xp_bind t e1 e2) (xp_bind t e1' e2) d

| xstepp_bind2 {U P G t e1 e2 e2' d}
    : init e1
      -> xstepp U P (G; cl_vl t) e2 e2' (shift1 d)
      -> xstepp U P G (xp_bind t e1 e2) (xp_bind t e1 e2') d

| xstepp_bind3 {U P G t v e}
    : value G v
      -> xstepp U P G (xp_bind t (xp_ret v) e) (subst1 v e) tr_nothing

| xstepp_spec2 {U P G e e' d v1 v2}
    : xstepp U P G e e' d
      -> xstepp U P G (xp_spec v1 v2 e) (xp_spec v1 v2 e') d

| xstepp_spec3 {U P G v e}
    : xstepp U P G (xp_spec v v e) e tr_nothing

| xstepp_new {U P G e b T1 T2}
    : xstepp U P G (xp_new b e) (subst (dot (tg_lit T2) (dot (tg_lit T1) id)) e) (tr_new b T1 T2)

| xstepp_fence1 {U P G f e e' d}
    : xstepp U P G e e' d
      -> xstepp U P G (xp_fence f e) (xp_fence f e') (addfence f d)

| xstepp_fence2 {U P G f e}
    : init e
      -> xstepp U P G (xp_fence f e) e tr_nothing
.


Inductive exstepp : tgtp -> lctp -> xstate -> xstate -> (* tr *) term -> thread -> Prop :=
| exstepp_hit {U P ex p e e' d}
    : xstepp U P nil e e' d
      -> exstepp U P (ex; (p, e)) (ex; (p, e')) d p

| exstepp_miss {U P ex ex' p e d p'}
    : exstepp U P ex ex' d p'
      -> exstepp U P (ex; (p, e)) (ex'; (p, e)) d p'
.


Inductive stepp : state -> state -> Prop :=
| stepp_sync {U P I U' P' I' H H' ex ex' d p}
    : exstepp U P ex ex' d p
      -> trans H d p H'
      -> evolve U P I d p U' P' I'
      -> stepp (U, P, I, H, ex) (U', P', I', H', ex')
.


Hint Constructors xstepp exstepp : progress.


(* Progress defns *)

Inductive within : (* xp *) term -> (* xp *) term -> Prop :=
| within_refl {e}
    : within e e

| within_bind1 {t e e1 e2}
    : within e e1
      -> within e (xp_bind t e1 e2)

| within_bind2 {t e e1 e2}
    : within e e2
      -> within e (xp_bind t e1 e2)

| within_spec {e v1 v2 e'}
    : within e e'
      -> within e (xp_spec v1 v2 e')

| within_fence {f e e'}
    : within e e'
      -> within e (xp_fence f e')
.


Inductive basic : (* xp *) term -> Prop :=
| basic_ret {m}
    : basic (xp_ret m)

| basic_force {m}
    : basic (xp_force m)

| basic_read {m}
    : basic (xp_read m)

| basic_write {m1 m2}
    : basic (xp_write m1 m2)

| basic_rw {m1 m2}
    : basic (xp_rw m1 m2)

| basic_rmw {t m1 m2}
    : basic (xp_rmw t m1 m2)

| basic_push
    : basic xp_push

| basic_nop
    : basic xp_nop

| basic_action {i}
    : basic (xp_action i)
.


Definition blocked H i :=
  ~ (exists H' v p, trans H (tr_exec i v) p H').


Inductive inactive : history -> (* xp *) term -> Prop :=
| inactive_open {H e}
    : basic e
      -> ~ closed e
      -> inactive H e

| inactive_blocked {H i}
    : blocked H i
      -> inactive H (xp_action i)

| inactive_bind1 {H t e1 e2}
    : inactive H e1
      -> inactive H e2
      -> inactive H (xp_bind t e1 e2)

| inactive_bind2 {H t e1 v}
    : inactive H e1
      -> value nil v
      -> inactive H (xp_bind t e1 (xp_ret v))

| inactive_bind3 {H t e1 e2}
    : inactive H e1
      -> ~ init e1
      -> inactive H (xp_bind t e1 e2)

| inactive_spec1 {H v1 v2 e}
    : v1 <> v2  (* either moot, or the values aren't closed yet *)
      -> inactive H e
      -> inactive H (xp_spec v1 v2 e)

| inactive_spec2 {H v1 v2 v}
    : v1 <> v2  (* either moot, or the values aren't closed yet *)
      -> value nil v
      -> inactive H (xp_spec v1 v2 (xp_ret v))

| inactive_fence {H f e}
    : inactive H e
      -> inactive H (xp_fence f e)
.


Hint Constructors basic inactive : progress.


(* Canonical forms *)

Lemma canon_base :
  forall U P G m,
    tmof U P G m tp_base
    -> value nil m
    -> exists c, m = tm_lit c.
Proof.
intros U P G m Hof Hvalue.
invert Hof;
try (solve [
           intros; subst m; invert Hvalue
           |
           intros; subst; repeat eexists; reflexivity
           ]).

(* var *)
intros; subst m.
invert Hvalue; intros ? Hindex.
invert Hindex; done.
Qed.


Lemma canon_susp :
  forall U P G m t,
    tmof U P G m (tp_susp t)
    -> value nil m
    -> exists e, m = tm_susp e.
Proof.
intros U P G m t Hof Hvalue.
invert Hof;
try (solve [
           intros; subst m; invert Hvalue
           |
           intros; subst; repeat eexists; reflexivity
           ]).

(* var *)
intros; subst m.
invert Hvalue; intros ? Hindex.
invert Hindex; done.
Qed.


Lemma canon_ref :
  forall U P G m t,
    tmof U P G m (tp_ref t)
    -> value nil m
    -> exists l, m = tm_loc l.
Proof.
intros U P G m t Hof Hvalue.
invert Hof;
try (solve [
           intros; subst m; invert Hvalue
           |
           intros; subst; repeat eexists; reflexivity
           ]).

(* var *)
intros; subst m.
invert Hvalue; intros ? Hindex.
invert Hindex; done.
Qed.


Lemma canon_arrow :
  forall U P G m t1 t2,
    tmof U P G m (tp_arrow t1 t2)
    -> value nil m
    -> (exists t m', m = tm_lam t m')
       \/ (exists t1 t2 m', m = tm_fun t1 t2 m').
Proof.
intros U P G m t1 t2 Hof Hvalue.
invert Hof;
try (solve [
           intros; subst m; invert Hvalue
           |
           intros; subst; left; repeat eexists; reflexivity
           |
           intros; subst; right; repeat eexists; reflexivity
           ]); [].

(* var *)
intros; subst m.
invert Hvalue; intros ? Hindex.
invert Hindex; done.
Qed.


(* Term rightness/progress *)

Lemma tmof_right :
  forall U P G m t,
    tmof U P G m t
    -> right_tm G m.
Proof.
intros U P G m t Hof.
induct Hof; try (intros; eauto using xpof_context_closedn with progress; done).

(* prim *)
intros U P G m1 m2 Hof1 IH1 Hof2 IH2.
apply right_tm_prim; auto; [|];
intro; eapply canon_base; eauto; done.

(* if *)
intros U P G m m1 m2 t Hofm IH Hofm1 IH1 Hofm2 IH2.
apply right_tm_if; auto; [].
intro; eapply canon_base; eauto; done.

(* lam *)
intros U P G m t1 t2 H H0 H1.
apply right_tm_lam; eauto using tpof_context_closedn with progress; [].
repl (S (length G)) with (length (G; cl_vl t1)) by trivial.
eapply tmof_context_closedn; eauto; done.

(* fun *)
intros U P G m t1 t2 H H0 H1 H2.
apply right_tm_fun; eauto using tpof_context_closedn with progress; [].
repl (2 + length G) with (length (G; cl_vl (tp_arrow t1 t2); cl_vl (shift1 t1))) by trivial.
eapply tmof_context_closedn; eauto; done.

(* app *)
intros U P G m1 m2 t t' Hof1 IH1 Hof2 IH2.
apply right_tm_app; auto; [].
intro Hvalue1.
eapply canon_arrow; eauto; done.
Qed.


Lemma right_tm_progress :
  forall m,
    right_tm nil m
    -> value nil m \/ (exists m', tstep nil m m').
Proof.
intros m Hright.
remember nil as G eqn:HeqG in Hright.
revert HeqG; induct Hright;
try (intro G; intros; subst G; eauto with static; done).

(* prim *)
intros G m1 m2 Hr1 IH1 Hr2 IH2 Hcanon1 Hcanon2 ->.
right.
destruct (IH1 (eq_refl _)) as [Hval1 | (m1' & Hstep1)].
  destruct (IH2 (eq_refl _)) as [Hval2 | (m2' & Hstep2)].
    so (Hcanon1 Hval1) as (c1 & ->).
    so (Hcanon2 Hval2) as (c2 & ->).
    exists (tm_lit (prim c1 c2)).
    apply tstep_prim3; done.

    exists (tm_prim m1 m2').
    apply tstep_prim2; auto; done.

  exists (tm_prim m1' m2).
  apply tstep_prim1; auto; done.

(* if *)
intros G m m1 m2 Hr IH _ _ _ _ Hcanon ->.
right.
destruct (IH (eq_refl _)) as [Hval | (m' & Hstep)].
  so (Hcanon Hval) as (c & ->).
  so (eq_base_dec c base0) as [Heq | Hneq].
    subst c.
    exists m1.
    apply tstep_if2; done.

    exists m2.
    apply tstep_if3; auto; done.
    
  exists (tm_if m' m1 m2).
  apply tstep_if1; auto; done.

(* app *)
intros G m1 m2 Hr1 IH1 Hr2 IH2 Hcanon ->.
toshow (value nil (tm_app m1 m2) \/
        (exists m' : term, tstep nil (tm_app m1 m2) m')).
right.
destruct (IH1 (eq_refl _)) as [Hval1 | (m1' & Hstep1)].
  have (value nil m1) as Hval1.
  destruct (IH2 (eq_refl _)) as [Hval2 | (m2' & Hstep2)].
    have (value nil m2) as Hval2.
    so (Hcanon Hval1) as [(t & m1' & Heq) | (t1 & t2 & m1' & Heq)].
      have (m1 = tm_lam t m1') as Heq.
      rewrite -> Heq.
      toshow (exists m' : term, tstep nil (tm_app (tm_lam t m1') m2) m').
      exists (subst1 m2 m1').
      apply tstep_app3; auto; done.

      rewrite -> Heq.
      exists (subst (dot m2 (dot (tm_fun t1 t2 m1') id)) m1').
      apply tstep_app3'; auto; done.

    have (tstep nil m2 m2') as Hstep2.
    exists (tm_app m1 m2').
    apply tstep_app2; auto; done.

  have (tstep nil m1 m1') as Hstep1.
  exists (tm_app m1' m2).
  apply tstep_app1; auto; done.
Qed.


(* Tag rightness *)

Lemma tgof_right :
  forall H U G T,
    tgtp_sat H U
    -> tgof U G T
    -> right_tg H G T.
Proof.
intros H U G T HsatU Hof.
revert HsatU; induct Hof; try (intros; eauto with progress; done).

(* lit *)
intros U G T Hin HsatU.
destruct HsatU as (_ & Hindecl & _).
eauto with progress.
Qed.


(* Fence rightness *)

Lemma fcof_right :
  forall H U G f,
    tgtp_sat H U
    -> fcof U G f
    -> right_fc H G f.
Proof.
intros H U G f HsatU Hof.
revert HsatU; induct Hof; try (intros; eauto using tgof_right with progress; done).
Qed.


(* Action rightness *)

Lemma acof_right :
  forall H U P i a t,
    trco H
    -> tgtp_sat H U
    -> lctp_sat U P H P
    -> acof U P nil a t
    -> H {{ ev_is i a }}
    -> right_ident H i.
Proof.
intros H U P i a t Htrco HsatU HsatP Hof His.
remember nil as G eqn:HeqG in Hof.
revert HsatU HsatP HeqG His.
#2 induct Hof; try (intros; eauto with progress; done).

(* read *)
intros l U P G t Hlookup HsatU HsatP -> His.
thus (In (l, t) P) as Hin by lookup_iff_in.
destruct HsatP as (_ & Hallinit & _).
so (Hallinit _ _ Hin) as (iw & ip & (v & Hisw) & Hisp & Hvos & Hexecp & Hallto).
apply (@right_ident_read H i l iw); auto; [].
thus (reads H i l) as Hread by @reads_read.
assert (writesto H iw l) as Hwrite by (eexists; eauto).
eapply wrp_vt; eauto; [].
apply vop_vt; [].
eapply plus_trans; eauto; [].
apply plus_one; [].
right; right.
split; auto; done.

(* rw *)
intros U P G l v t Hlookup _ _ HsatU HsatP -> His.
thus (In (l, t) P) as Hin by lookup_iff_in.
destruct HsatP as (_ & Hallinit & _).
so (Hallinit _ _ Hin) as (iw & ip & (v' & Hisw) & Hisp & Hvos & Hexecp & Hallto).
apply (@right_ident_rw H i l v iw); auto; [].
thus (reads H i l) as Hread by @reads_rw.
assert (writesto H iw l) as Hwrite by (eexists; eauto).
eapply wrp_vt; eauto; [].
apply vop_vt; [].
eapply plus_trans; eauto; [].
apply plus_one; [].
right; right.
split; auto; done.
Qed.


(* Deciding executable actions *)

Lemma trco_executes_itemizable :
  forall H i,
    trco H
    -> itemizable (fun h => trco (H; h) /\ acyclic (co (H; h)) /\ executes h i).
Proof.
intros H i Htrco.
so (exec_decidable H i) as [Hexec | Hnexec].
  apply itemizable_none; [].
  intros (h & Htrco' & Hcoacy' & Hexecutes).
  so (trco_executes_invert _#3 Htrco' Hexecutes) as (_ & _ & Hnexec & _).
  contradiction.

  assert (decidable (forall i', xo H i' i -> exec H i')) as [Hreqs | Hnreqs].
    #2 apply itemizable_decidable_forall.
      apply xo_itemizable_left; auto; done.
      
      intros i' Hxo.
      apply exec_decidable; done.

    #2 apply (itemizable_from_univ _ _ (ev_exec i :: map (fun i' => ev_rf i' i) (ids H))).
      intros h Hin.
      assert (decidable (trco (H; h)) /\ executes h i) as ([Htrco' | Hntrco] & Hexecutes).
        destruct Hin as [Heq | Hin].
          subst h.
          split; [| left; reflexivity].
          so (present_1_decidable _ _ (match_is_i_decidable i) H) as [(a & His) | Hnis].
            so (match_ac_read_decidable a) as [(l & Heqread) | Hneqread].
              subst a.
              right.
              intros Htrco'.
              invert Htrco'.
              intros a' _ His' _ _ Hneqread _.
              destruct Hneqread.
              exists l.
              so (is_fun _#4 Htrco His His'); subst; reflexivity.
  
              so (match_ac_rw_decidable a) as [(l & v & Heqrw) | Hneqrw].
                subst a.
                right.
                intros Htrco'.
                invert Htrco'.
                intros a' _ His' _ _ _ Hneqrw.
                destruct Hneqrw.
                exists l, v.
                so (is_fun _#4 Htrco His His'); subst; reflexivity.
  
                left.
                eapply trco_exec; eauto; done.

            right; intro Htrco'; invert Htrco'.
            intros a _ His _ _ _ _.
            destruct Hnis; eauto; done.
                
          so (in_map_iff _#5 andel Hin) as (iw & Heq & Hinids).
          subst h; clear Hin.
          split; [| right; eauto; done].
          so (exec_decidable H iw) as [Hexecw | Hnexecw].
            so (reads_decidable_i H i) as [(l & Hreads) | Hnreads].
              so (writesto_decidable H iw l) as [Hwritesto | Hnwritesto].
                left.
                eapply trco_rf; eauto; done.

                right; intro Htrco'; invert Htrco'.
                intros l' _ Hwritesto Hreads' _ _ _.
                so (reads_fun _#4 Htrco Hreads Hreads'); subst l'.
                contradiction.

              right; intro Htrco'; invert Htrco'.
              intros l _ _ Hreads _ _ _.
              destruct Hnreads; eauto; done.

            right.
            intro Htrco'.
            invert Htrco'.
            intros _ _ _ _ Hexecw _ _.
            contradiction.

        so (acyclic_co_decidable _ Htrco') as [Hcoacy' | Hncoacy].
          left; auto; done.

          right.
          intros (_ & Hcoacy' & _).
          contradiction.

        right.
        intros (Htrco' & _).
        contradiction.

      intros h (Htrco' & _ & Hexecutes).
      destruct Hexecutes as [Heq | (i' & Heq)].
        subst h.
        left.
        reflexivity.

        subst.
        right.
        apply in_map_iff; [].
        exists i'.
        split; auto; [].
        apply init_ids; [].
        so (trco_rf_invert _#3 Htrco') as (l & Hwritesto & _).
        so (writesto_inited _#3 Htrco Hwritesto) as (p' & Hinit').
        eauto; done.

    apply itemizable_none; [].
    intros (h & Htrco' & Hcoacy' & Hexecutes).
    so (trco_executes_invert _#3 Htrco' Hexecutes) as (_ & _ & _ & Hreqs).
    contradiction.
Qed.


Lemma trco_executes_decidable :
  forall H i,
    trco H
    -> decidable (exists h, trco (H; h) /\ acyclic (co (H; h)) /\ executes h i).
Proof.
intros H i Htrco.
apply itemizable_decidable_exists; [].
apply trco_executes_itemizable; auto; done.
Qed.


Lemma trco_trans :
  forall H h i p,
    trco (H; h)
    -> acyclic (co (H; h))
    -> executes h i
    -> H {{ ev_init i p }}
    -> exists v, trans H (tr_exec i v) p (H; h).
Proof.
intros H h i p Htrco Hcoacy Hexecutes Hinit.
invert Hexecutes; clear Hexecutes.

(* exec *)
intros ->.
invert Htrco.
intros a Htrco' His Hnexec Hpriors Hneqread Hneqrw.
so (is_isac _#3 Htrco' His) as Hisac.
#3 invert Hisac; try (intros; subst; solve [destruct Hneqread; eauto | destruct Hneqrw; eauto]).
  (* write *)
  intros l v <-.
  exists tm_unit.
  eapply trans_write; eauto; [].
  split; auto; done.

  (* push *)
  intros <-.
  exists tm_unit.
  eapply trans_push; eauto; [].
  split; auto; done.

  (* nop *)
  intros <-.
  exists tm_unit.
  eapply trans_nop; eauto; [].
  split; auto; done.

(* rf *)
intros (iw & ->).
invert Htrco.
intros l Htrco' Hwritesto Hreads Hexecw Hnexec Hpriors.
destruct Hwritesto as (v & Hwrite).
exists v.
invert Hreads.
  (* read *)
  intro His.
  eapply trans_read; eauto; [].
  split; auto; done.

  (* rw *)
  intros v' His.
  eapply trans_rw; eauto; [].
  split; auto; done.
Qed.


Lemma trans_exec_decidable :
  forall H i p,
    trco H
    -> H {{ ev_init i p }}
    -> decidable (exists H' v, trans H (tr_exec i v) p H').
Proof.
intros H i p Htrco Hinit.
so (trco_executes_decidable H i Htrco) as [(h & Htrco' & Hcoacy' & Hexecutes) | Hcond].
  left.
  so (trco_trans _#4 Htrco' Hcoacy' Hexecutes Hinit) as (v & Htrans).
  eauto; done.

  right.
  intros (v & H' & Htrans).
  destruct Hcond.
  so (trans_exec_form _#5 Htrans) as (h & -> & Hexecutes).
  exists h.
  exploit (preservation_trco_ninit _#4 Htrans) as Htrco'; [discform | auto |].
  do2 2 split; auto; [].
  invert Htrans; intros; subst; assumption.
Qed.


Lemma ident_progress :
  forall H i p,
    trco H
    -> H {{ ev_init i p }}
    -> right_ident H i  (* don't actually need this *)
    -> blocked H i \/ exists H' v, trans H (tr_exec i v) p H'.
Proof.
intros H i p Htrco Hinit _.
so (trans_exec_decidable _#3 Htrco Hinit) as [Htrans | Hntrans].
  right; assumption.

  left.
  unfold blocked.
  intros (H' & v & p' & Htrans); destruct Hntrans; [].
  exists H', v.
  assert (H {{ ev_init i p' }}) as Hinit' by (invert Htrans; auto).
  so (init_unique _#4 Htrco Hinit Hinit') as Heq.
  subst p'; assumption.
Qed.


(* Expression rightness *)

Lemma idtp_resp_po :
  forall H U P I1 I2 I3 i i' t p,
    trco H
    -> restricteq I2 p = nil; (i, (t, p))
    -> idtp_sat U P H (I1;; I2;; I3)
    -> po H i' i
    -> ~ exec H i'
    -> indom i' (restricteq I1 p).
Proof.
intros H U P I1 I2 I3 i i' t p Htrco HeqI2p HsatI Hpo Hnexec'.
so (po_init _#3 Hpo) as (p' & Hinit' & Hinit).
assert (p = p').
  eapply init_unique; eauto; [].
  refine (idtp_sat_init _#8 Htrco HsatI _); [].
  rewrite -> !restricteq_app.
  apply lookup_app_right.
  apply lookup_app_left.
  rewrite -> HeqI2p.
  apply lookup_hit; eauto; done.
subst p'.
assert (In (i, (t, p)) I2) as HiniI2.
  apply lookup_iff_in; [].
  eapply (restricteq_contain _ p).
  rewrite -> HeqI2p.
  apply lookup_hit; done.
destruct HsatI as (_ & Hininit & Hinitin & Horder).
so (Hinitin _ _ Hinit' Hnexec') as Hindom.
so (indom_iff_in _#4 andel Hindom) as ((t', p'), Hin).
assert (p = p').
  refine (init_unique _#4 Htrco Hinit' _); [].
  so (Hininit _#3 Hin) as (_ & Hinit'' & _).
  exact Hinit''.
subst p'.
so (in_app_or _#4 Hin) as [Hin3 | Hin12].
  have (I3 {{ (i', (t', p)) }}) as Hin3.
  exfalso.
  so (in_split _#3 Hin3) as (I3b & I3a & HeqI3).
  so (in_split _#3 HiniI2) as (I2b & I2a & HeqI2).
  refine (po_irreflex _ i Htrco _).
  eapply (po_trans _ _ i'); eauto; [].
  refine (Horder (I1;; I2a) (I2b;; I3a) I3b i t i' t' p _); [].
  subst I2 I3.
  repeat (progress (rewrite <- ?app_assoc; rewrite <- ?app_comm_cons)).
  reflexivity.

  so (in_app_or _#4 Hin12) as [Hin2 | Hin1].
    have (I2 {{ (i', (t', p)) }}) as Hin2.
    exfalso.
    so (in_restricteq _#4 Hin2) as Hin2'.
    rewrite -> HeqI2p in Hin2'.
    destruct Hin2' as [Heq | ?]; auto.
    injection Heq; intros; subst.
    exact (po_irreflex _ _ Htrco Hpo).

    have (I1 {{ (i', (t', p)) }}) as Hin1.
    apply indom_iff_in; [].
    exists (t', p).
    apply in_restricteq; auto; done.
Qed.


Lemma restricteq_app_pushout :
  forall I I1 I2 p,
    restricteq I p = I1;; I2
    -> exists I1' I2',
         I = I1';; I2'
         /\ restricteq I1' p = I1
         /\ restricteq I2' p = I2.
Proof.
intros I I1 I2 p Heq.
revert I2 Heq.
induction I as [| (i, (t, p')) I IH].
(* nil *)
intros.
simpl in Heq.
symmetry in Heq.
so (app_eq_nil _#3 Heq) as (? & ?).
exists nil, nil.
repeat split; auto; done.

(* cons *)
intros.
simpl in Heq.
remember (eq_thread_dec p p') as b.
destruct b.
  subst p'.
  so (app_eq_cons _#5 Heq) as [(I2a & HeqI2 & Heq') | (HeqI2 & HeqI1)].
    so (IH _ Heq') as (I1' & I2' & Heq'' & Heq1 & Heq2).
    exists I1', (I2'; (i, (t, p))).
    do2 2 split.
      subst I.
      rewrite <- app_comm_cons.
      reflexivity.

      assumption.

      simpl.
      rewrite <- Heqb.
      subst I2.
      f_equal; auto; done.

    exists (I; (i, (t, p))), nil.
    do2 2 split.
      reflexivity.

      simpl.
      rewrite <- Heqb.
      auto; done.

      subst I2.
      reflexivity.

  so (IH _ Heq) as (I1' & I2' & Heq' & Heq1 & Heq2).
  exists I1', (I2'; (i, (t, p'))).
  do2 2 split.
    subst I.
    rewrite <- app_comm_cons.
    reflexivity.

    assumption.

    simpl.
    rewrite <- Heqb.
    assumption.
Qed.


Lemma xpof_right :
  forall H U P I1 I2 I3 p G e t,
    trco H
    -> tgtp_sat H U
    -> lctp_sat U P H P
    -> idtp_sat U P H (I1;; I2;; I3)
    -> xpof U P (restricteq I2 p) G e t
    -> right_xp H p (map car (restricteq I1 p)) U P G e.
Proof.
intros H U P I1 I2 I3 p G e t Htrco HsatU HsatP HsatI Hof.
remember (restricteq I2 p) as I2' eqn:HeqI2'.
revert I1 I2 I3 HeqI2' HsatU HsatP HsatI.
#4 induct Hof;
try (intros; eauto using tpof_context_closedn, tmof_context_closedn, tmof_right, fcof_right, canon_susp, canon_ref with progress; done).

(* bind *)
intros U P I G t1 e1 e2 t2 Hoft1 Hofe1 IH1 Hofe2 IH2 I1 I2 I3 HeqI2' HsatU HsatP HsatI.
#2 apply right_xp_bind;
eauto using tpof_context_closedn with progress.
  erewrite <- xpof_idtp; eauto; [].
  rewrite -> HeqI2'.
  rewrite <- map_app.
  rewrite <- restricteq_app.
  apply (IH2 (I1;; I2) nil I3); auto; done.

  intro Hactions.
  destruct Hactions.
  rewrite <- (xpof_idtp _#6 Hofe2).
  reflexivity.

(* bind_init *)
intros U P I2a I2b G t1 e1 e2 t2 Hinit Hoft1 Hofe1 IH1 Hofe2 IH2 I1 I2 I3 HeqI2' HsatU HsatP HsatI.
symmetry in HeqI2'. 
so (restricteq_app_pushout _#4 HeqI2') as (I2a' & I2b' & HeqI2 & HeqI2a & HeqI2b).
subst I2.
rewrite <- app_assoc in HsatI.
#2 apply right_xp_bind;
eauto using tpof_context_closedn with progress.
  apply (IH1 I1 I2a' (I2b';; I3)); auto; [].
  rewrite <- app_assoc.
  assumption.

  erewrite <- xpof_idtp; eauto; [].
  rewrite <- HeqI2a.
  rewrite <- map_app.
  rewrite <- restricteq_app.
  apply (IH2 (I1;; I2a') I2b' I3); auto; done.

(* rmw *)
intros U P G t m1 m2 Hoft Hofm1 Hofm2 I1 I2 I3 HeqI2' HsatU HsatP HsatI.
apply right_xp_rmw; eauto using tpof_context_closedn, tmof_right; [].
intro Hvaluem1.
so (canon_ref _#5 Hofm1 Hvaluem1) as (l & Heqm1).
rewrite -> Heqm1 in Hofm1.
invert Hofm1; [].
intros t' Hlookup Heqt.
so (lookup_iff_in _#5 ander Hlookup) as Hin.
destruct HsatP as (_ & Hinitialized & Hwf & _).
so (Hinitialized _ _ Hin) as (iw & _ & (v & Hwrites) & _ & _ & _).
so (Hwf _#4 Hin Hwrites) as (Hvaluev & Hofv).
exists v.
thus (closed v) as Hclosed by tmof_context_closed.
split.
  so (weaken_tmof _ _ _ G _ _ Hofv) as Hofv'.
  rewrite -> Heqt in Hofv'.
  rewrite -> closed_subst in Hofv'; auto; [].
  rewrite -> app_nil_r in Hofv'.
  exact Hofv'.

  so (weaken_value _ G _ Hvaluev) as Hvaluev'.
  rewrite -> closed_subst in Hvaluev'; auto; [].
  rewrite -> app_nil_r in Hvaluev'.
  exact Hvaluev'.

(* action *)
intros U P G i t p' I1 I2 I3 HeqI2' HsatU HsatP HsatI.
assert (lookup i (restricteq I2 p) (t, p')) as Hlookup.
  rewrite <- HeqI2'.
  apply lookup_hit; done.
thus (p = p') as Heqp by lookup_restricteq.
subst p'.
thus (lookup i I2 (t, p)) as Hlookup' by restricteq_contain.
thus (In (i, (t, p)) I2) as Hin by lookup_iff_in.
so HsatI as (_ & Hininit  & _).
exploit (Hininit i t p) as (a & Hinit & His & Hofa & Hnexec); [|].
  apply in_or_app; right.
  apply in_or_app; left.
  assumption.
eapply right_xp_action; eauto using acof_right; [].
intros i'.
split.
  intros (Hprereq & Hnexec').
  apply indom_iff_in_car.
  toshow (indom i' (restricteq I1 p)).
  eapply idtp_resp_po; eauto; done.

  intro Hindom.
  so (in_map_iff _#5 andel Hindom) as ((i'', (t', p')) & Heq & Hin').
  simpl in Heq; subst i''.
  so (lookup_restricteq _#5 (lookup_iff_in _#5 andel Hin')) as Heqp.
  subst p'.
  so (lookup_iff_in _#5 ander (restricteq_contain _#4 (lookup_iff_in _#5 andel Hin'))) as Hin''.
  clear Hindom Hin'.
  split.
    so (in_split _#3 Hin) as (I2a & I2b & HeqI2).
    so (in_split _#3 Hin'') as (I1a & I1b & HeqI1).
    apply (HsatI ander ander ander I1b (I1a;; I2b) (I2a;; I3) i' t' i t p); [].
    subst; autorewrite with canonlist; auto; done.

    exploit (HsatI ander andel i' t' p) as (_ & _ & _ & _ & Hnexec'); [|].
      apply in_or_app; right; apply in_or_app; right; assumption.

      assumption.
Qed.


(* Expression progress *)

Lemma closed_value_context :
  forall G v,
    closed v
    -> value G v
    -> value nil v.
Proof.
intros G v Hclo Hvalue.
induction Hvalue; auto with static; [].
destruct (open_var i); assumption.
Qed.


Lemma closed_right_tm_context :
  forall G m,
    closed m
    -> right_tm G m
    -> right_tm nil m.
Proof.
intros G m Hclo Hrt.
revert Hclo.
induct Hrt;
try (complete
       (intros;
        tryall ltac:(fun H => reduce_closed in H);
        eauto with progress)).
(* var *)
intros G i t Hindex Hclo.
destruct (open_var i); assumption.
Qed.


Lemma weaken_tstep :
  forall G1 G2 m m',
    tstep G1 m m'
    -> tstep (G1;; G2) (subst (sh (length G2)) m) (subst (sh (length G2)) m').
Proof.
intros G1 G2 m m' Hstep.
revert G2.
#2 induct Hstep; try (intros; simp_sub; eauto using weaken_value with dynamic; done).

(* app3 *)
intros G1 t m1 v2 Hvalue G2.
simp_sub.
(match goal with
 | |- tstep _ (tm_app (tm_lam ?t ?b) ?a) _
   => exploit (@tstep_app3 (G1;; G2) t b a) as Hstep
 end).
  apply weaken_value; assumption.

  simp_sub in Hstep.
  replace (length G2 + 1) with (S (length G2)) in Hstep by omega.
  rewrite -> compose_sh_succ_dot in Hstep.
  simp_sub in Hstep.
  exact Hstep.

(* app3' *)
intros G1 t1 t2 m1 v2 Hvalue G2.
simp_sub.
(match goal with
 | |- tstep _ (tm_app (tm_fun ?t1 ?t2 ?b) ?a) _
   => exploit (@tstep_app3' (G1;; G2) t1 t2 b a) as Hstep
 end).
  apply weaken_value; assumption.

  simp_sub in Hstep.
  replace (length G2 + 2) with (S (S (length G2))) in Hstep by omega.
  rewrite -> !compose_sh_succ_dot in Hstep.
  simp_sub in Hstep.
  exact Hstep.
Qed.


Lemma right_tm_progress_closed :
  forall G m,
    right_tm G m
    -> closed m
    -> value nil m \/ (exists m', tstep G m m').
Proof.
intros G m Hrt Hclo.
exploit (right_tm_progress m) as [Hvalue | (m' & Hstep)].
  eapply closed_right_tm_context; eauto; done.

  auto; done.
  
  right.
  exists (subst (sh (length G)) m').
  so (weaken_tstep _ G _ _ Hstep) as Hstep'.
  autorewrite with canonlist in Hstep'.
  rewrite -> closed_subst in Hstep' at 1; auto; done.
Qed.


Lemma fresh_ident :
  forall H, exists i, ~ H {{ ev_init i p | p }}.
Proof.
intros H.
so (ident_infinite (ids H)) as (i & Hfresh).
exists i.
contradict Hfresh.
apply init_ids; assumption.
Qed.


Lemma fence_trans :
  forall H d p H' f,
    trans H d p H'
    -> exists H'', trans H (addfence f d) p H''.
Proof.
intros H d p H' f Htrans.
revert Htrans.
case d; clear d; try (exists H'; assumption); [].
intros fl i a Htrans.
invert Htrans; clear Htrans.
intros Hfresh <-.
simpl.
exists (H; ev_init i p; ev_is i a;; map (fun f' => ev_fence f' i) (fc_list (fl_cons f fl))).
apply trans_init; auto; done.
Qed.


Lemma addfence_closed :
  forall f d,
    closed f
    -> closed d
    -> closed (addfence f d).
Proof.
intros f d Hclf Hcld.
revert Hcld.
case d; auto; [].
intros fl i a Hcld.
simpl.
reduce_closed in Hcld.
intros Hcla Hclfl.
reduce_closed; auto; [].
unfold shift1; rewrite -> !closed_subst; auto.
Qed.


Lemma right_tg_notag_closed :
  forall H G T,
    right_tg H G T
    -> ~ In cl_tg G
    -> closed T.
Proof.
intros H G T Hrt.
induct Hrt.

(* var *)
intros _ G i Hindex Hnotag.
destruct Hnotag.
eapply index_in; eauto; done.

(* lit *)
intros H G T H0 H1.
reduce_closed.
Qed.


Lemma right_fc_notag_closed :
  forall H G f,
    right_fc H G f
    -> ~ In cl_tg G
    -> closed f.
Proof.
intros H G f Hrt.
induct Hrt; try (intros; reduce_closed; done).
intros.
reduce_closed; [].
eapply right_tg_notag_closed; eauto; done.
Qed.


Lemma writes_closed :
  forall H i l v,
    trco H
    -> writes H i l v
    -> closed v.
Proof.
intros H i l v Htrco Hwrites.
invert Hwrites;
intro His;
so (is_closed _#3 Htrco His) as Hclosed;
reduce_closed in Hclosed;
auto.
Qed.


Lemma trans_tr_exec_closed :
  forall H i v p H',
    trco H
    -> trans H (tr_exec i v) p H'
    -> closed v.
Proof.
intros H i v p H' Htrco Htrans.
invert Htrans; try (intros; subst; reduce_closed; auto; done);
intros; eapply writes_closed; eauto.
Qed.


Lemma right_xp_progress :
  forall H p L U P G e,
    trco H
    -> ~ In cl_tg G
    -> right_xp H p L U P G e
    -> inactive H e
       \/ (exists H' e' d,
             xstepp U P G e e' d
             /\ trans H d p H'
             /\ closed d)
       \/ (exists v, value nil v /\ e = xp_ret v).
Proof.

Ltac closure_progress m H :=
  so (closed_decidable m) as [H | H];
  [| left;
     eapply inactive_open; eauto with progress;
     contradict H;
     reduce_closed in H; auto; done ].
  
Ltac simple_progress :=
  subst;
  right; left;
  do 3 eexists;
  do2 2 split; [eauto using weaken_value_all with progress dynamic | apply trans_nothing | reduce_closed; reflexivity]; done.

Ltac init_progress H p a :=
  let i := fresh "i"
  in
    subst;
    right; left;
    so (fresh_ident H) as (i & ?);
    exists (H; ev_init i p; ev_is i a;; map (fun f => ev_fence f i) (fc_list fl_nil)), (xp_action i), (tr_init fl_nil i a);
    do2 2 split; [eauto using weaken_value_all with progress | eapply trans_init; eauto | reduce_closed; unfold shift1; rewrite -> closed_subst; auto];
    done.
  
intros H p L U P G e Htrco Hnotag Hrte.
revert Htrco Hnotag.
#13 induct Hrte.

(* ret *)
intros H p L U P G m Hrtm Htrco _.
closure_progress m Hclo.
so (right_tm_progress_closed _ _ Hrtm Hclo) as [Hval | (m' & Hstep)].
  right; right.
  eexists; eauto using weaken_value_all; done.

  simple_progress.

(* force *)
intros H p L U P G m Hrtm Hform Htrco _.
closure_progress m Hclo.
so (right_tm_progress_closed _ _ Hrtm Hclo) as [Hval | (m' & Hstep)].
  so (Hform Hval) as (e & Heqm).
  simple_progress.

  simple_progress.

(* bind *)
intros H p L U P G t e1 e2 _ _ IH1 _ IH2 _ Htrco Hnotag.
so (IH1 Htrco Hnotag) as [Hinactive1 |[ (H' & e1' & d & Hstep & Htrans & Hclosed) | (v & Hvalue & Heqe1)]].
  so (init_decidable e1) as [Hinit | Hninit].
    assert (~ In cl_tg (G; cl_vl t)) as Hnotag'.
      contradict Hnotag.
      discriminate_in Hnotag; auto; done.
    so (IH2 Htrco Hnotag') as [Hinactive2 |[ (H' & e2' & d & Hstep & Htrans & Hclosed) | (v & Hvalue & Heqe2)]].
      left.
      eauto with progress; done.
  
      right; left.
      exists H', (xp_bind t e1 e2'), d.
      do2 2 split; auto; [].
      apply xstepp_bind2; auto; [].
      unfold shift1.
      rewrite -> closed_subst; auto; done.

      left.
      subst; eauto with progress; done.

    left.
    auto with progress; done.

  right; left.
  exists H', (xp_bind t e1' e2), d.
  split; auto with progress; done.

  simple_progress.

(* spec *)
intros H p L U P G v1 v2 e _ _ Hval1 Hval2 Hrte IH Htrco Hnotag.
so (eq_dec_decidable _ eq_term_dec v1 v2) as [Heq | Hneq].
  simple_progress.

  so (IH Htrco Hnotag) as [Hinactive |[ (H' & e' & d & Hstep & Htrans) | (v & Hvalue & Heqe)]].
    left.
    eauto with progress; done.

    right; left.
    exists H', (xp_spec v1 v2 e'), d.
    split; auto with progress; done.

    left.
    subst; eauto with progress; done.

(* read *)
intros H p L U P G m Hrtm Hform Htrco _.
closure_progress m Hclo.
so (right_tm_progress_closed _ _ Hrtm Hclo) as [Hval | (m' & Hstep)].
  so (Hform Hval) as (l & Heqm).
  init_progress H p (ac_read l).

  simple_progress.

(* write *)
intros H p L U P G m1 m2 Hrtm1 Hrtm2 Hform Htrco _.
closure_progress m1 Hclo1.
closure_progress m2 Hclo2.
so (right_tm_progress_closed _ _ Hrtm1 Hclo1) as [Hval1 | (m1' & Hstep1)].
  so (right_tm_progress_closed _ _ Hrtm2 Hclo2) as [Hval2 | (m2' & Hstep2)].
    so (Hform Hval1) as (l & Heqm1).
    init_progress H p (ac_write l m2).

    simple_progress.

  simple_progress.

(* rw *)
intros H p L U P G m1 m2 Hrtm1 Hrtm2 Hform Htrco _.
closure_progress m1 Hclo1.
closure_progress m2 Hclo2.
so (right_tm_progress_closed _ _ Hrtm1 Hclo1) as [Hval1 | (m1' & Hstep1)].
  so (right_tm_progress_closed _ _ Hrtm2 Hclo2) as [Hval2 | (m2' & Hstep2)].
    so (Hform Hval1) as (l & Heqm1).
    init_progress H p (ac_rw l m2).

    simple_progress.

  simple_progress.

(* rmw *)
intros H p L U P G t m1 m2 Hclosedn Hrt1 Hrt2 Hform Htrco Hnotag.
closure_progress m1 Hclo1.
so (right_tm_progress_closed _ _ Hrt1 Hclo1) as [Hval1 | (m1' & Hstep1)].
  so (Hform Hval1) as (v & Hofv & Hvaluev).
  right; left.
  exists H, (xp_bind t (xp_rw m1 (subst1 v m2)) (xp_spec (var 0) (shift1 v) (xp_ret (shift1 v)))), tr_nothing.
  do2 2 split.
    apply xstepp_rmw2; auto using weaken_value_all; done.

    apply trans_nothing; done.

    reduce_closed; reflexivity.

  simple_progress.

(* push *)
intros H p L G Htrco.
init_progress H p ac_push.

(* nop *)
intros H p L G Htrco.
init_progress H p ac_nop.

(* action *)
intros H p L U P G i Hinit Hnexec Hpriors Hrti Htrco.
so (ident_progress _#3 Htrco Hinit Hrti) as [Hblocked | (H' & v & Htrans)].
  left.
  eapply inactive_blocked; eauto with progress; done.

  right; left.
  exists H', (xp_ret v), (tr_exec i v).
  do2 2 split; auto with progress; [].
  reduce_closed.
  eapply trans_tr_exec_closed; eauto; done.

(* new *)
intros H p L U P G b e Hrte _ Htrco _.
right; left.
so (tag_infinite (tags H)) as (T1 & Hfresh1).
so (tag_infinite (T1 :: tags H)) as (T2 & Hfresh2).
exists (H; ev_edge b T1 T2), (subst (dot (tg_lit T2) (dot (tg_lit T1) id)) e), (tr_new b T1 T2).
do2 2 split.
  apply xstepp_new; done.

  apply trans_new; [| |].
    contradict Hfresh1.
    apply tag_declared_tags; assumption.

    contradict Hfresh2.
    right.
    apply tag_declared_tags; assumption.

    contradict Hfresh2.
    left; assumption.

  reduce_closed; done.

(* fence *)
intros H p L U P G f e Hrtf Hrte IH Htrco Hnotag.
so (IH Htrco Hnotag) as [Hinactive |[ (H' & e' & d & Hstep & Htrans & Hclosed) | (v & Hvalue & Heqe)]].
  left.
  eauto with progress; done.

  right; left.
  so (fence_trans _#4 f Htrans) as (H'' & Htrans').
  exists H'', (xp_fence f e'), (addfence f d).
  do2 2 split.
    apply xstepp_fence1; auto; done.

    assumption.

    apply addfence_closed; auto; [].
    eapply right_fc_notag_closed; eauto; done.

  simple_progress.
Qed.


(* Execution state rightness *)

Lemma exof_ind_restrict :
  forall
  (U : tgtp) (P : lctp) (I : idtp)
  (Q : xstate -> Prop),
    (Q nil)
    -> (forall ex p e t,
          ~ indom p ex
          -> Q ex
          -> xpof U P (restricteq I p) nil e t
          -> Q (ex; (p, e)))
    -> forall ex, exof U P I ex -> Q ex.
Proof.
intros U P I Q Hnil Hcons ex Hofex.
rewrite <- (filter_all _ I) in Hofex.
pose (f := (fun (_ : thread) => true)).
remember (filter (fun _ => true) I) as I' eqn:HeqI'.
rewrite <- (filter_iff _ (fun itp => match itp with (_, (_, p)) => f p end)) in HeqI';
  [| intros (i, (t, p)); reflexivity].
assert (forall p, indom p ex -> f p = true) as Hspecf.
  intros p _.
  reflexivity.
clearbody f.
revert f Hspecf HeqI' Hcons.
induct Hofex.

(* nil *)
intros.
exact Hnil.

(* cons *)
intros U P I' ex p e t Hnindom Hofex IH Hofe f Hspecf HeqI' Hcons.
subst I'.
apply (Hcons ex p e t); [| |].
  assumption.

  apply (IH (fun p' => andb (negb (eq_thread_bool p p')) (f p'))); [| |].
    intros p' Hindom.
    rewrite -> (eq_thread_bool_false p p' andel).
      rewrite -> Bool.andb_true_l.
      apply Hspecf.
      apply indom_miss.
      assumption.
  
      intro; subst p'; contradiction.

    rewrite -> restrictneq_filter.
    rewrite -> filter_and.
    apply filter_iff; [].
    intros (i', (t', p')).
    unfold eq_thread_bool.
    generalize (eq_thread_dec p p'); intros [Heq | Hneq].
      subst p'.
      rewrite Hspecf; auto; [].
      apply indom_hit; auto; done.

      reflexivity.

    apply Hcons; done.

    replace (restricteq (filter (fun itp => match itp with (_, (_, p')) => f p' end) I) p) with (restricteq I p) in Hofe.
      assumption.
    rewrite -> !restricteq_filter.
    rewrite -> filter_and.
    apply filter_iff; [].
    intros (i', (t', p')).
    unfold eq_thread_bool.
    generalize (eq_thread_dec p p'); intros [Heq | Hneq].
      subst p'.
      rewrite Hspecf; auto; [].
      apply indom_hit; auto; done.

      rewrite -> Bool.andb_false_l.
      reflexivity.
Qed.


Lemma exof_right :
  forall H U P I ex,
    trco H
    -> tgtp_sat H U
    -> lctp_sat U P H P
    -> idtp_sat U P H I
    -> exof U P I ex
    -> right_ex H U P ex.
Proof.
intros H U P I ex Htrco HsatU HsatP HsatI Hexof.
induct Hexof using exof_ind_restrict.

(* nil *)
apply right_ex_nil; done.

(* cons *)
intros ex p e t Hnindom IH Hofe.
#3 apply right_ex_cons.
  assumption.

  eapply (xpof_right _#3 nil I nil); eauto; [].
  rewrite -> app_nil_l.
  rewrite -> app_nil_r.
  assumption.

  intros i Hinit Hnexec.
  erewrite <- xpof_idtp; eauto; [].
  apply indom_iff_in_car; [].
  so (init_idtp_sat_in _#6 Htrco HsatI Hinit Hnexec) as (t' & Hin).
  apply indom_iff_in; [].
  exists (t', p).
  apply in_restricteq; auto; done.
Qed.


Lemma exof_thread_in :
  forall U P I ex i t p,
    exof U P I ex
    -> In (i, (t, p)) I
    -> indom p ex.
Proof.
intros U P I ex i t p Hexof Hin.
revert Hin.
induct Hexof.

(* nil *)
intros U P Hin.
destruct Hin; done.

(* cons *)
intros U P I ex p' e t' _ _ IH _ Hin.
#2 so (eq_thread_dec p p') as [Heq | Hneq].
  subst p'.
  apply indom_hit; auto; done.

  apply indom_miss; [].
  apply IH; [].
  apply in_restrictneq; auto; done.
Qed.


Theorem ok_right :
  forall S, ok S -> right S.
Proof.
intros ((((U, P), I), H), ex) Hok.
invert Hok; clear Hok.
intros Hhok Hexof.
so Hhok as (HsatU & HsatP & HsatI & Htrco & _).
apply right_i; eauto using exof_right; [].
intros i p Hinit Hnexec.
so (init_idtp_sat_in _#6 Htrco HsatI Hinit Hnexec) as (t & Hin).
eapply exof_thread_in; eauto; done.
Qed.


(* Execution state progress *)

Lemma right_tm_closedn :
  forall G m,
    right_tm G m
    -> closedn (length G) m.
Proof.
intros G m Hrt.
induct Hrt;
try (complete (intros; reduce_closed; trivial; rewrite <- ?under_succ; try eapplyall; eauto));
[|].

(* var *)
intros G i t Hindex.
thus (i < length G) as Hlt by index_length.
unfold closedn.
simp_sub.
rewrite -> project_under_lt; auto; done.

(* fun *)
intros.
reduce_closed; trivial; [].
rewrite <- under_succ2.
assumption.
Qed.


Lemma right_tg_closedn :
  forall H G T,
    right_tg H G T
    -> closedn (length G) T.
Proof.
intros H G T Hrt.
induct Hrt; try (complete (intros; reduce_closed; trivial)).

(* var *)
intros H G i Hindex.
thus (i < length G) as Hlt by index_length.
unfold closedn.
simp_sub.
rewrite -> project_under_lt; auto; done.
Qed.


Lemma right_fc_closedn :
  forall H G f,
    right_fc H G f
    -> closedn (length G) f.
Proof.
intros H G f Hrt.
induct Hrt; intros; reduce_closed; eauto using right_tg_closedn.
Qed.


Lemma right_xp_closedn :
  forall H p L U P G e,
    right_xp H p L U P G e
    -> closedn (length G) e.
Proof.
intros H p L U P G e Hrt.
induct Hrt;
try (complete (intros; reduce_closed; trivial; rewrite <- ?under_succ; rewrite <- ?under_succ2; try eapplyall; eauto using right_fc_closedn, right_tm_closedn)); [].

(* rmw *)
intros H p L U P G t m1 m2 Hclosedn Hm1 Hm2 _.
reduce_closed; eauto using right_tm_closedn; [].
rewrite <- under_succ.
replace (S (length G)) with (length (G; cl_vl t)) by reflexivity.
apply right_tm_closedn; auto; done.
Qed.


Lemma inactive_all_blocked :
  forall H p L U P G e,
    trco H
    -> right_xp H p L U P G e
    -> inactive H e
    -> forall i, In i (xp_actions e) -> blocked H i.
Proof.
intros H p L U P G e _ Hrt Hinactive i Hin.
revert L G Hrt Hin.
#5 induct Hinactive;
try (intros; invert Hrt; intros; eauto; done).

(* basic *)
intros H e Hbasic Hnclosed _ _ _ Hin.
invert Hbasic; try (intros; subst; destruct Hin; done); [].
intros.
subst e.
destruct Hnclosed.
reduce_closed; done.

(* blocked *)
intros H i0 Hblocked _ _ _ Hin.
destruct Hin as [-> | Hfalse]; [| destruct Hfalse].
assumption.

(* bind1 *)
intros H t e1 e2 _ IH1 _ IH2 L G Hrt Hin.
simpl in Hin.
invert Hrt.
intros _ Hrt1 Hrt2.
so (in_app_or _#4 Hin) as [|]; eauto; done.

(* bind3 *)
intros H t e1 e2 _ IH Hninit L G Hrt Hin.
invert Hrt.
intros _ Hrt1 _ Hinit.
simpl in Hin.
so (in_app_or _#4 Hin) as [Hin' |]; eauto; [].
destruct Hninit.
apply Hinit; [].
intro Heq.
rewrite -> Heq in Hin'.
destruct Hin'; done.

(* spec2 *)
intros H v1 v2 v H0 H1 L G Hrt Hin.
destruct Hin; done.
Qed.


Lemma closed_inactive :
  forall H e,
    closed e
    -> inactive H e
    -> (exists v1 v2 e',
          within (xp_spec v1 v2 e') e
          /\ closed v1
          /\ closed v2
          /\ v1 <> v2)
       \/ (exists i, In i (xp_actions e)).
Proof.
intros H e Hclosed Hinactive.
revert Hclosed.
induct Hinactive.

(* open *)
intros; contradiction.

(* blocked *)
intros H i Hblocked Hclosed.
right.
exists i.
left; reflexivity.

(* bind1 *)
intros H t e1 e2 _ IH _ _ Hclosed.
reduce_closed in Hclosed.
intros _ Hclosed' _.
so (IH Hclosed') as [(v1 & v2 & e' & Hwithin & Hclosed1 & Hclosed2 & Hneq) | (i & Hwithin)].
  left.
  exists v1, v2, e'.
  repeat2 split; auto; [].
  apply within_bind1; auto; done.

  right.
  exists i.
  simpl.
  apply in_or_app; right; assumption.

(* bind2 *)
intros H t e1 v _ IH _ Hclosed.
reduce_closed in Hclosed.
intros _ Hclosed' _.
so (IH Hclosed') as [(v1 & v2 & e' & Hwithin & Hclosed1 & Hclosed2 & Hneq) | (i & Hwithin)].
  left.
  exists v1, v2, e'.
  repeat2 split; auto; [].
  apply within_bind1; auto; done.

  right.
  exists i.
  simpl; assumption.

(* bind3 *)
intros H t e1 e2 _ IH _ Hclosed.
reduce_closed in Hclosed.
intros _ Hclosed' _.
so (IH Hclosed') as [(v1 & v2 & e' & Hwithin & Hclosed1 & Hclosed2 & Hneq) | (i & Hwithin)].
  left.
  exists v1, v2, e'.
  repeat2 split; auto; [].
  apply within_bind1; auto; done.

  right.
  exists i.
  simpl.
  apply in_or_app; right; assumption.

(* spec1 *)
intros H v1 v2 e Hneq _ _ Hclosed.
left.
reduce_closed in Hclosed.
intros _ Hclosed2 Hclosed1.
exists v1, v2, e.
repeat2 split; auto using within_refl; done.

(* spec2 *)
intros H v1 v2 v Hneq _ Hclosed.
left.
reduce_closed in Hclosed.
intros _ Hclosed2 Hclosed1.
exists v1, v2, (xp_ret v).
repeat2 split; auto using within_refl; done.

(* fence *)
intros H f e _ IH Hclosed.
reduce_closed in Hclosed.
intros Hclosed' _.
so (IH Hclosed') as [(v1 & v2 & e' & Hwithin & Hclosed1 & Hclosed2 & Hneq) | (i & Hwithin)].
  left.
  exists v1, v2, e'.
  repeat2 split; auto; [].
  apply within_fence; auto; done.

  right.
  exists i.
  simpl; assumption.
Qed.


Lemma right_xp_actions :
  forall H p L U P G e i,
    right_xp H p L U P G e
    -> In i (xp_actions e)
    -> H {{ ev_init i p }} /\ ~ exec H i.
Proof.
intros H p L U P G e i Hrte Hin.
revert Hin.
#2 induct Hrte; auto;
try (intros;
     match goal with
     | Hin:In _ (xp_actions _) |- _ => destruct Hin; done
     end).

(* bind *)
intros H p L U P G t e1 e2 H0 H1 H2 H3 H4 H5 Hin.
simpl in Hin.
so (in_app_or _#4 Hin) as [|]; auto; done.

(* action *)
intros H p L U P G i' Hinit Hnexec _ _ Hin.
simpl in Hin.
destruct Hin as [-> | Hfalse]; [| destruct Hfalse].
auto; done.
Qed.


Lemma exec_blocked :
  forall H i,
    exec H i
    -> blocked H i.
Proof.
intros H i Hexec.
intros (H' & v & p & Htrans).
assert (executable H i) as Hexecutable by (invert Htrans; auto).
destruct Hexecutable as (Hnexec & _).
contradiction.
Qed.
    

Theorem right_ex_progress :
  forall H U P ex,
    trco H
    -> right_ex H U P ex
    -> Forall (fun pe : thread * term =>
               let (p, e) := pe
               in
                 (exists v1 v2 e',
                    within (xp_spec v1 v2 e') e
                    /\ closed v1
                    /\ closed v2
                    /\ v1 <> v2)
                 \/ ((exists i, H {{ ev_init i p }} /\ ~ exec H i)
                     /\ (forall i, H {{ ev_init i p }} -> blocked H i))
                 \/ (exists H' e' d,
                       xstepp U P nil e e' d
                       /\ trans H d p H')
                 \/ (exists v, value nil v /\ e = xp_ret v)) ex.
Proof.
intros H U P ex Htrco Hrt.
revert Htrco.
induct Hrt.

(* nil *)
intros H U P Htrco.
apply Forall_nil; done.

(* cons *)
intros H U P ex p e _ IH Hrte Hinactions Htrco.
apply Forall_cons; auto; [].
assert (~ In cl_tg nil) as Hnotag by (intro Hin; destruct Hin).
so (right_xp_progress _#7 Htrco Hnotag Hrte) as [Hinactive | [(H' & e' & d & Hstep & Htrans & _) | (v & Hvalue & Heqe)]].
  so (closed_closedn _ ander (right_xp_closedn _#7 Hrte)) as Hclosed.
  so (closed_inactive _ _ Hclosed Hinactive) as [Hmoot | Hact].
    left; assumption.

    right; left.
    split.
    destruct Hact as (i & Hwithin).
    exists i.
    eapply right_xp_actions; eauto; done.
    
    intros i Hinit.
    so (exec_decidable H i) as [Hexec | Hnexec].
    apply exec_blocked; assumption.

    eapply inactive_all_blocked; eauto; done.

    right; right; left.
    eauto; done.

    right; right; right; eauto; done.
Qed.


Theorem right_progress :
  forall U P I H ex,
    right (U, P, I, H, ex)
    -> (Exists (fun pe : thread * term =>
                let (p, e) := pe
                in
                  exists v1 v2 e',
                    within (xp_spec v1 v2 e') e
                    /\ closed v1
                    /\ closed v2
                    /\ v1 <> v2) ex)
       \/
       ((exists i, H {{ ev_init i p | p }} /\ ~ exec H i)
        /\ (forall i, H {{ ev_init i p | p }} -> blocked H i))
       \/
       (exists H' ex' d p,
          exstepp U P ex ex' d p
          /\ trans H d p H')
       \/
       (Forall (fun pe : thread * term =>
                let (p, e) := pe
                in
                  exists v, value nil v /\ e = xp_ret v) ex).
Proof.
intros U P I H ex Hrt.
invert Hrt.
intros Htrco Hrtex Hinitin.
so (right_ex_progress _#4 Htrco Hrtex) as Hprogress.
clear Hrt.
assert ((exists H' ex' d p,
            exstepp U P ex ex' d p
            /\ trans H d p H')
         \/
         (Exists (fun pe : thread * term =>
                 let (p, e) := pe
                  in
                    exists v1 v2 e',
                      within (xp_spec v1 v2 e') e
                      /\ closed v1
                      /\ closed v2
                      /\ v1 <> v2) ex)
         \/
         ((forall i p, H {{ ev_init i p }} -> indom p ex -> blocked H i)
          /\
          ((exists i, H {{ ev_init i p | p }} /\ ~ exec H i)
           \/
           (Forall (fun pe : thread * term =>
                    let (p, e) := pe
                    in
                      exists v, value nil v /\ e = xp_ret v) ex)))) as Hgoal.
  clear Hinitin.
  revert Hrtex.
  #2 induct Hprogress.
    (* nil *)
    intros _.
    right; right.
    split.
      intros ? ? ? Hindom.
      invert Hindom; done.

      right; apply Forall_nil; done.

    (* cons *)
    intros (p, e) ex Hprogress _ IH Hrtex.
    invert Hrtex; clear Hrtex; [].
    intros Hrtex _ Hinitp.
    #3 so (IH Hrtex) as [Hcondtl | [Hcondtl | Hcondtl]]; clear IH.
      left.
      destruct Hcondtl as (H' & ex' & d & p' & Hstep & Htrans).
      exists H', (ex'; (p, e)), d, p'.
      split; auto; [].
      apply exstepp_miss; assumption.

      right; left.
      apply Exists_cons_tl; assumption.

      #4 destruct Hprogress as [Hcond | [Hcond | [Hcond | Hcond]]].
        right; left.
        apply Exists_cons_hd; assumption.

        right; right.
        destruct Hcond as (Hexi & Halli).
        destruct Hcondtl as (Hallitl & Hcondtl).
        split.
          intros i p' Hinit Hindom.
          #2 invert Hindom.
            intro; subst p'.
            apply Halli; auto; done.

            intro Hindom'.
            eapply Hallitl; eauto; done.

          left.
          destruct_all Hexi; eauto; done.

        left.
        destruct Hcond as (H' & e' & d & Hstep & Htrans).
        exists H', (ex; (p, e')), d, p.
        auto with progress; done.

        right; right.
        destruct Hcondtl as (Hallitl & Hcondtl).
        split.
          intros i p' Hinit Hindom.
          #2 invert Hindom.
            intro; subst p'.
            destruct Hcond as (v & _ & ->).
            so (exec_decidable H i) as [Hexec | Hnexec].
              apply exec_blocked; auto; done.

              exfalso.
              eapply Hinitp; eauto; done.
            
            intro Hindom'.
            eapply Hallitl; eauto; done.

          destruct Hcondtl as [Hcondtl | Hcondtl].
            left; assumption.

            right.
            apply Forall_cons; auto; done.
destruct Hgoal as [Hcond | [Hcond | (Halli & [Hcond | Hcond])]].
  right; right; left; assumption.

  left; assumption.

  right; left.
  split; auto; [].
  intros i (p & Hinit).
  so (exec_decidable H i) as [Hexec | Hnexec].
    apply exec_blocked; auto; done.
    
    eapply Halli; eauto; done.

    right; right; right; assumption.
Qed.


(* Peroration *)

Lemma evolve_progress :
  forall H U P I ex d p H' ex',
    hok H U P I
    -> exof U P I ex
    -> exstep U P ex ex' d p
    -> trans H d p H'
    -> exists U' P' I', evolve U P I d p U' P' I'.
Proof.
intros H U P I ex d p H' ex' Hhok Hofex Hstep Htrans.
thus (trofo U P (restricteq I p) nil d) as Htrofo by exstep_trofo.
clear Hofex.
invert Htrofo.

(* nothing *)
intros <-.
exists U, P, I; apply evolve_nothing; done.

(* init *)
intros fl i a t Hoffl Hofa <-.
exists U, P, (I; (i, (t, p))).
apply evolve_init; auto; [].
toshow (~ indom i I).
invert Htrans; intros Hninit _.
contradict Hninit.
rename Hninit into Hindom.
so (indom_iff_in _#4 andel Hindom) as ((t', p') & Hin).
exists p'.
destruct Hhok as (_ & _ & (_ & Hininit & _) & _).
so (Hininit _#3 Hin) as (_ & Hinit & _).
exact Hinit.

(* exec *)
intros i v <-.
assert (H {{ ev_init i p }}) as Hinit by (invert Htrans; intros; assumption).
assert (exists tp, lookup i I tp) as ((t, p') & Hlookup).
  apply indom_lookup.
  destruct Hhok as (_ & _ & HsatI & _).
  destruct HsatI as (_ & _ & Hinitin & _).
  apply (Hinitin i p); [|].
    exact Hinit.

    assert (executable H i) as (Hnexec & _) by (invert Htrans; intros; assumption).
    exact Hnexec.
assert (p = p').
  destruct Hhok as (_ & _ & HsatI & Htrco & Hcoacy).
  destruct HsatI as (_ & Hininit & _).
  so (Hininit _#3 (lookup_iff_in _#5 ander Hlookup)) as (_ & Hinit' & _).
  exact (init_unique _#4 Htrco Hinit Hinit').
subst p'.
so (lookup_iff_in _#5 ander Hlookup) as Hin.
so (in_split _#3 Hin) as (I3 & I1 & Heq).
subst I.
exists U, P, (I1;; I3).
apply evolve_exec; auto; [].
#2 invert Htrans;
try (intros;
     destruct Hhok as (_ & _ & HsatI & Htrco & Hcoacy);
     (match goal with
      | His : In (ev_is _ _) H |- _ => so (idtp_sat_resolve _#8 Htrco HsatI His Hlookup) as Hofa
      end);
     invert Hofa;
     intros; subst; apply tmof_unit; done).
  (* read *)
  intros l iw _ His _ Hwrites _ _ Heq1.
  destruct Hhok as (_ & HsatP & HsatI & Htrco & Hcoacy).
  so (idtp_sat_resolve _#8 Htrco HsatI His Hlookup) as Hofa.
  invert Hofa.
  intros t' Hlookupl Heqt.
  rewrite <- Heqt.
  simpl.
  simp_sub.
  so (lctp_sat_resolve _#8 HsatP Hwrites Hlookupl) as (_ & Hofv).
  assumption.

  (* rw *)
  intros vw l iw _ His _ Hwrites _ _ Heq1.
  destruct Hhok as (_ & HsatP & HsatI & Htrco & Hcoacy).
  so (idtp_sat_resolve _#8 Htrco HsatI His Hlookup) as Hofa.
  invert Hofa.
  intros t' Hlookupl _ _ Heqt.
  rewrite <- Heqt.
  simpl.
  simp_sub.
  so (lctp_sat_resolve _#8 HsatP Hwrites Hlookupl) as (_ & Hofv).
  assumption.

(* new *)
intros b T T' <-.
exists (U; T; T'), P, I.
invert Htrans; intros HndeclT HndeclT' Hneq _.
destruct Hhok as ((_ & Hindecl & _) & _).
apply evolve_new; [| |].
  contradict HndeclT.
  apply Hindecl; auto; done.

  contradict HndeclT'.
  apply Hindecl; auto; done.

  assumption.
Qed.


Lemma xstepp_impl_xstep :
  forall U P G e e' d,
    xstepp U P G e e' d
    -> xstep U P G e e' d.
Proof.
intros U P G e e' d Hstep.
induction Hstep; auto with dynamic.
Qed.


Lemma exstepp_impl_exstep :
  forall U P ex ex' d p,
    exstepp U P ex ex' d p
    -> exstep U P ex ex' d p.
Proof.
intros U P ex ex' d p Hstep.
induction Hstep; auto using xstepp_impl_xstep with dynamic.
Qed.


Theorem progress :
  forall U P I H ex,
    ok (U, P, I, H, ex)
    -> (Exists (fun pe : thread * term =>
                let (p, e) := pe
                in
                  exists v1 v2 e',
                    within (xp_spec v1 v2 e') e
                    /\ closed v1
                    /\ closed v2
                    /\ v1 <> v2) ex)
       \/
       ((exists i, H {{ ev_init i p | p }} /\ ~ exec H i)
        /\ (forall i, H {{ ev_init i p | p }} -> blocked H i))
       \/
       (exists S, stepp (U, P, I, H, ex) S)
       \/
       (Forall (fun pe : thread * term =>
                let (p, e) := pe
                in
                  exists v, value nil v /\ e = xp_ret v) ex).
Proof.
intros U P I H ex Hok.
invert Hok.
intros Hhok Hofex.
so (ok_right _ Hok) as Hright.
so (right_progress U P I H ex Hright) as [Hmoot | [Hdeadlock | [Hstep | Hfinal]]].
  left; assumption.

  right; left; assumption.

  right; right; left.
  destruct Hstep as (H' & ex' & d & p & Hstepp & Htrans).
  so (exstepp_impl_exstep U P _#4 Hstepp) as Hstep.
  so (evolve_progress _#9 Hhok Hofex Hstep Htrans) as (U' & P' & I' & Hevolve).
  exists (U', P', I', H', ex').
  eapply stepp_sync; eauto; done.

  right; right; right; assumption.
Qed.
