
Require Import Tactics.
Require Import Dynamic.
Require Import Sequence.
Require Import Relation.
Require Import Hwf.
Require Import History.
Require Import Match.


(* Truncation *)

(* can truncate: is, exec, rf, edge *)
Lemma so_tail :
  forall H h b i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists f i, h = ev_fence f i)
    -> so b (H; h) i1 i2
    -> so b H i1 i2.
Proof.
intros H h b i1 i2 Htrco Hneqinit Hneqfence Hso.
invert Hso.

(* tag *)
intros T1 T2 Hpo Hfence1 Hfence2 Hedge.
apply (@so_tag _#4 T1 T2); [| | |].
  eapply po_tail; eauto; done.

  destruct Hfence1; auto.
  destruct Hneqfence.
  repeat eexists; eauto; done.

  destruct Hfence2; auto.
  destruct Hneqfence.
  repeat eexists; eauto; done.

  destruct Hedge as [Heqh | Hedge]; auto; [].
  exfalso.
  subst h.
  invert Htrco.
  intros _ Hndecl _ _.
  destruct Hndecl.
  discriminate_in Hfence1; intro Hfence1.
  so (fence_wf_prop _#3 (trco_tail _ _ Htrco) Hfence1) as (Hdecl & _).
  invert Hdecl; auto; done.

(* post *)
intros Hpo Hfence.
apply so_post; [|].
  eapply po_tail; eauto; done.

  destruct Hfence; auto; [].
  destruct Hneqfence.
  repeat eexists; eauto; done.

(* pre *)
intros Hpo Hfence.
apply so_pre; [|].
  eapply po_tail; eauto; done.

  destruct Hfence; auto; [].
  destruct Hneqfence.
  repeat eexists; eauto; done.
Qed.


(* can truncate: is, exec, rf, edge *)
Lemma so_truncate :
  forall H H' b i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> so b (H;; H') i1 i2
    -> so b H i1 i2.
Proof.
intros H H' b i1 i2 Htrco Hninit Hnedge Hso.
induction H'.

(* nil *)
assumption.

(* cons *)
apply IHH';
try (intro HH;
     repeat (destruct HH as (? & HH));
     (match goal with
      | Hn : ~ _ |- _ => destruct Hn; repeat eexists; right; eassumption
      end)); [|].
  autorewrite with canonlist in Htrco; prove_trco.
autorewrite with canonlist in Hso.
eapply so_tail; eauto;
intro HH;
repeat (match goal with He : (exists _, _) |- _ => destruct He end);
(match goal with
 | Hn : ~ _ |- _ => destruct Hn; repeat eexists; left; eassumption
 end).
Qed.


(* can truncate: is, exec, rf, edge *)
Lemma sop_tail :
  forall H h b i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists f i, h = ev_fence f i)
    -> sop b (H; h) i1 i2
    -> sop b H i1 i2.
Proof.
intros H h b i1 i2 Htrco Hneqinit Hneqfence Hsop.
pose proof (plus_plusi _#4 Hsop) as Hsop'; clear Hsop.
elim Hsop'; clear i1 i2 Hsop'.
(* one *)
intros i1 i2 Hso.
apply plus_one.
eapply so_tail; eauto.

(* step *)
intros i1 i2 i3 Hso _ IH.
exists i2; split.
  eapply so_tail; eauto.

  apply plus_star; auto.
Qed.


Lemma so_execed_tail :
  forall H h b i1 i2,
    trco (H; h)
    -> exec H i2
    -> so b (H; h) i1 i2
    -> so b H i1 i2.
Proof.
intros H h b i1 i2 Htrco Hexeced2 Hso.
assert (po (H; h) i1 i2) as Hpo by (inversion Hso; auto).
thus (po H i1 i2) as Hpo' by po_execed_tail.
assert (forall T i, (i = i1 \/ i = i2) -> (H; h) {{ ev_fence T i }} -> H {{ ev_fence T i }}) as Hstillfence.
  intros T i Heqi Hfence.
  simpl in Hfence; destruct Hfence as [Heq | ?]; auto.
  subst h.
  so (trco_fence_invert_nexec _#3 Htrco) as Hnexec.
  destruct Hnexec.
  destruct Heqi; subst i; [|].
    so (trco_fence_invert_npo _#3 i2 Htrco) as Hnpo.
    destruct Hnpo.
    assumption.

    assumption.
clear Hpo.
remember (H; h) as H0.
revert Hexeced2 Hpo' HeqH0 Htrco Hstillfence.
induct Hso.

(* tag *)
intros b H0 i1 i2 T1 T2 _ Hfence1 Hfence2 Hedge Hexeced2 Hpo HeqH0 Htrco Hstillfence.
subst H0.
apply (@so_tag _#4 T1 T2).
  assumption.

  apply Hstillfence; auto.

  apply Hstillfence; auto.

  simpl in Hedge; destruct Hedge as [Heq | ?]; auto.
  subst h.
  so (trco_edge_invert_nfence _#4 Htrco) as (Hnfence1 & Hnfence2).
  destruct Hnfence1.
  exists i1.
  simpl in Hfence1; destruct Hfence1 as [Heq | Hin].
    discriminate Heq.

    assumption.

(* post *)
intros b H0 i i' _ Hfence Hexeced2 Hpo HeqH0 Htrco Hstillfence.
subst H0.
apply so_post.
  assumption.

  apply Hstillfence; auto.

(* pre *)
intros b H0 i i' _ Hfence Hexeced2 Hpo HeqH0 Htrco Hstillfence.
subst H0.
apply so_pre.
  assumption.

  apply Hstillfence; auto.
Qed.


Lemma so_execed_truncate :
  forall H1 H2 b i1 i2,
    trco (H1;; H2)
    -> exec H1 i2
    -> so b (H1;; H2) i1 i2
    -> so b H1 i1 i2.
Proof.
intros H1 H2 b i1 i2 Htrco Hexec Hso.
induction H2 as [| h].
(* nil *)
simpl in Hso; auto.

(* cons *)
apply IHlist.
  simpl in Htrco.
  eapply trco_tail; eauto.

  simpl in Hso.
  eapply so_execed_tail; eauto.
  apply exec_app_incl; assumption.
Qed.  


Lemma so_init_tail :
  forall b H i p i1 i2,
    trco (H; ev_init i p)
    -> so b (H; ev_init i p) i1 i2
    -> so b H i1 i2 \/ i = i2.
Proof.
intros b H i p i1 i2 Htrco Hso.
so (po_init_tail _#5 Htrco (so_po _#4 Hso)) as [Hpo' | Heq].
  left.
  invert Hso.
    intros T T' _ Hfence1 Hfence2 Hedge.
    discriminate_in Hfence1; discriminate_in Hfence2; discriminate_in Hedge; intros.
    eapply so_tag; eauto; done.
  
    intros _ Hfence.
    discriminate_in Hfence; intros.
    eapply so_post; eauto; done.

    intros _ Hfence.
    discriminate_in Hfence; intros.
    eapply so_pre; eauto; done.

  right; assumption.
Qed.


Lemma so_po_tail :
  forall H h b i1 i2 i3,
    trco (H; h)
    -> po (H; h) i2 i3
    -> so b (H; h) i1 i2
    -> so b H i1 i2.
Proof.
intros H h b i1 i2 i3 Htrco Hpo23 Hso.
invert Hso.
  intros T T' Hpo Hfence1 Hfence2 Hedge.
  apply (@so_tag _#4 T T'); eauto using po_po_tail, fence_po_tail, edge_fence_tail_1; done.

  intros Hpo Hfence.
  apply so_post; eauto using po_po_tail, fence_po_tail; done.

  intros Hpo Hfence.
  apply so_pre; eauto using po_po_tail, fence_po_tail; done.
Qed.


Lemma so_po_truncate :
  forall b H H' i1 i2 i3,
    trco (H;; H')
    -> po H i2 i3
    -> so b (H;; H') i1 i2
    -> so b H i1 i2.
Proof.
intros b H1 H2 i1 i2 i3 Htrco Hpo Hso.
induction H2 as [| h H2 IH].
(* nil *)
simpl in Hso; auto; done.

(* cons *)
simpl in Htrco, Hso.
apply IH; [|].
  prove_trco.

  eapply so_po_tail; eauto; [].
  apply po_cons_incl; eauto; [].
  apply po_app_incl; eauto; done.
Qed.


Lemma sos_po_tail :
  forall b H h i1 i2 i3,
    trco (H; h)
    -> po (H; h) i2 i3
    -> star (so b (H; h)) i1 i2
    -> star (so b H) i1 i2.
Proof.
intros b H h i1 i2 i3 Htrco Hpo Hsos.
so (star_starr _#4 Hsos) as Hsos'; renameover Hsos' into Hsos.
revert i3 Hpo.
elim Hsos; clear i1 i2 Hsos.
(* refl *)
intros; eapply star_refl; eauto; done.

(* step *)
intros i1 i2 i3 _ IH Hso23 i4 Hpo.
assert (so b H i2 i3) as Hso23'.
  eapply so_po_tail; eauto; done.
apply (star_trans _#3 i2).
  apply (IH i3); [].
  eapply so_po; eauto; done.

  apply star_one.
  assumption.
Qed.


Lemma sop_po_tail :
  forall b H h i1 i2 i3,
    trco (H; h)
    -> po (H; h) i2 i3
    -> sop b (H; h) i1 i2
    -> sop b H i1 i2.
Proof.
intros b H h i1 i2 i3 Htrco Hpo Hsop.
so (plus_plusri _#4 Hsop) as Hsop'; renameover Hsop' into Hsop.
revert i3 Hpo.
elim Hsop; clear i1 i2 Hsop.
(* refl *)
intros i1 i2 Hso i3 Hpo.
apply plus_one.
eapply so_po_tail; eauto; done.

(* step *)
intros i1 i2 i3 _ IH Hso23 i4 Hpo.
apply plusr_plus.
assert (so b H i2 i3) as Hso23'.
  eapply so_po_tail; eauto; done.
exists i2; split.
  apply plus_star.
  apply (IH i3); [].
  eapply so_po; eauto; done.

  assumption.
Qed.


Lemma xo_po_truncate :
  forall H H' i1 i2 i3,
    trco (H;; H')
    -> po H i2 i3
    -> xo (H;; H') i1 i2
    -> xo H i1 i2.
Proof.
intros H H' i1 i2 i3 Htrco Hpo23 Hxo.
invert Hxo.
intros b Hso.
eapply xo_specified; [].
eapply so_po_truncate; eauto; done.
Qed.


Lemma xop_po_truncate :
  forall H H' i1 i2 i3,
    trco (H;; H')
    -> po H i2 i3
    -> xop (H;; H') i1 i2
    -> xop H i1 i2.
Proof.
intros H H' i1 i2 i3 Htrco Hpo Hxop.
so (plus_plusi _#4 Hxop) as Hxop'; clear Hxop.
revert i3 Hpo.
induct Hxop'.

(* one *)
intros.
apply plus_one; [].
eapply xo_po_truncate; eauto; done.

(* step *)
intros i1 i2 i3 H12 H23 IH i4 Hpo34.
apply (plus_trans _#3 i2); [|].
  apply plus_one; [].
  eapply (xo_po_truncate _#4 i3); eauto; [].
  eapply (po_po_truncate _#4 i4); eauto; [].
  apply xop_po; auto; [].
  apply plusi_plus; auto; done.

  eapply IH; eauto; done.
Qed.


Lemma so_rf_tail :
  forall b H h i1 i2 i3,
    trco (H; h)
    -> rf H i2 i3
    -> so b (H; h) i1 i2
    -> so b H i1 i2.
Proof.
intros b H h i1 i2 i3 Htrco Hrf Hso.
eapply so_execed_tail; eauto; [].
eapply rf_execed_fst; eauto; prove_trco.
Qed.


Lemma so_rf_truncate :
  forall b H H' i1 i2 i3,
    trco (H;; H')
    -> rf H i2 i3
    -> so b (H;; H') i1 i2
    -> so b H i1 i2.
Proof.
intros b H1 H2 i1 i2 i3 Htrco Hrf Hso.
induction H2 as [| h H2 IH].
(* nil *)
simpl in Hso; auto; done.

(* cons *)
simpl in Htrco, Hso.
apply IH; [|].
  prove_trco.

  eapply so_rf_tail; eauto; [].
  apply in_or_app; right.
  eassumption.
Qed.


Lemma sop_rf_tail :
  forall b H h i1 i2 i3,
    trco (H; h)
    -> rf H i2 i3
    -> sop b (H; h) i1 i2
    -> sop b H i1 i2.
Proof.
intros b H h i1 i3 i4 Htrco Hrf Hsop.
so (plus_plusr _#4 Hsop) as (i2 & Hsop12 & Hso23).
apply plusr_plus.
assert (so b H i2 i3) as Hso23'.
  eapply so_rf_tail; eauto; done.
exists i2; split.
  eapply sos_po_tail; eauto; [].
  eapply so_po; eauto; done.

  assumption.
Qed.


Lemma sop_rf_truncate :
  forall b H H' i1 i2 i3,
    trco (H;; H')
    -> rf H i2 i3
    -> sop b (H;; H') i1 i2
    -> sop b H i1 i2.
Proof.
intros b H1 H2 i1 i2 i3 Htrco Hrf Hsop.
induction H2 as [| h H2 IH].
(* nil *)
simpl in Hsop; auto; done.

(* cons *)
simpl in Htrco, Hsop.
apply IH; [|].
  prove_trco.

  eapply sop_rf_tail; eauto; [].
  apply in_or_app; right.
  eassumption.
Qed.


Lemma rwf_tail :
  forall H h i1 i2,
    ~ (exists i a, h = ev_is i a)
    -> ~ (exists i i', h = ev_rf i i')
    -> rwf (H; h) i1 i2
    -> rwf H i1 i2.
Proof.
intros H h i1 i2 Hnis Hnrf Hrwf.
destruct Hrwf as (Hrf & (l & v & His)).
destruct Hrf as [Heq | Hrf].
  subst; destruct Hnrf; eauto; done.

  destruct His as [Heq | His].
    subst; destruct Hnis; eauto; done.

    unfold rwf; eauto.
Qed.


Lemma rwf_truncate :
  forall H H' i1 i2,
    ~ H' {{ ev_is i a | i a }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> rwf (H;; H') i1 i2
    -> rwf H i1 i2.
Proof.
intros H H' i1 i2 Hnis Hnrf Hrwf.
induction H'.

(* nil *)
assumption.

(* cons *)
#3 apply IHH'.
  contradict Hnis.
  destruct_all Hnis.
  repeat eexists; right; eassumption.

  contradict Hnrf.
  destruct_all Hnrf.
  repeat eexists; right; eassumption.

  autorewrite with canonlist in Hrwf.
  #2 eapply rwf_tail; eauto.
    contradict Hnis.
    destruct_all Hnis.
    subst.
    repeat eexists; left; reflexivity.

    contradict Hnrf.
    destruct_all Hnrf.
    subst.
    repeat eexists; left; reflexivity.
Qed.


Lemma rwfs_truncate :
  forall H H' i1 i2,
    ~ H' {{ ev_is i a | i a }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> rwfs (H;; H') i1 i2
    -> rwfs H i1 i2.
Proof.
intros H H' i1 i2.
intros.
apply (star_mono _ (rwf (H;; H'))); eauto using rwf_truncate.
Qed.


Lemma to_cons_invert :
  forall H h i1 i2,
    trco (H; h)
    -> to (H; h) i1 i2
    -> to H i1 i2
       \/ (executes h i1 /\ ~ exec H i2).
Proof.
intros H h i1 i2 Htrco Hto.
destruct Hto as (H1 & H2 & HeqHh & Hexec & Hnexec).
so (app_eq_cons _#5 HeqHh) as [(H2' & _ & HeqH) | (_ & HeqH1)].
  left.
  exists H1, H2'.
  repeat2 eexists; eauto; done.

  rewrite -> HeqH1 in Hexec.
  so (exec_invert _ _ Hexec) as (h' & Hin & Hexecutes).
  destruct Hin as [Heqh | Hin].
    subst h'.
    right.
    split; auto; [].
    contradict Hnexec.
    rewrite -> HeqH1.
    apply exec_cons_incl; assumption.

    left.
    apply to_from_nexec; [|].
      eapply executes_exec; eauto; done.

      contradict Hnexec.
      rewrite -> HeqH1.
      apply exec_cons_incl; assumption.
Qed.


Lemma pio_cons_invert :
  forall H h i1 i2,
    trco (H; h)
    -> pio (H; h) i1 i2
    -> pio H i1 i2 \/
       (h = ev_exec i1 /\ H {{ ev_is i1 ac_push }} /\ ~ exec H i2).
Proof.
intros H h i1 i2 Htrco Hpio.
destruct Hpio as (His1 & Hto).
destruct His1 as [Heq | His1].
  subst h.
  destruct (trco_is_invert_nexec _#3 Htrco); [].
  so (to_execed_fst _#3 Hto) as Hexec.
  eapply exec_tail; eauto; discform.

  so (to_cons_invert _#4 Htrco Hto) as [Hto' | (Heqh & Hnexec)].
    left.
    split; auto; done.
    
    right.
    split; auto; [].
    invert Heqh; auto; [].
    intros (iw & ->).
    exfalso.
    so (rf_isw_isr _#3 Htrco (or_introl (eq_refl _))) as (_ & Hisr).
    eapply isr_isp_contra; eauto; [].
    right; assumption.
Qed.



(* can truncate: init, is, co, rf, edge, fence *)
Lemma pio_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i, h = ev_exec i)
    -> pio (H; h) i1 i2
    -> pio H i1 i2.
Proof.
intros H h i1 i2 Htrco Hnexec Hpio.
so (pio_cons_invert _#4 Htrco Hpio) as [Hpio' | (Heq & _)].
  assumption.

  destruct Hnexec.
  subst.
  eexists; reflexivity.
Qed.


Lemma pio_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_exec i | i }}
    -> pio (H;; H') i1 i2
    -> pio H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hnexec Hpio.
induction H' as [| h H' IH].

(* nil *)
auto; done.

(* cons *)
autorewrite with canonlist in *.
apply IH; [| |].
  eapply trco_tail; eauto; done.

  contradict Hnexec.
  destruct Hnexec.
  eexists.
  right; eassumption.

  eapply pio_tail; eauto; [].
  contradict Hnexec.
  destruct Hnexec.
  eexists.
  left; eassumption.
Qed.


Lemma pio_execed_tail :
  forall H h i1 i2,
    trco (H; h)
    -> exec H i2
    -> pio (H; h) i1 i2
    -> pio H i1 i2.
Proof.
intros H h i1 i2 Htrco Hexec Hpio.
destruct Hpio as (Hpush & Hto).
thus (to H i1 i2) as Hto' by to_execed_tail.
split.
  destruct Hpush as [Heq | ?]; auto; [].
  subst h.
  destruct (trco_is_invert_nexec _#3 Htrco).
  eapply to_execed_fst; eauto; done.

  exact Hto'.
Qed.


(* can truncate is, edge *)
Lemma vo_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> vo (H; h) i1 i2
    -> vo H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvo.
destruct Hvo as [Hso | [Hrf | Hpio]].
  left.
  eapply so_tail; eauto; done.

  right; left.
  destruct Hrf.
    destruct Hnrf; eauto; done.

    assumption.

  right; right.
  eapply pio_tail; eauto.
Qed.


Lemma vo_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_exec i | i }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> vo (H;; H') i1 i2
    -> vo H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvo.
induction H' as [| h H' IH].
  assumption.

  apply IH;
  try (intro HH; destruct_all HH;
       (match goal with
          HHn : ~ _ |- _ => destruct HHn; repeat eexists; right; eassumption
        end)); [|].
    eapply trco_tail; eauto; done.

    apply (vo_tail _ h); auto;
    try (intro HH; destruct_all HH;
         (match goal with
            HHn : ~ _ |- _ => destruct HHn; repeat eexists; left; eassumption
          end)).
Qed.


Lemma vop_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> vop (H; h) i1 i2
    -> vop H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvop.
elim (plus_plusi _#4 Hvop); clear i1 i2 Hvop.
(* one *)
intros i1 i2 Hvo.
apply plus_one.
eapply vo_tail; eauto; done.

(* trans *)
intros i1 i2 i3 Hvo _ IH.
apply (plus_trans _#3 i2).
  apply plus_one.
  eapply vo_tail; eauto; done.

  assumption.
Qed.


Lemma vop_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_exec i | i }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> vop (H;; H') i1 i2
    -> vop H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvop.
elim (plus_plusi _#4 Hvop); clear i1 i2 Hvop.
(* one *)
intros i1 i2 Hvo.
apply plus_one.
eapply vo_truncate; eauto; done.

(* trans *)
intros i1 i2 i3 Hvo _ IH.
apply (plus_trans _#3 i2).
  apply plus_one.
  eapply vo_truncate; eauto; done.

  assumption.
Qed.


Lemma vos_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> vos (H; h) i1 i2
    -> vos H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvos.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst i2.
  apply star_refl; done.

  apply plus_star; [].
  eapply vop_tail; eauto; done.
Qed.


Lemma vos_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_exec i | i }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> vos (H;; H') i1 i2
    -> vos H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvos.
elim Hvos; clear i1 i2 Hvos.
(* refl *)
intro i.
apply star_refl; done.

(* trans *)
intros i1 i2 i3 Hvo _ IH.
apply (star_trans _#3 i2).
  apply star_one.
  eapply vo_truncate; eauto; done.

  assumption.
Qed.


(* can truncate: exec, rf, edge *)
Lemma xo_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists f i, h = ev_fence f i)
    -> xo (H; h) i1 i2
    -> xo H i1 i2.
Proof.
intros H h i1 i2 Htrco Hneqinit Hneqfence Hxo.
invert Hxo; [].
eauto using @xo_specified, @so_tail; done.
Qed.


(* can truncate: exec, rf, edge *)
Lemma xo_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> xo (H;; H') i1 i2
    -> xo H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hninit Hnfence Hxo.
induction H' as [| h H' IH].
  assumption.

  simpl in Htrco.
  apply IH;
  try (intro HH; destruct_all HH;
       (match goal with
          HHn : ~ _ |- _ => destruct HHn; repeat eexists; right; eassumption
        end));
  try prove_trco; [].
  apply (xo_tail _ h);
  try (intro HH; destruct_all HH;
       (match goal with
          HHn : ~ _ |- _ => destruct HHn; repeat eexists; left; eassumption
        end));
  try prove_trco; done.
Qed.


Lemma xos_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists f i, h = ev_fence f i)
    -> xos (H; h) i1 i2
    -> xos H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnfence Hxos.
apply (star_mono _ (xo (H; h))); auto; [].
intros; eapply xo_tail; eauto.
Qed.


(* Executedness *)

Lemma so_xo :
  forall H b i i',
    so b H i i'
    -> xo H i i'.
Proof.
intros H b i i' Hso.
eapply xo_specified; eauto; done.
Qed.


Lemma so_execed_2 :
  forall H b i i',
    trco H
    -> so b H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H b i1 i2 Htrco Hso Hexec2.
so (exec_invert _ _ Hexec2) as (h & Hin & Hexecutes).
so (in_split _#3 Hin) as (H2 & H1 & HeqH).
subst H.
so (trco_executes_invert _#3 (trco_truncate _ _ Htrco) Hexecutes) as (_ & _ & _ & Hprior).
apply exec_app_incl; apply exec_cons_incl; [].
apply Hprior; [].
apply (so_xo _ b); [].
apply (so_tail _ h); try discform; [prove_trco |].
eapply so_execed_truncate; eauto; [].
eapply executes_exec; eauto; [].
left; reflexivity.
Qed.


Lemma executable_app :
  forall H H' i,
    trco (H;; H')
    -> ~ H' {{ ev_init i p | i p }}
    -> ~ H' {{ ev_is i a | i a }}
    -> ~ H' {{ ev_exec i | i }}
    -> ~ H' {{ ev_rf i i' | i i' }}
    -> ~ H' {{ ev_fence f i | f i }}
    -> executable H i
    -> executable (H;; H') i.
Proof.
intros H H' i Htrco Hnoinit Hnois Hnoexec Hnorf Hnofence (Hnexec & Hpriors).
split.
  intro Hexec.
  destruct Hnexec.
  so (exec_invert _ _ Hexec) as (h & Hin & Hexecutes).
  so (in_app_or _#4 Hin) as [Hin' | Hin'].
    destruct Hexecutes as [-> | (? & ->)].
      destruct Hnoexec.
      eexists; eassumption.

      destruct Hnorf.
      repeat eexists; eassumption.

    eapply executes_exec; eauto; done.

  intros i' Hxo.
  apply exec_app_incl; [].
  apply Hpriors.
  eapply xo_truncate; eauto; done.
Qed.


(* Imposing trace order *)

Lemma so_to :
  forall H b i i',
    trco H
    -> so b H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H b i1 i2 Htrco Hso Hexec2.
so (exec_invert _ _ Hexec2) as (h & Hin & Hexecutes).
so (in_split _#3 Hin) as (H2 & H1 & Heq).
so (trco_executes_invert _#3 (trco_truncate _ _ (eqconv Heq Htrco)) Hexecutes) as (_ & _ & Hnexec & Hpriors).
exists H1, (nil; h;; H2).
autorewrite with canonlist.
repeat2 split; auto; [].
apply Hpriors; [].
apply (@xo_specified b); [].
apply (so_tail _ h); try discform; [prove_trco |].
apply (so_execed_truncate _ H2); subst H; auto; [].
eapply executes_exec; eauto; [].
left; reflexivity.
Qed.


Lemma so_to' :
  forall H b i i',
    trco H
    -> so b H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros H b i1 i2 Htrco Hso Hexec1.
so (exec_decidable H i2) as [Hexec2 | Hnexec2].
  eapply so_to; eauto; done.

  apply to_from_nexec; auto; done.
Qed.


Lemma sop_to :
  forall b H i i',
    trco H
    -> sop b H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros b H i i' Htrco Hsop Hexec.
pose proof (plus_plusi _#4 Hsop) as Hsop'; clear Hsop.
revert Hexec.
elim Hsop'; clear i i' Hsop'.
(* one *)
intros i i' Hso Hexec.
eapply so_to; eauto; done.

(* step *)
intros i1 i2 i3 Hso _ IH Hexec.
pose proof (IH Hexec) as Hto23.
apply (to_trans _ _ i2); auto.
eapply so_to; eauto.
apply (to_execed_fst H i2 i3); auto.
Qed.


Lemma sop_to' :
  forall b H i i',
    trco H
    -> sop b H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros b H i1 i2 Htrco Hsop Hexec1.
so (exec_decidable H i2) as [Hexec2 | Hnexec2].
  eapply sop_to; eauto; done.

  apply to_from_nexec; auto; done.
Qed.


Lemma sop_execed_truncate :
  forall H1 H2 b i1 i2,
    trco (H1;; H2)
    -> exec H1 i2
    -> sop b (H1;; H2) i1 i2
    -> sop b H1 i1 i2.
Proof.
intros H1 H2 b i1 i2 Htrco Hexec Hsop.
pose proof (plus_plusi _#4 Hsop) as Hsop'; clear Hsop.
revert Hexec.
elim Hsop'; clear i1 i2 Hsop'.
(* one *)
intros.
apply plus_one.
eapply so_execed_truncate; eauto.

(* step *)
intros i1 i2 i3 Hso _ IH Hexec.
exists i2; split.
  eapply so_execed_truncate; eauto.
  apply (to_execed_fst H1 i2 i3).
  eapply sop_to; eauto.
  eapply trco_truncate; eauto.

  apply plus_star.
  apply IH; auto.
Qed.


Lemma sop_execed_2 :
  forall b H i i',
    trco H
    -> sop b H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros b H i i' Htrco Hso Hexec.
assert (to H i i')  as Hto by (eapply sop_to; eauto).
eapply to_execed_fst; eauto; done.
Qed.


Lemma sop_vop :
  forall H i1 i2,
    sop at_v H i1 i2
    -> vop H i1 i2.
Proof.
intros H i1 i2 Hsop.
pose proof (plus_plusi _#4 Hsop) as Hsop'; clear Hsop.
elim Hsop'; clear i1 i2 Hsop'.
(* one *)
intros.
apply plus_one.
left; auto.

(* step *)
intros i1 i2 i3 Hso _ IH.
exists i2; split.
  left; auto.

  apply plus_star; auto.
Qed.


Lemma xo_execed_tail :
  forall H h i1 i2,
    trco (H; h)
    -> exec H i2
    -> xo (H; h) i1 i2
    -> xo H i1 i2.
Proof.
intros H h i1 i2 Htrco Hexec Hxo.
invert Hxo; [].
intros; eauto using @xo_specified, @so_execed_tail; done.
Qed.


Lemma xo_execed_truncate :
  forall H H' i1 i2,
    trco (H;; H')
    -> exec H i2
    -> xo (H;; H') i1 i2
    -> xo H i1 i2.
Proof.
intros H H' i1 i2 Htrco Hexec Hxo.
induction H' as [| h H'].
  assumption.

  apply IHH'.
    simpl in Htrco.
    eapply trco_tail; eauto.

    simpl in Hxo.
    eapply xo_execed_tail; eauto; [].
    apply exec_app_incl; auto; done.
Qed.


Lemma xo_to :
  forall H i i',
    trco H
    -> xo H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H i1 i2 Htrco Hxo Hexec2.
so (exec_invert _ _ Hexec2) as (h & Hin & Hexecutes).
so (in_split _#3 Hin) as (H2 & H1 & ->).
eapply to_from_before; eauto; [].
so (trco_executes_invert _#3 (trco_truncate _ _ Htrco) Hexecutes) as (_ & _ & _ & Hpriors).
apply Hpriors; [].
apply (xo_tail _ h); try discform; [prove_trco |].
apply (xo_execed_truncate _ H2); auto; [].
eapply executes_exec; eauto; [].
left; reflexivity.
Qed.


Lemma xo_to' :
  forall H i i',
    trco H
    -> xo H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros H i i' Htrco Hxo Hexec.
so (exec_decidable H i') as [Hexec' | Hnexec].
  apply xo_to; auto; done.

  apply to_from_nexec; auto; done.
Qed.


Lemma xop_to :
  forall H i i',
    trco H
    -> plus (xo H) i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H i i' Htrco Hxop Hexec.
so (plus_plusri _#4 Hxop) as Hxop'; renameover Hxop' into Hxop.
induction Hxop as [i1 i2 Hxo | i1 i2 i3 _ IH Hxo].

(* one *)
apply xo_to; auto; done.

(* step *)
assert (to H i2 i3) as Hto23 by (apply xo_to; auto).
apply (to_trans _ _ i2); auto; [].
apply IH; [].
eapply to_execed_fst; eauto; done.  
Qed.


Lemma xop_to' :
  forall H i i',
    trco H
    -> xop H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros H i i' Htrco Hxop Hexec.
so (exec_decidable H i') as [Hexec' | Hnexec].
  apply xop_to; auto; done.

  apply to_from_nexec; auto; done.
Qed.


Lemma xos_toq :
  forall H i i',
    trco H
    -> star (xo H) i i'
    -> exec H i'
    -> toq H i i'.
Proof.
intros H i i' Htrco Hxos Hexec.
so (star_plus _#4 Hxos) as [Heq | Hxop].
  subst i'.
  left; reflexivity.

  right.
  apply xop_to; auto; done.
Qed.


Lemma xos_toq' :
  forall H i i',
    trco H
    -> star (xo H) i i'
    -> exec H i
    -> toq H i i'.
Proof.
intros H i i' Htrco Hxos Hexec.
so (exec_decidable H i') as [Hexec' | Hnexec].
  apply xos_toq; auto; done.

  right.
  apply to_from_nexec; auto; done.
Qed.


Lemma xo_execed_2 :
  forall H i i',
    trco H
    -> xo H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H i i' Htrco Hxo Hexeci'.
eapply to_execed_fst; [].
eapply xo_to; eauto; done.
Qed.


Lemma xop_execed_2 :
  forall H i i',
    trco H
    -> plus (xo H) i i'
    -> exec H i'
    -> exec H i.
Proof.
intros.
eapply to_execed_fst.
eapply xop_to; eauto; done.
Qed.


Lemma xos_execed_2 :
  forall H i i',
    trco H
    -> star (xo H) i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H i i' Htrco Hxos Hexeci'.
so (star_plus _#4 Hxos) as [Heq | Hxop].
  subst; assumption.

  eapply xop_execed_2; eauto; done.
Qed.


Lemma xos_execed_tail :
  forall H h i1 i2,
    trco (H; h)
    -> exec H i2
    -> xos (H; h) i1 i2
    -> xos H i1 i2.
Proof.
intros H h i1 i2 Htrco Hexec Hxos.
revert Hexec.
induct Hxos.

(* refl *)
intros; apply star_refl; done.

(* step *)
intros i1 i2 i3 Hxo _ IH Hexec3.
assert (xos H i2 i3) as Hxos by (auto using IH).
apply (@star_step _ _ _ i2); auto; [].
eapply xo_execed_tail; eauto; [].
eapply xos_execed_2; eauto; prove_trco.
Qed.


(* Visibility order *)

Lemma vo_to_or_svo :
  forall H i i',
    trco H
    -> vo H i i'
    -> to H i i' \/ so at_v H i i'.
Proof.
intros H i i' Htrco Hvo.
destruct Hvo as [Hsvo | [Hrf | Hpio]].
  right; assumption.

  left.
  apply rf_to; auto; done.

  left.
  apply pio_to; auto; done.
Qed.


Lemma vo_to :
  forall H i i',
    trco H
    -> vo H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H i i' Htrco Hvo Hexec.
so (vo_to_or_svo _#3 Htrco Hvo) as [| Hsvo]; auto; [].
eapply so_to; eauto; done.
Qed.


Lemma vop_to_or_svop :
  forall H i i',
    trco H
    -> vop H i i'
    -> to H i i' \/ sop at_v H i i'.
Proof.
intros H i i' Htrco Hvop.
so (plus_plusi _#4 Hvop) as Hvop'; clear Hvop.
induct Hvop'.

(* one *)
intros i1 i2 Hvo.
so (vo_to_or_svo _#3 Htrco Hvo) as [Hto | Hsvo].
  left.
  assumption.

  right.
  apply plus_one; auto; done.

(* step *)
intros i1 i2 i3 Hvo12 _ IH.
destruct IH as [Hto23 | Hsvop23].
  left.
  eapply to_trans; eauto; [].
  apply vo_to; auto; [].
  eapply to_execed_fst; eauto; done.

  so (vo_to_or_svo _#3 Htrco Hvo12) as [Hto12 | Hsvo12].
    left.
    so (exec_decidable H i3) as [Hexec3 | Hnexec3].
      eapply to_trans; eauto; [].
      eapply sop_to; eauto; done.

      apply to_from_nexec; auto; [].
      eapply to_execed_fst; eauto; done.
  
    right.
    exists i2; split; auto; [].
    apply plus_star; auto; done.
Qed.


Lemma vop_execed_fst_or_svop :
  forall H i i',
    trco H
    -> vop H i i'
    -> exec H i \/ sop at_v H i i'.
Proof.
intros H i i' Htrco Hvop.
so (vop_to_or_svop _#3 Htrco Hvop) as [Hto | Hsop].
  left.
  eapply to_execed_fst; eauto; done.

  right; assumption.
Qed.


Lemma vos_execed_fst_or_svos :
  forall H i i',
    trco H
    -> vos H i i'
    -> exec H i \/ star (so at_v H) i i'.
Proof.
intros H i i' Htrco Hvos.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst i'.
  right; apply star_refl; done.

  so (vop_execed_fst_or_svop _#3 Htrco Hvop) as [Hexeced | Hsvop]; auto using plus_star; done.
Qed.


Lemma vos_execed_fst_or_xos :
  forall H i i',
    trco H
    -> vos H i i'
    -> exec H i \/ xos H i i'.
Proof.
intros H i i' Htrco Hvos.
so (vos_execed_fst_or_svos _#3 Htrco Hvos) as [? | Hsvos]; auto; [].
right.
apply (star_mono _ (so at_v H)); eauto using @xo_specified; done.
Qed.


Lemma vop_to :
  forall H i i',
    trco H
    -> vop H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H i1 i2 Htrco Hvop Hexec2.
so (plus_plusi _#4 Hvop) as Hvop'; clear Hvop.
revert Hexec2.
induct Hvop'.

(* one *)
intros i1 i2 Hvo Hexec2.
so (vo_to_or_svo _#3 Htrco Hvo) as [Hto | Hsvo].
  assumption.

  eapply so_to; eauto; done.

(* step *)
intros i1 i2 i3 Hvo12 _ IH Hexec3.
so (IH Hexec3) as Hto23.
eapply to_trans; eauto; [].
so (vo_to_or_svo _#3 Htrco Hvo12) as [Hto12 | Hsvo12].
  assumption.

  eapply so_to; eauto; [].
  eapply to_execed_fst; eauto; done.
Qed.


Lemma vop_to' :
  forall H i i',
    trco H
    -> vop H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros H i i' Htrco Hvop Hexec.
so (exec_decidable H i') as [Hexec' | Hnexec].
  apply vop_to; auto; done.

  apply to_from_nexec; auto; done.
Qed.


Lemma vos_toq :
  forall H i i',
    trco H
    -> vos H i i'
    -> exec H i'
    -> toq H i i'.
Proof.
intros H i i' Htrco Hvos Hexec.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst; left; reflexivity.

  right.
  eapply vop_to; eauto; done.
Qed.


Lemma vop_irreflex :
  forall H i,
    trco H
    -> ~ vop H i i.
Proof.
intros H i Htrco Hvop.
so (vop_execed_fst_or_svop _#3 Htrco Hvop) as [Hexec | Hsvop].
  destruct (to_irreflex H i).
  apply vop_to; auto; done.

  destruct (po_irreflex _ i Htrco).
  eapply sop_po; eauto; done.
Qed.


Lemma vo_irreflex :
  forall H i,
    trco H
    -> ~ vo H i i.
Proof.
intros H i Htrco Hvo.
edestruct vop_irreflex; eauto.
apply plus_one; eauto.
Qed.


Lemma vos_to_trans :
  forall H i1 i2 i3,
    trco H
    -> vos H i1 i2
    -> to H i2 i3
    -> to H i1 i3.
Proof.
intros H i1 i2 i3 Htrco Hvos Hto23.
eapply toq_to_trans; eauto; [].
apply vos_toq; auto; [].
eapply to_execed_fst; eauto; done.
Qed.


Lemma vop_execed_2 :
  forall H i i',
    trco H
    -> vop H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H i i' Htrco Hvop Hexec.
eapply to_execed_fst; eauto; [].
eapply vop_to; eauto; done.
Qed.


Lemma vos_execed_2 :
  forall H i i',
    trco H
    -> vos H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H i i' Htrco Hvos Hexec.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst i'.
  assumption.

  eapply vop_execed_2; eauto; done.
Qed.


Lemma vo_execed_tail :
  forall H h i i',
    trco (H; h)
    -> exec H i'
    -> vo (H; h) i i'
    -> vo H i i'.
Proof.
intros H h i i' Htrco Hexec Hvo.
destruct Hvo as [Hsvo | [Hrf | Hpio]].
  left.
  eapply so_execed_tail; eauto; done.

  right; left.
  destruct Hrf as [Heq | Hrf].
    subst h.
    invert Htrco.
    intros _ _ _ _ _ Hnexec _.
    contradiction.

    assumption.

  right; right.
  eapply pio_execed_tail; eauto; done.
Qed.


Lemma vo_execed_truncate :
  forall H H' i i',
    trco (H;; H')
    -> exec H i'
    -> vo (H;; H') i i'
    -> vo H i i'.
Proof.
intros H1 H2 i i' Htrco Hexec Hvo.
induction H2 as [| h H2 IH].

(* nil *)
assumption.

(* cons *)
autorewrite with canonlist in *.
apply IH; [prove_trco |].
eapply vo_execed_tail; eauto; [].
apply exec_app_incl; auto; done.
Qed.


Lemma vop_execed_tail :
  forall H h i i',
    trco (H; h)
    -> exec H i'
    -> vop (H; h) i i'
    -> vop H i i'.
Proof.
intros H h i i' Htrco Hexec Hvop.
so (plus_plusi _#4 Hvop) as Hvop'; clear Hvop.
revert Hexec.
induct Hvop'.

(* one *)
intros.
apply plus_one; [].
eapply vo_execed_tail; eauto; done.

(* step *)
intros i1 i2 i3 H12 _ IH Hexec3.
so (IH Hexec3) as Hvop23.
assert (exec H i2) as Hexec2 by (eapply vop_execed_2; eauto; prove_trco).
exists i2.
split.
  eapply vo_execed_tail; eauto; done.

  apply plus_star; [].
  apply IH; auto; done.
Qed.


Lemma vos_execed_tail :
  forall H h i i',
    trco (H; h)
    -> exec H i'
    -> vos (H; h) i i'
    -> vos H i i'.
Proof.
intros H h i i' Htrco Hexec Hvos.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst; apply star_refl; done.

  apply plus_star; [].
  eapply vop_execed_tail; eauto; done.
Qed.


Lemma vop_execed_truncate :
  forall H H' i i',
    trco (H;; H')
    -> exec H i'
    -> vop (H;; H') i i'
    -> vop H i i'.
Proof.
intros H1 H2 i i' Htrco Hexec Hvop.
so (plus_plusi _#4 Hvop) as Hvop'; clear Hvop.
revert Hexec.
induct Hvop'.

(* one *)
intros.
apply plus_one; [].
eapply vo_execed_truncate; eauto; done.

(* step *)
intros i1 i2 i3 H12 _ IH Hexec.
exists i2.
split.
  eapply vo_execed_truncate; eauto; [].
  eapply vop_execed_2; eauto; prove_trco.

  apply plus_star; [].
  apply IH; auto; done.
Qed.


Lemma vos_execed_truncate :
  forall H H' i i',
    trco (H;; H')
    -> exec H i'
    -> vos (H;; H') i i'
    -> vos H i i'.
Proof.
intros H1 H2 i i' Htrco Hexec Hvos.
so (star_plus _#4 Hvos) as [Heq | Hvop].
  subst; apply star_refl; done.

  apply plus_star; [].
  eapply vop_execed_truncate; eauto; done.
Qed.


Lemma vop_extend_exec_nisp_xop :
  forall H h i i',
    trco (H; h)
    -> executes h i
    -> vop (H; h) i i'
    -> ~ isp H i
    -> xop H i i'.
Proof.
intros H h i1 i3 Htrco Hexecutes Hvop Hnisp.
destruct Hvop as (i2 & Hvo12 & Hvos23).
exists i2; split.
  destruct Hvo12 as [Hsvo | [Hrf | Hpio]].
    eapply xo_specified; [].
    eapply so_tail; eauto; discform.

    exfalso.
    so (trco_executes_invert _#3 Htrco Hexecutes) as (_ & _ & Hnexec & _).
    destruct Hnexec.
    destruct Hrf as [Heq | Hrf].
      subst h.
      invert Htrco.
      intros; assumption.

      eapply rf_execed_fst; eauto; prove_trco.

    exfalso.
    destruct Hpio as (Hisp & _).
    discriminate_in Hisp; intro Hisp; [|].
      discriminate_executes Hexecutes; done.
    contradiction.

  so (vos_execed_fst_or_xos _#3 Htrco Hvos23) as [Hexec2 | Hxos23].
    so (exec_cons_invert _#3 Hexec2) as [Hexecutes' | Hexec2'].
      so (executes_fun_1 _#3 Hexecutes Hexecutes'); subst i2.
      edestruct vo_irreflex; eauto; done.

      exploit (vop_to (H; h) i1 i2) as Hto; auto; [|].
        apply plus_one; assumption.

        edestruct to_cons_nexec; eauto; done.

    eapply xos_tail; eauto; discform.
Qed.


(* Visible-to *)

(* can truncate: is, edge *)
Lemma vt_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> vt (H; h) i1 i2
    -> vt H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnexec Hnrf Hnfence Hvt.
destruct Hvt as (i & Hvop & Hxos).
exists i; split.
  eapply vop_tail; eauto; done.

  eapply xos_tail; eauto; done.
Qed.


Lemma vt_execed_tail :
  forall H h i1 i2,
    trco (H; h)
    -> exec H i2
    -> vt (H; h) i1 i2
    -> vt H i1 i2.
Proof.
intros H h i1 i3 Htrco Hexec3 Hvt.
destruct Hvt as (i2 & Hvop & Hxos).
exists i2; split.
  eapply vop_execed_tail; eauto; [].
  so (star_plus _#4 Hxos) as [Heq | Hxop].
    subst i2; assumption.
  eapply (to_front_half _ _ (nil; h)); eauto; [].
  autorewrite with canonlist.
  apply xop_to; auto; [].
  apply exec_cons_incl; assumption.

  eapply xos_execed_tail; eauto; done.
Qed.


Lemma vt_execed_2 :
  forall H i i',
    trco H
    -> vt H i i'
    -> exec H i'
    -> exec H i.
Proof.
intros H i i' Htrco Hvt Hexec.
destruct Hvt as (i'' & Hvop & Hxos).
eapply vop_execed_2; eauto; [].
eapply xos_execed_2; eauto; done.
Qed.


Lemma vt_to_or_xop :
  forall H i i',
    trco H
    -> vt H i i'
    -> to H i i' \/ xop H i i'.
Proof.
intros H i1 i3 Htrco Hvt.
destruct Hvt as (i2 & Hvop & Hxos).
so (vop_to_or_svop _#3 Htrco Hvop) as [Hto | Hsvop].
  left.
  so (exec_decidable H i3) as [Hexec | Hnexec].
    eapply to_toq_trans; eauto; [].
    apply xos_toq; auto; done.

    apply to_from_nexec; auto; [].
    eapply to_execed_fst; eauto; done.

  right.
  eapply plus_star_trans; eauto; [].
  apply (plus_mono _ (so at_v H)); auto; [].
  eauto using xo_specified; done.
Qed.


Lemma vt_execed_fst_or_xop :
  forall H i i',
    trco H
    -> vt H i i'
    -> exec H i \/ xop H i i'.
Proof.
intros H i1 i3 Htrco Hvt.
so (vt_to_or_xop _#3 Htrco Hvt) as [Hto | Hxop]; eauto using to_execed_fst.
Qed.


Lemma vt_to :
  forall H i i',
    trco H
    -> vt H i i'
    -> exec H i'
    -> to H i i'.
Proof.
intros H i i' Htrco Hvt Hexec.
so (vt_to_or_xop _#3 Htrco Hvt) as [Hto | Hxop].
  assumption.

  apply xop_to; auto; done.
Qed.


Lemma vt_to' :
  forall H i i',
    trco H
    -> vt H i i'
    -> exec H i
    -> to H i i'.
Proof.
intros H i i' Htrco Hvt Hexec.
so (exec_decidable H i') as [Hexec' | Hnexec].
  apply vt_to; auto; done.

  apply to_from_nexec; auto; done.
Qed.


(* Coherence order *)

Lemma wrp_execed_fst_or_po :
  forall H i i',
    trco H
    -> wrp H i i'
    -> exec H i \/ po H i i'.
Proof.
intros H i i' Htrco Hwrp.
invert Hwrp; auto; [].
intros l Hvt _ _.
so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop]; auto; [].
right.
apply xop_po; auto; done.
Qed.


Lemma wwp_execed_fst_or_po :
  forall H i i',
    trco H
    -> wwp H i i'
    -> exec H i \/ po H i i'.
Proof.
intros H i i' Htrco Hwwp.
invert Hwwp; auto; [|].
  intros l Hvt _ _.
  so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop]; auto; [].
  right.
  apply xop_po; auto; done.

  intros ir Hwrp Hrwp.
  so (wrp_execed_fst_or_po _#3 Htrco Hwrp) as [Hexec | Hpo]; auto; [].
  right.
  eapply po_trans; eauto using rwp_po; done.
Qed.


(* can truncate: edge *)
Lemma wrp_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i a, h = ev_is i a)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> wrp (H; h) i1 i2
    -> wrp H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnis Hnexec Hnrf Hnfence Hwrp.
invert Hwrp.

(* vt *)
intros l Hvt Hwrite Hread.
eapply wrp_vt; eauto using vt_tail, writesto_tail, reads_tail.

(* po *)
intros l Hpo Hwrite Hread.
eapply wrp_po; eauto using writesto_tail, reads_tail; [].
eapply po_tail; eauto.
Qed.


Lemma rwp_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i a, h = ev_is i a)
    -> rwp (H; h) i1 i2
    -> rwp H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnis Hrwp.
destruct Hrwp as (l & Hpo & Hread & Hwrite).
exists l; repeat2 split; eauto using reads_tail, writesto_tail; [].
eapply po_tail; eauto.
Qed.


Lemma wwp_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i a, h = ev_is i a)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> wwp (H; h) i1 i2
    -> wwp H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnis Hnexec Hnrf Hnfence Hwwp.
invert Hwwp.

(* vt *)
intros l Hvt Hwrite Hwrite'.
eapply wwp_vt; eauto using vt_tail, writesto_tail; done.

(* po *)
intros l Hpo Hwrite Hwrite'.
eapply wwp_po; eauto using writesto_tail; [].
eapply po_tail; eauto.

(* read *)
intros ir Hwrp Hrwp.
eapply wwp_read; eauto using wrp_tail, rwp_tail; done.
Qed.


Lemma co_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i a, h = ev_is i a)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> ~ (exists i i', h = ev_co i i')
    -> co (H; h) i1 i2
    -> co H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnis Hnexec Hnrf Hnfence Hnco Hco.
remember (H; h) as H' eqn:HeqH' in Hco.
revert HeqH'.
induct Hco.

(* read *)
intros H' i1 i2 ir Hwrp Hrf Hneq ->.
eapply co_read; eauto using wrp_tail; [].
destruct Hrf as [Heqh | ?]; auto; [].
destruct Hnrf; eauto; done.

(* write *)
intros H' i1 i2 Hwwp ->.
eapply co_write; eauto using wwp_tail; done.

(* rw *)
intros H' i1 i2 iw Hrwf _ IH Hneq ->.
eapply co_rw; eauto using rwf_tail; done.

(* fiat *)
intros H' i i' Hin ->.
#2 destruct Hin.
  subst; destruct Hnco; eauto; done.

  apply co_fiat; auto; done.
Qed.


Lemma cop_tail :
  forall H h i1 i2,
    trco (H; h)
    -> ~ (exists i p, h = ev_init i p)
    -> ~ (exists i a, h = ev_is i a)
    -> ~ (exists i, h = ev_exec i)
    -> ~ (exists i i', h = ev_rf i i')
    -> ~ (exists f i, h = ev_fence f i)
    -> ~ (exists i i', h = ev_co i i')
    -> cop (H; h) i1 i2
    -> cop H i1 i2.
Proof.
intros H h i1 i2 Htrco Hninit Hnis Hnexec Hnrf Hnfence Hnco Hvop.
elim (plus_plusi _#4 Hvop); clear i1 i2 Hvop.
(* one *)
intros i1 i2 Hvo.
apply plus_one.
eapply co_tail; eauto; done.

(* trans *)
intros i1 i2 i3 Hco _ IH.
apply (plus_trans _#3 i2).
  apply plus_one.
  eapply co_tail; eauto; done.

  assumption.
Qed.


Lemma co_cons_invert :
  forall H ia ib i1 i2,
    trco (H; ev_co ia ib)
    -> co (H; ev_co ia ib) i1 i2
    -> co H i1 i2 \/ (rwfs H ia i1 /\ i2 = ib).
Proof.
intros H ia ib i1 i2 Htrco Hco.
remember (H; ev_co ia ib) as H' eqn:Heq.
revert Heq Htrco.
induct Hco.

(* read *)
intros H0 i i' ir H1 Hrf H3 -> Htrco.
left.
#2 apply (@co_read _ _ _ ir); auto.
  eapply wrp_tail; eauto; discform.

  unfold rf in Hrf.
  discriminate_in Hrf; auto; done.

(* write *)
intros H0 i i' H1 -> Htrco.
left.
apply co_write; [].
eapply wwp_tail; eauto; discform.

(* rw *)
intros H0 i i' iw Hrwf _ IH H4 -> Htrco.
so (IH (eq_refl _) Htrco) as [Hco | (Hrwfs & Heq)].
  left.
  eapply @co_rw; eauto; [].
  eapply rwf_tail; eauto; discform.

  right; split; auto; [].
  eapply star_trans; eauto; [].
  apply star_one; [].
  eapply rwf_tail; eauto; discform.

(* fiat *)
intros H0 i i' Hin -> Htrco.
destruct Hin as [Heq | Hin].
  injection Heq; intros; subst.
  right; split; auto; [].
  apply star_refl; done.

  left.
  apply co_fiat; auto; done.
Qed.


Lemma wrp_vt_co :
  forall H iw ir iw' l,
    trco H
    -> wrp H iw ir
    -> vt H ir iw'
    -> reads H ir l
    -> writesto H iw' l
    -> cop H iw iw'.
Proof.
intros H iw1 ir iw2 l Htrco Hwrp Hvt Hisr Hisw2.
so (vt_execed_fst_or_xop _#3 Htrco Hvt) as [Hexec | Hxop].
  so (exec_read_rf _ _ Htrco Hexec (eintro l Hisr)) as (iw & Hrf).
  apply (star_plus_trans _#3 iw); [|].
    eapply wrp_rf_cos; eauto; done.

    apply plus_one; [].
    apply co_write; [].
    so (rf_loc_2 _#4 Htrco Hrf Hisr) as Hisw1.
    eapply wwp_vt; eauto; [].
    eapply vos_vt_trans; eauto; [].
    apply star_one; [].
    right; left; assumption.

  apply plus_one; [].
  apply co_write; [].
  eapply wwp_read; eauto; [].
  eexists; eauto using xop_po; done.
Qed.
