
Require Import Tactics.
Require Import Sequence.
Require Import Dynamic.
Require Import Hwf.
Require Import Relation.
Require Import History.
Require Import Truncate.
Require Import Match.
Require Import Decide.


(* Visibility *)

Lemma visible_execed :
  forall H p i,
    trco H
    -> visible H p i
    -> H {{ ev_exec i }}.
Proof.
intros H p i Htrco Hvisible.
destruct Hvisible as (i' & Hvos & Hexec & _).
eapply vos_execed_2; eauto; done.
Qed.


Lemma visible_at_execed :
  forall H p i l,
    trco H
    -> visible_at H p i l
    -> H {{ ev_exec i }}.
Proof.
intros H p i l Htrco (? & _).
eapply visible_execed; eauto.
Qed.


Lemma visible_avail :
  forall H p i,
    trco H
    -> isw H i
    -> visible H p i
    -> H {{ ev_rel i }} \/ H {{ ev_init i p }}.
Proof.
intros H p i Htrco Hisw Hvis.
destruct Hvis as (i' & Hvos & Hexec & [Hinit | Hisp]).
  clear Hexec.
  revert Hinit.
  thus (vos H i i) as Hvos' by @star_refl.
  revert Hisw Hvos'.
  generalize i at 1 2 4.
  induct Hvos.

  (* refl *)
  intros i2 i1 Hisw Hvos Hinit.
  right; assumption.
  
  (* step *)
  intros i2 i3 i4 Hvo23 Hvos34 IH i1 Hisw1 Hvos12 Hinit4.
  assert (vos H i1 i3) as Hvos13.
    eapply star_trans; eauto; [].
    apply star_one; auto; done.
  so (IH i1 Hisw1 Hvos13 Hinit4) as [Hrel | Hinit3].
    left; assumption.
  destruct Hvo23 as [Hsvo | [Hrf | Hpio]].
    right.
    apply (po_init_2 _ _ i3); auto; [].
    eapply so_po; eauto; done.

    so (rf_rel_or_init _#4 Htrco Hinit3 Hrf) as [Hrel | Hinit2].
      left.
      apply (vos_rel_released _ _ i2); auto; done.

      right; assumption.

    destruct Hpio as (Hpush & Hto).
    so (to_execed_fst _#3 Hto) as Hexec2.
    left.
    eapply vos_push_released; eauto; done.

  left.
  eapply vos_push_released; eauto; done.
Qed.


(* Write visibility *)

(* If  i1 and i2 write to the same location
   and i1 is visible to i2 (ie, i2 is visible to i2's thread before i2 executes)
   and i2 is released
   then 
       i1 -co->+ i2  or  i2 -po-> i1
*)
Lemma visible_to_write :
  forall H i1 i2 l H1 H2 p,
    trco H
    -> H {{ ev_is i1 (ac_write l ve) | ve }}
    -> H {{ ev_is i2 (ac_write l ve) | ve }}
    -> H {{ ev_init i2 p }}
    -> H = H1; ev_exec i2;; H2
    -> H {{ ev_rel i2 }}
    -> visible H1 p i1
    -> cop H i1 i2 \/ po H i2 i1.
Proof.
intros H i1 i2 l H1 H2 p Htrco (v1 & Hwrite1) (v2 & Hwrite2) Hinit2 HeqH Hrel Hvis.
so (in_split _ _ (rel_split_right _#4 HeqH Htrco Hrel)) as (H4 & H3 & HeqH2).
subst H2.
autorewrite with canonlist in HeqH.
invert (trco_truncate _ _ (eqconv HeqH Htrco)).
intros p' l' _ Hinit2' (v2' & Hwrite2') _ Htrichot _ _ _ Hnrel.
assert (p = p').
  eapply (init_unique H); eauto; [].
  subst H.
  apply in_or_app; right; right; assumption.
subst p'; clear Hinit2'.
assert (ac_write l v2 = ac_write l' v2') as Heqwrite.
  eapply is_fun; eauto; [].
  subst H.
  apply in_or_app; right; right; assumption.
injection Heqwrite; intros; subst l'; clear dependent v2'.
assert (i1 <> i2) as Hneq.
  intro; subst i2.
  apply (trco_exec_invert_nexec H1 i1); [prove_trco |].
  eapply visible_execed; eauto; [].
  prove_trco.
exploit (Htrichot i1) as [Hcop | [Heq | Hcop]].
  split; [|].
    apply visible_app_incl; apply visible_cons_incl; assumption.
  exists v1.
  apply (is_execed_truncate _ (nil; ev_rel i2;; H4)); autorewrite with canonlist; rewrite <- ?HeqH; auto; [].
  apply in_or_app; right; right.
  eapply visible_execed; eauto; [].
  prove_trco.

  exploit (cop_rel_or_po (H1; ev_exec i2;; H3) i2 i1) as [Hrel' | Hpo]; [| | | | |].
    prove_trco.

    exists l, v2.
    apply (is_execed_truncate _ (nil; ev_rel i2;; H4)); autorewrite with canonlist; try prove_trco; [].
    apply in_or_app; right; left; reflexivity.

    apply in_or_app; right; left; reflexivity.

    assumption.

    contradiction.

    right.
    subst H.
    apply po_app_incl; apply po_cons_incl; assumption.

  destruct Hneq; auto; done.

  left.
  subst H.
  apply cop_app_incl; apply cop_cons_incl; assumption.
Qed.


Lemma rel_prior_cop_trichot :
  forall H H1 H2 i1 i2 p l,
    trco H
    -> H = H1; ev_rel i2;; H2
    -> H {{ ev_init i2 p }}
    -> H {{ ev_is i2 (ac_write l ve) | ve }}
    -> visible_at H1 p i1 l
    -> cop H1 i2 i1 \/ i1 = i2 \/ cop H1 i1 i2.
Proof.
intros H H1 H2 i1 i2 p l Htrco HeqH Hinit2 (v & His2) Hvisat.
invert (trco_truncate _ _ (eqconv HeqH Htrco)).
intros p' l' _ Hinit2' (v' & His2') _ Hvisatco _ _.
eassert Prop as Heqp; [refine (init_unique _#4 Htrco Hinit2 _) |].
  subst H.
  apply in_or_app; right; right; eassumption.
subst p'.
assert (ac_write l v = ac_write l' v') as Heqwrite.
  eapply is_fun; eauto; [].
  subst H.
  apply in_or_app; right; right; eassumption.
injection Heqwrite; clear Heqwrite; intros; subst l' v'.
so (Hvisatco _ Hvisat) as [? | [? | ?]]; auto; done.
Qed.


Theorem write_visibility :
  forall H l i1 i2 i3,
    trco H
    -> vos H i1 i2
    -> xop H i2 i3
    -> H {{ ev_is i1 (ac_write l ve) | ve }}
    -> H {{ ev_is i3 (ac_write l ve) | ve }}
    -> H {{ ev_rel i3 }}
    -> cop H i1 i3.
Proof.
intros H l i1 i2 i3 Htrco Hvos12 Hxop23 Hwrite1 Hwrite3 Hrel3.
so (po_init _#3 (xop_po _#3 Htrco Hxop23)) as (p & Hinit2 & Hinit3).
so (rel_execed _ _ Htrco Hrel3) as Hexec3.
so (in_split _#3 Hexec3) as (H2 & H1 & HeqH).
assert (H1 {{ ev_exec i2 }}) as Hexec2_3.
  eapply exec_tail_to; eauto; [].
  apply xop_to; auto; done.
assert (H1 {{ ev_exec i1 }}) as Hexec1_3.
  eapply exec_tail_to; eauto; [].
  apply (toq_to_trans _ _ i2); auto; [|].
    apply vos_toq; auto; [].
    eapply xop_execed_2; eauto; done.

    apply xop_to; auto; done.
assert (vos H1 i1 i2) as Hvos12_3.
  apply (vos_execed_tail _ (ev_exec i3)); auto; [| |].
    prove_trco.

    subst H.
    eapply vos_execed_truncate; eauto; done.

    discform.
exploit (visible_to_write H i1 i3 l H1 H2 p) as [? | Hpo31]; auto; [|].
  exists i2.
  do2 2 split; auto; [].
  left.
  subst H.
  exploit (init_execed_truncate (H1; ev_exec i3) H2 i2 p) as Hinit2'; auto; [|].
    right; assumption.

    discriminate_in Hinit2'; auto; done.

  exfalso.
  assert (trco H1) as HtrcoH1 by prove_trco.
  assert (H1 {{ ev_is i1 (ac_write l ve) | ve }}) as Hwrite1_3.
    destruct Hwrite1 as (v & His1).
    exists v.
    apply (is_execed_tail _ (ev_exec i3)); auto; [prove_trco |].
    subst H; eapply is_execed_truncate; eauto; [].
    right; assumption.
  so (vos_rel_or_poq _#3 HtrcoH1 (eintro l Hwrite1_3) Hvos12_3) as [Hrel1_3 | Hpoq12_3].
    apply (trco_exec_invert_nexec H1 i3); [prove_trco |].
    apply rel_execed; auto; [].
    eapply po_rel_released; eauto; [|].
      destruct Hwrite3 as (v & His3).
      exists v.
      cut ((H1; ev_exec i3) {{ ev_is i3 (ac_write l v) }}).
        intro His; discriminate_in His; auto; done.
      subst H; eapply is_execed_truncate; eauto; [].
      left; reflexivity.

      apply (po_tail _ (ev_exec i3)); [|].
        discform.
      subst H; eapply po_execed_truncate; eauto; [].
      right; assumption.

    apply (po_irreflex H i1); auto; [].
    eapply po_trans; eauto; [].
    apply (poq_po_trans _ _ i2); auto; [|].
      subst H; apply poq_app_incl; apply poq_cons_incl; assumption.

      apply xop_po; auto; done.
Qed.


(* Read visibility *)

Theorem read_visibility :
  forall H l i1 i2 i3 i,
    trco H
    -> vos H i1 i2
    -> xop H i2 i3
    -> H {{ ev_is i1 (ac_write l ve) | ve }}
    -> H {{ ev_is i3 (ac_read l) }}
    -> rf H i i3
    -> cos H i1 i.
Proof.
intros H l i1 i2 i3 i0 Htrco Hvos12 Hxop23 Hwrite1 Hread3 Hrf03.
(* Note that i2 -po-> i3 and i2 -to-> i3.
   Let p be the thread of i2 and i3.
   Since i1 -vo->* i2 and i2 -to-> i3, we have i1 -vo->* i2 when i3 executes.
   Thus i1 is visible to p when i3 executes.

   Therefore i1 -co->* i0  or  i3 -rw-> i4 -co->* i1.
   If the former we're done, so suppose the latter.
   Note that i1 and i4 write to the same location.
   Either i4 -po->? i1  or  H (rel i4).

     Suppose i4 -po->? i1.
     i4 cannot release until i2 executes
     Therefore, i1 cannot release until i2 executes.
     Since i1 -vo->* i2 without i1 released, i1 -po->? i2 (by vos_rel_or_poq).
     Hence i1 -po->? i2 -po-> i3 -po-> i4 -po->? i1.

     Suppose H (rel i4).
     i4 cannot release until i2 executes, at which time i1 is visible to p.
     Thus, when i4 releases, either i1 -co->+ i4  or  i1 = i4  or  i4 -co->+ i1.
       If i1 -co->+ i4 then i1 -co->+ i4 -co->* i1.

       If i1 = i4 then as above.

       If i4 -co->+ i1 before i4 releases, then i4 -po-> i1, so as above.
*)
so (xop_po _#3 Htrco Hxop23) as Hpo23.
thus (H {{ ev_exec i3 }}) as Hexec3 by rf_execed_snd.
thus (to H i2 i3) as Hto23 by xop_to.
so (po_init _#3 Hpo23) as (p & Hinit2 & Hinit3).
so (in_split _#3 Hexec3) as (H3a & H3b & Heq3).
so (rf_split _#5 Heq3 Htrco Hrf03) as (H3a' & ->).
autorewrite with canonlist in Heq3.
invert (trco_truncate _ _ (eqconv Heq3 Htrco)).
intros l' p' _ _ _ His3' _ Hinit3' _ Hviscos _.
eassert Prop as Heqp; [refine (init_unique _#4 Htrco Hinit3 _) |].
  subst H.
  apply in_or_app; right; right; eassumption.
subst p'.
assert (ac_read l = ac_read l') as Heqread.
  eapply is_fun; eauto; [].
  subst H.
  apply in_or_app; right; right; eassumption.
injection Heqread; clear Heqread; intros; subst l'.  
clear Hinit3' His3'.
exploit (Hviscos i1) as [Hcos | Hubcos]; clear Hviscos; [| |].
  assert ((H3b; ev_exec i3) {{ ev_exec i2 }}) as Hexec2_3.
    subst H.
    eapply (exec_tail_to' _ (nil; ev_rf i0 i3;; H3a')); autorewrite with canonlist; eauto; [].
    left; reflexivity.
  split.
    exists i2.
    do2 2 split; auto; [|].
      subst H.
      apply (vos_to_truncate _ (nil; ev_rf i0 i3;; H3a') _ _ i3); autorewrite with canonlist; auto; [].
      apply (to_execed_truncate _ (nil; ev_rf i0 i3;; H3a')); autorewrite with canonlist; auto; [].
      left; reflexivity.

      left.
      subst H.
      apply (init_execed_truncate _ (nil; ev_rf i0 i3;; H3a')); autorewrite with canonlist; auto; done.

    destruct Hwrite1 as (v1 & His1).
    exists v1.
    subst H.
    apply (is_execed_truncate _ (nil; ev_rf i0 i3;; H3a')); autorewrite with canonlist; auto; [].
    eapply (exec_tail_toq' _ (nil; ev_rf i0 i3;; H3a')); autorewrite with canonlist; eauto; [].
    eapply vos_toq; eauto; [].
    apply in_or_app; right; right; assumption.

  rewrite -> Heq3.
  apply cos_app_incl; apply cos_cons_incl; assumption.

  exfalso.
  assert (ubcos H i3 i1) as Hubcos'.
    rewrite -> Heq3.
    apply ubcos_app_incl; apply ubcos_cons_incl; assumption.
  clear Hubcos H3b H3a' Heq3.
  destruct Hubcos' as (i4 & Hrwo & Hcos41).
  so (rwo_wf_prop_1 _#4 Htrco Hrwo Hread3) as Hwrite4.
  assert (~ poq H i4 i1) as Hnpoq41.
    (* case 1 *)
    intros Hpoq41.
    cut (poq H i1 i2).
      intro Hpoq12.
      destruct (po_irreflex H i1); auto; [].
      eapply poq_po_trans; eauto; [].
      eapply po_trans; eauto; [].
      eapply po_poq_trans; eauto; [].
      apply rwo_po; auto; done.
    so (to_execed_fst _#3 Hto23) as Hexec2.
    so (in_split _#3 Hexec2) as (H2a & H2b & Heq2).
    (* This part is messy. *)
    assert (exists Hh H2a', H2a = Hh;; H2a' /\ vos (H2b; ev_exec i2;; Hh) i1 i2 /\ ~ Hh {{ ev_rel i1 }}) as Hprop.
      so (list_invert_snoc _ H2a) as [-> | (H2a' & h & ->)].
        exfalso.
        so (exec_back_to _#5 Heq2 Htrco Hto23) as Hinnil.
        destruct Hinnil; done.
      autorewrite with canonlist in Heq2.
      so (eq_dec_decidable _ eq_event_dec h (ev_rel i1)) as [Heqh | Hneqh].
        exists nil, (nil; h;; H2a').
        subst h.
        autorewrite with canonlist.
        do2 2 split; auto; [].
        apply (vos_tail _ (ev_rel i1)); try discform; [prove_trco |].
        apply (vos_execed_truncate _ _ H2a'); [prove_trco | |].
          subst H; assumption.
          
          left; reflexivity.
          
        exists (nil; h), H2a'.
        autorewrite with canonlist.
        do2 2 split; auto; [|].
          apply (vos_execed_truncate _ _ H2a'); [prove_trco | |].
            subst H; assumption.

            left; reflexivity.

          intro Hin.
          destruct Hin; auto; done.
    (* Back to business. *)
    destruct Hprop as (Hh & H2a' & -> & Hvos12' & Hnrel).
    autorewrite with canonlist in Heq2.
    exploit (vos_rel_or_poq (H2b; ev_exec i2;; Hh) i1 i2) as [Hrel1 | Hpoq]; eauto; try prove_trco; [| |].
      destruct Hwrite1 as (v1 & His1).
      exists l, v1.
      apply (is_execed_truncate _ H2a'); [prove_trco | |].
        eapply vos_execed_2; eauto; [prove_trco |].
        apply in_or_app; right; left; reflexivity.

        subst H; assumption.

      exfalso.
      discriminate_in Hrel1; [|].
        intro; contradiction.
        
        intro Hrel1'.
        assert (H2b {{ ev_exec i1 }}) as Hexec1 by (eapply rel_execed; eauto; prove_trco).
        cut (H2b {{ ev_rel i4 }}).
          intro Hrel4.
          assert (H2b {{ ev_exec i4 }}) as Hexec4 by (eapply rel_execed; eauto; prove_trco).
          invert (trco_truncate _ _ (trco_truncate _ _ (eqconv Heq2 Htrco))).
          intros _ _ _ Hnexec2 _ _.
          destruct Hnexec2.
          apply (xoprwo_rel_execed _ _ i3 i4); auto; [prove_trco | |].
            apply (xop_po_truncate _ (nil; ev_exec i2;; Hh;; H2a') _ _ i4); autorewrite with canonlist; rewrite <- ?Heq2; eauto; [].
            apply (po_execed_truncate _ (nil; ev_exec i2;; Hh;; H2a')); autorewrite with canonlist; rewrite <- ?Heq2; eauto; [].
            apply rwo_po; auto; done.
            
            apply (rwo_execed_truncate _ (nil; ev_exec i2;; Hh;; H2a')); autorewrite with canonlist; rewrite <- ?Heq2; eauto; done.
        destruct Hpoq41 as [Heq | Hpo41].
          subst; assumption.
        assert (po H2b i4 i1) as Hpo41' by (apply (po_execed_truncate _ (nil; ev_exec i2;; Hh;; H2a')); autorewrite with canonlist; rewrite <- ?Heq2; eauto).
        apply (po_rel_released _ _ i1 l); auto; [prove_trco | |].
          destruct Hwrite4 as (v4 & His4).
          exists v4.
          apply (is_po_truncate _ (nil; ev_exec i2;; Hh;; H2a') _ i1); autorewrite with canonlist; rewrite <- ?Heq2; eauto; done.

          destruct Hwrite1 as (v1 & His1).
          exists v1.
          apply (is_execed_truncate _ (nil; ev_exec i2;; Hh;; H2a')); autorewrite with canonlist; rewrite <- ?Heq2; eauto; done.

      rewrite -> Heq2.
      apply poq_app_incl; assumption.
  so (cos_rel_or_poq _#3 Htrco Hcos41) as [Hrel4 |]; [| contradiction].
  so (in_split _#3 Hrel4) as (H4a & H4b & Heq4).
  assert (H4b {{ ev_exec i4 }}) as Hexec4_4.
    invert (trco_truncate _ _ (eqconv Heq4 Htrco)); intros; assumption.
  assert (H4b {{ ev_exec i2 }}) as Hexec2_4.
    assert (trco (H4b; ev_rel i4)) as Htrco4bh by (subst; prove_trco).
    cut ((H4b; ev_rel i4) {{ ev_exec i2 }}).
      intros [Heqcontra | Hexec].
        discriminate Heqcontra.

        assumption.
    apply (xoprwo_rel_execed _ _ i3 i4); auto; [| |].
      apply (xop_po_truncate _ H4a _ _ i4); autorewrite with canonlist; rewrite <- ?Heq4; eauto; [].
      apply (po_execed_truncate _ H4a); autorewrite with canonlist; rewrite <- ?Heq4; eauto.
        right; assumption.

        apply rwo_po; assumption.

      apply (rwo_execed_truncate _ H4a); autorewrite with canonlist; rewrite <- ?Heq4; eauto; [].
      right; assumption.

      left; reflexivity.
    assert (H {{ ev_init i4 p }}) as Hinit4.
      eapply po_init_1; eauto; [].
      apply rwo_po; auto; done.
    eassert Prop as Hcond; [refine (rel_prior_cop_trichot _#3 i1 _#3 Htrco Heq4 Hinit4 Hwrite4 _) |]. 
      split.
        exists i2.
        do2 2 split; auto; [|].
          apply (vos_execed_tail _ (ev_rel i4)); auto; [| |].
            subst; prove_trco.

            apply (vos_execed_truncate _ _ H4a); rewrite <- ?Heq4; auto; done.

            intros (ii & Heqcontra); discriminate Heqcontra.

          left.
          apply (init_execed_truncate _ (nil; ev_rel i4;; H4a)); autorewrite with canonlist; rewrite <- ?Heq4; auto; done.

        destruct Hwrite1 as (v1 & His1).
        exists v1.
        apply (is_execed_truncate _ (nil; ev_rel i4;; H4a)); autorewrite with canonlist; rewrite <- ?Heq4; auto; [].
        so (to_execed_fst _#3 Hto23) as Hexec2.
        so (vos_toq _#3 Htrco Hvos12 Hexec2) as Htoq12.
        destruct Htoq12 as [Heq | Hto12].
          subst; assumption.
        apply (to_execed_fst _ _ i2); [].
        apply (to_execed_truncate _ (nil; ev_rel i4;; H4a)); autorewrite with canonlist; rewrite <- ?Heq4; auto; done.
    destruct Hcond as [Hcop41 | [Heq14 | Hcop14]].
      exploit (cop_rel_or_po H4b i4 i1) as [Hrel4' | Hpo41]; auto; [prove_trco | | |].
        destruct Hwrite4 as (v & His4).
        exists l, v.
        apply (is_execed_truncate _ (nil; ev_rel i4;; H4a)); autorewrite with canonlist; subst H; auto; done.

        invert (trco_truncate _ _ (eqconv Heq4 Htrco)).
        intros _ _ _ _ _ _ _ _ _ _ Hnrel.
        contradiction.

        (* refer to case 1 *)
        destruct Hnpoq41.
        right.
        subst H.
        apply po_app_incl; apply po_cons_incl; assumption.

      (* refer to case 1 *)
      subst i4.
      destruct Hnpoq41.
      left; reflexivity.

      destruct (trco_acyclic H Htrco).
      exists i1.
      eapply plus_star_trans; eauto; [].
      subst H.
      apply cop_app_incl; apply cop_cons_incl; assumption.
Qed.
